/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/ 
#include "FMDB.h"
#include "mMesh.h"
#include "mEntityGroup.h"

struct debug_struct
{
  int i;
  int j;
  int source_partid;
  mEntity* ent1;
  mEntity* ent2;
};

int FMDB_Mesh_Verify (pMeshMdl mesh, int *isValid)
{
  *isValid=1;
  pPart part;

  pmEntity* pe;
  mEntity *vertex, *vertex0, *vertex1;
  mEntity* edge;
  mEntity* face;
  mEntity *region, *region0, *region1;
  int ret=SCUtil_SUCCESS, gLevel, numAdjE, numAdjF, numAdjR, mypid=SCUTIL_CommRank(), partDim;
  double value0, value1;
  int numPart=1, meshDim;
  int intersectFaces=0,  sameSidedRegionPair=0, vtxEdgeAdjMismatch=0, edgeFaceAdjMismatch=0, faceRgAdjMismatch=0, 
      negVol=0, gTagMismatch=0, gTypeMismatch=0, nonManifold=0, gclasMismatch=0, pclasMismatch=0, rmtCopyMismatch=0, 
      psetWithNoPO=0, psetWithV=0,  psetWithE=0,  psetWithF=0, psetMismatch=0, entWithNoPsetAttached=0, 
      numV, numE, numF, numR;

  FMDB_Mesh_GetNumPart(mesh, &numPart);
  FMDB_Mesh_GetDim(mesh, &meshDim);

#ifdef FMDB_PARALLEL  // data for communication
    int numRC;

    IPComMan *CM = ParUtil::Instance()->ComMan();
    CM->set_comm_validation(IPComMan::Neighbors);

    CM->set_tag(0);
    CM->set_fixed_msg_size(sizeof(debug_struct));
    debug_struct* msg_send = (debug_struct*)CM->alloc_msg(sizeof(debug_struct));
#endif
  void* pset_ptr;

  for (mMesh::partIter pit=mesh->partBegin(); pit!=mesh->partEnd(); ++pit)
  {
    part = *pit;  
    FMDB_Part_GetDim (part, &partDim);

    // check p-set
    for(mPart::psetIter pset_it=part->beginPSet(); pset_it!=part->endPSet(); ++pset_it)
    {
      ((mEntityGroup*)(*pset_it))->getNumEntity(&numV, &numE, &numF, &numR);
      if (numV) ++psetWithV;
      if (numE) ++psetWithE;
      if (meshDim==3 && numF) ++psetWithF;
      if (meshDim==2 && numV+numE) ++psetWithNoPO;
      if (meshDim==3 && numV+numE+numF) ++psetWithNoPO;

      for (mEntityGroup::entIter ent_it=((mEntityGroup*)(*pset_it))->begin();
          ent_it!=((mEntityGroup*)(*pset_it))->end(); ++ent_it)
      {
        if (!EN_getDataPtr(*ent_it, FMDB_Util::Instance()->getPSet(), &pset_ptr))
          ++entWithNoPsetAttached;
        else if (((mEntityGroup*)pset_ptr)!=((mEntityGroup*)(*pset_it)))
          ++psetMismatch;
      }
    }  

  // check vertices
  for (mPart::iterall it=part->beginall(0); it!=part->endall(0);++it)
  {
    vertex = *it;
    numAdjE=V_numEdges(vertex);
    if (partDim==3 && numAdjE<3)
    {  
      ++vtxEdgeAdjMismatch;
      if (numAdjE==0) nonManifold=1;
    }
    if (partDim==2 && numAdjE==0)
    {  
      ++vtxEdgeAdjMismatch;
      nonManifold=1;
    }

    pe = vertex->getPClassification();
#ifndef FMDB_PARALLEL 
    assert(!pe);
#else 
    numRC = vertex->getNumRemoteCopies();     
    if (!pe)
    { 
      if (numRC) // non-part bdry vertex
      ++pclasMismatch; 
    }
    else // part bdry vertex
    {
      if (pe && (numRC+1!=pe->getNumBPs()))  // CB vertex
        ++rmtCopyMismatch;
      for (mEntity::RCIter rciter = vertex->rcBegin();
          rciter!=vertex->rcEnd(); ++rciter)
      {
        msg_send->ent1 = vertex;
        msg_send->ent2 = rciter->second;
        msg_send->i=GEN_type(vertex->getClassification());
	msg_send->j=GEN_tag(vertex->getClassification());
        msg_send->source_partid=FMDB_Part_GlobID(part);
        CM->send(rciter->first/numPart, (void*)msg_send);
      } // for rciter
   } // else
#endif    
  }
  
  // check edges
  for (mPart::iterall it=part->beginall(1); it!=part->endall(1);++it)
  {
    edge = *it;
    numAdjF = E_numFaces(edge);
    
    if (partDim==3)
    {
      switch (numAdjF)
      { 
        case 0: ++edgeFaceAdjMismatch;
#ifdef DEBUG
               cout<<"["<<SCUTIL_CommRank()<<"] "<<__func__<<" "<<edge->getUid()<<" has 0 adj face\n";
#endif
	       nonManifold=1;
	       break;
        case 1: if (edge->getPClassification()==NULL && edge->getGhostIndex()<0)	       
               {
	         ++edgeFaceAdjMismatch;
#ifdef DEBUG
               cout<<"["<<SCUTIL_CommRank()<<"] "<<__func__<<" "<<edge->getUid()<<" is internal and adjacent to "<<edge->get(2,1)->getUid()<<"\n";
#endif
                 nonManifold=1;
	       }
	       break;
        default: break;
      } 
      gLevel=GEN_type(edge->getClassification());
      if (gLevel==1 || gLevel==2)
      {
        int bf_counter=0; // number of boundary faces adjacent
        for (int i=0; i<numAdjF; ++i)
        {
	  face = edge->get(2,i);
  	  if (GEN_type(face->getClassification())==2)
	    bf_counter++;
        }
        if (bf_counter!=2 && edge->getPClassification()==NULL && edge->getGhostIndex()<0)
        {
  	  ++edgeFaceAdjMismatch;
          nonManifold=1;
        }
      } 
    }
    else if (partDim==2) // edge in 2D
    {
      if (numAdjF==0)
      {  
        ++edgeFaceAdjMismatch;
#ifdef DEBUG
        cout<<"["<<SCUTIL_CommRank()<<"] "<<__func__<<" "<<edge->getUid()<<" has 0 adj face\n";
#endif
        nonManifold=1;
      }
    
      gLevel=GEN_type(edge->getClassification());
      switch(gLevel)
      {
        case 1: // model edge
   	        if (numAdjF!=1) 
		{
  		  ++edgeFaceAdjMismatch;
#ifdef DEBUG
                  cout<<"["<<SCUTIL_CommRank()<<"] "<<__func__<<" "<<edge->getUid()<<" classified on model edge and # adj face="<<numAdjF<<"\n";
#endif
                  nonManifold=1;
	        }
		break;
        case 2:  if (edge->getPClassification() && edge->getGhostIndex()<0 && numAdjF!=1) // CB edge
                 {
                   ++edgeFaceAdjMismatch;
#ifdef DEBUG
                  cout<<"["<<SCUTIL_CommRank()<<"] "<<__func__<<" "<<edge->getUid()<<" is on part bdry classified and # adj face="<<numAdjF<<"\n";
#endif

                 }
	      break;
        default: nonManifold=1;
                 break;
      } // switch
    } // else
    
    pe = edge->getPClassification();
#ifndef FMDB_PARALLEL 
    assert(!pe);
#else 
    numRC = edge->getNumRemoteCopies();     
    if (!pe)
    { 
      if (numRC) // non-CB vertex
      ++pclasMismatch;
    }
    else  // CB vertex
    {
      if (numRC+1!=pe->getNumBPs())
        ++rmtCopyMismatch;
      for (mEntity::RCIter rciter = edge->rcBegin();
          rciter!=edge->rcEnd(); ++rciter)
      {
        msg_send->ent1 = edge;
        msg_send->ent2 = rciter->second;
        msg_send->i = GEN_type(edge->getClassification());
        msg_send->j = GEN_tag(edge->getClassification());
        msg_send->source_partid=FMDB_Part_GlobID(part);
        CM->send(rciter->first/numPart, (void*)msg_send);
      } // for rciter
    } // else
#endif     
  } // edges
 
  // check faces
  for (mPart::iterall it=part->beginall(2); it!=part->endall(2);++it)
  {
    face = *it;
#ifdef MTOOL
    IntersectionData idata;
    for (int k=0; k<3; ++k)
    {
       vtx = face->get(0,k);
       pPList vfaces=V_faces(vtx);
       iter=0;
       while ((adjfc=(mEntity*)PList_next(vfaces, &iter)))
       {
         if (face==adjfc) continue;
         idata = M_intersectionData_new() ;
	 if (1==M_intersectFaces(face, 0, adjfc, 0, &idata))
	   ++intersectFaces;
         M_intersectionData_delete(idata);	       
       }
       PList_delete(vfaces);
    }
#endif    
    numAdjR = F_numRegions(face);
 
    if (partDim==3)
    {
      gLevel=GEN_type(face->getClassification());
      switch(gLevel)
      {
        case 2: // face on model face
               if (numAdjR!=1)
	       {
                 ++faceRgAdjMismatch;
	         nonManifold=1;
               }
	       break;
        case 3: // interior face
               if (face->getPClassification()) // CB face
	       {
  	         if (numAdjR!=1)
                   ++faceRgAdjMismatch;
  	       } 
	       else if (face->getGhostIndex()<0)
	       {  
	         if (F_numRegions(face)!=2) 
                   nonManifold=1;
                 if (F_numRegions(face) == 0)
                   continue;
// check if two regions are on the same side;
/*
                 region0 = face->get(3,0);
                 if (F_numRegions(face)==2)
                 { 
		   region1 = face->get(3,1);
                   vertex0 = findVertexOppositeFace(region0, face);
                   vertex1 = findVertexOppositeFace(region1, face);
                   value0 = computeSignedDistance(vertex0, face);
                   value1 = computeSignedDistance(vertex1, face);

                   if (fabs(value0)>1E-5 && fabs(value1)>1E-5 && value0*value1>0.0 ) 
                    ++sameSidedRegionPair;
                 }
*/
	       }
               break;
        default: ++gclasMismatch;
                 break;
      }
    } // face in 3D
    else if (partDim==2)
    {
      if (numAdjR!=0)
        ++faceRgAdjMismatch;
      if (GEN_type(face->getClassification())!=2) 
        ++gclasMismatch;
    } // face in 2D

    pe = face->getPClassification();
#ifndef FMDB_PARALLEL 
    assert(!pe);
#else 
    numRC = face->getNumRemoteCopies();     
    if (!pe)
    {
      if (numRC) // non-CB vertex
      ++pclasMismatch;
    }
    else  // CB vertex
    {
      if (numRC+1!=pe->getNumBPs())
        ++rmtCopyMismatch;

      for (mEntity::RCIter rciter = face->rcBegin();
          rciter!=face->rcEnd(); ++rciter)
      {
        msg_send->ent1 = face;
        msg_send->ent2 = rciter->second;
        msg_send->i = GEN_type(face->getClassification());
        msg_send->j = GEN_tag(face->getClassification());
        msg_send->source_partid=FMDB_Part_GlobID(part);
        CM->send(rciter->first/numPart, (void*)msg_send);
      } // for rciter
    } // else
#endif         
  }
  
   // check regions
  for (mPart::iterall it=part->beginall(3); it!=part->endall(3);++it)
  {
    region = *it;
    assert(!(region->getPClassification()));
    assert(GEN_type(region->getClassification())==3);
    if (R_Volume2(region)<0.0)
      ++negVol;
  }
  } // for each part

#ifdef FMDB_PARALLEL
// check partition boundary
  CM->finalize_send();
  CM->free_msg(msg_send);
  void *msg_recv;
  int pid_from;
  while (CM->receive(msg_recv, &pid_from))
  {
    debug_struct* castbuf = (debug_struct*) msg_recv;
    if (GEN_type(castbuf->ent2->getClassification())!=castbuf->i)
      ++gTypeMismatch;
    if (GEN_tag(castbuf->ent2->getClassification())!=castbuf->j)
      ++gTagMismatch;
    CM->free_msg(msg_recv);
  }  // end of while(!AP...)
#endif
  
  int local_error=SCUtil_SUCCESS, global_error; 
  if (intersectFaces+sameSidedRegionPair+vtxEdgeAdjMismatch+edgeFaceAdjMismatch+faceRgAdjMismatch+
      negVol+gTagMismatch+gTypeMismatch+nonManifold+gclasMismatch+pclasMismatch+rmtCopyMismatch+
      entWithNoPsetAttached+psetMismatch+psetWithNoPO)
    local_error=1;
#ifdef FMDB_PARALLEL
  MPI_Allreduce(&local_error, &global_error, 1, MPI_INT, MPI_SUM, ParUtil::Instance()->getComm());
#else
   global_error=local_error;
#endif

  if (local_error)       
    cout<<"["<<SCUTIL_CommRank()<<"] intersectingFaces="<<intersectFaces<<", sameSidedRegionPair="<<sameSidedRegionPair<<", vtxEdgeAdjMismatch="<<vtxEdgeAdjMismatch<<", edgeFaceAdjMismatch="<<edgeFaceAdjMismatch<<", faceRgAdjMismatch="<<faceRgAdjMismatch<<",\n\tnegativeVol="<<negVol<<", geomIDMismatch="<<gTagMismatch<<", geomTypeMismatch="<<gTypeMismatch<<", nonManifold="<<nonManifold<<", geomClasMismatch="<<gclasMismatch<<", ptnClasMismatch="<<pclasMismatch<<",\n\trmtCopyMismatch="<<rmtCopyMismatch
        <<",\n\t entWithNoPsetAttached="<<entWithNoPsetAttached<<", psetMismatch="<<psetMismatch<<", psetWithVertex="<<psetWithV<<", psetWithEdge="<<psetWithE<<", psetWithFace="<<psetWithF<<endl; 

  if (global_error)
  {
    *isValid=0;
    if (!SCUTIL_CommRank())
      cout<<"\n[MESH VERIFICATION - FAIL]\n";
  }
  else 
  {
    if (!SCUTIL_CommRank())   
      cout<<"\n[MESH VERIFICATION - SUCCESS]\n";
  }

  return SCUtil_SUCCESS;
}

