/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "FMDB.h"
#include "mPart.h"
#include "mMesh.h"
#include "oldFMDB.h"
#include "FMDB_Internals.h"
#include "iUtil.h"
#include "mMeshIO.h"
#include <algorithm>
#include <assert.h>
#include <list>
#include <string.h>
#include "FMDB_cint.h"
#include "pmMigration.h"
#include "pmZoltanCallbacks.h"
/**********************************************
      DISRIBUTED MESH MANAGEMENT FUNCTIONS
***********************************************/

int FMDB_Mesh_CreatePart (pMeshMdl mesh, pPart& part)
{
  int num_part = 1;
#ifdef FMDB_PARALLEL
  num_part = ParUtil::Instance()->getTgtNumParts();
#endif

  // whenever part is manually created/deleted, part id shall be synch'ed through FMDB_Mesh_SyncPtn
  part = new mPart(SCUTIL_CommRank()*num_part+mesh->getNumPart(), mesh->getGeomModel(), mesh);
  mesh->addPart(part);
  return SCUtil_SUCCESS;
}


int FMDB_Mesh_DelPart (pMeshMdl mesh, pPart part)
{
  int numEnt = part->size(0)+part->size(1)+part->size(2)+part->size(3);
  if (mesh->getNumPart()==1 || numEnt)
    return SCUtil_FAILURE;
  mesh->delPart(part);
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_GetPart(pMeshMdl mesh, int ith, pPart& part)
{
  part = mesh->getPart(ith);
  if (!part) 
    return SCUtil_FAILURE;
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_GetNumPart (pMeshMdl mesh, int* numPart)
{
  *numPart = mesh->getNumPart();
  return SCUtil_SUCCESS; 
}

int FMDB_Mesh_SetNumPart (pMeshMdl mesh, int numPart)
{
  if(numPart < (int) mesh->getNumPart())  
    return SCUtil_INVALID_ARGUMENT;

#ifdef FMDB_PARALLEL
  ParUtil::Instance()->setTgtNumParts(numPart); 
#endif
  // update part ids' of existing parts based on the new TgtNumParts
  int counter=0;
  for (mMesh::partIter pit=mesh->partBegin(); pit!=mesh->partEnd(); ++pit)
  {
    (*pit)->setId(SCUTIL_CommRank()*numPart+counter);
    counter++;
  }

  // check # tgt parts changed and create new parts if so
  int curNumPart = mesh->getNumPart(); //== ParUtil::Instance()->getCurNumParts(); 
  for (int i=0; i<numPart-curNumPart; ++i)
  {
    pPart tmpPart;
    FMDB_Mesh_CreatePart(mesh, tmpPart);
  }
//  if (syncMesh)
//    FMDB_Mesh_SyncPtn(mesh);
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_GetAllPart(pMeshMdl mesh, int procId, std::vector<pPart>& vecPart)
{
  if(procId!=SCUTIL_CommRank())
  {
    cout<<__func__<<" not implemented yet with non-local process\n";
    return SCUtil_NOT_SUPPORTED;   
  }

  std::copy (mesh->partBegin(), mesh->partEnd(), std::back_inserter (vecPart));
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_IsPtnSynced(pMeshMdl mesh, int* synced)
{
  *synced=1;
#ifdef FMDB_PARALLEL
  if (ParUtil::Instance()->getTgtNumParts() != ParUtil::Instance()->getCurNumParts())
    *synced=0;
#endif
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_SyncPtn (pMeshMdl mesh)
{
#ifdef FMDB_PARALLEL
  int tgtNumParts = ParUtil::Instance()->getTgtNumParts();
  int curNumParts = ParUtil::Instance()->getCurNumParts();

  // STEP 2: reassign global part id 
  int counter=0;
  std::vector<std::pair<int, pMeshEnt> > vecRmtCopy;
  std::vector<int> ResPids;

  for (mMesh::partIter pit=mesh->partBegin(); pit!=mesh->partEnd(); ++pit)
  {
    (*pit)->setId(SCUTIL_CommRank()*ParUtil::Instance()->getTgtNumParts()+counter);
    counter++;
    // STEP 3: update part boudary entities' remote copy and partition classification
    for (int dim = 0; dim<ParUtil::Instance()->get_maxMeshDim(); ++dim)
    {
      for(mPart::iterall eit=(*pit)->beginall(dim); eit!=(*pit)->endall(dim); ++eit)
      {
        if ((*eit)->getPClassification()==NULL) continue;

        if (tgtNumParts != curNumParts)  // update remote copy part id
        {
          vecRmtCopy.clear();
          FMDB_Ent_GetAllRmt (*eit, vecRmtCopy);
          FMDB_Ent_ClrRmt (*eit);
          for (std::vector<std::pair<int, pMeshEnt> >::iterator rc_iter=vecRmtCopy.begin(); rc_iter!=vecRmtCopy.end(); ++rc_iter)
            FMDB_Ent_SetRmt(*eit, (rc_iter->first/curNumParts)*tgtNumParts+rc_iter->first%curNumParts, rc_iter->second);
        } // if (globalMaxNumParts != curNumParts)
        FMDB_Ent_SetPtnTopo(*eit, *pit);  // set partition classification
      } // for each entity
    } // for each dimension
  } // for each part

  // STEP 3: set IPComMan's nbor process info
  vector<int> NborProc;
  FMDB_Mesh_GetNborProcID (mesh, NborProc);
  std::set<int> SetNborProc;
  std::copy(NborProc.begin(), NborProc.end(), inserter(SetNborProc, SetNborProc.end()));
  ParUtil::Instance()->ComMan()->set_Nbrs(SetNborProc);

  // STEP 4: update ownership info
  pmModel::Instance()->updateOwnership(mesh);
  ParUtil::Instance()->setCurNumParts(tgtNumParts); 
#endif 
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_GetAllPartID (pMeshMdl mesh, int procId, std::vector<int>& vecPartID)
{
  cout<<__func__<<" not implemented yet\n";
  return SCUtil_FAILURE;
}

int FMDB_Mesh_LocalPtn(pMeshMdl mesh, pmMigrationCallbacks& migrCB)
{
  cout<<__func__<<" not implemented yet\n";
  return SCUtil_NOT_SUPPORTED;
}

int FMDB_Mesh_SetPtnParam (pMeshMdl mesh, int LBMethod, int LBApproach,
         double imbalance_tol, int debug_level)
{
#ifdef FMDB_PARALLEL
  mesh->LB_Method=LBMethod;
  mesh->LB_Approach=LBApproach;
  mesh->imbalance_tol = imbalance_tol;
  mesh->debug_level=debug_level;
#endif
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_GlobPtn(pMeshMdl mesh, pmMigrationCallbacks& migrCB, int ptnMethod, int ptnApproach,
                      double imbalance_tol, int debug_level)
{
  int err = SCUtil_SUCCESS;
#ifdef FMDB_PARALLEL
  std::map<pMeshEnt, std::pair<int, int> > EntToMigr;
  std::map<pPartSet, std::pair<int, int> > SetToMigr;
  getGlobPtnInfo (mesh, ptnMethod, ptnApproach, imbalance_tol, debug_level, EntToMigr, SetToMigr);
  FMDB_Mesh_Migr(mesh, migrCB, EntToMigr, SetToMigr);
#endif
  return err;
}

int FMDB_Mesh_SerialGlobPtn(pMeshMdl mesh, int ptnMethod, int ptnApproach,
                      double imbalance_tol, int debug_level)
{
  int err = SCUtil_SUCCESS;
#ifdef FMDB_PARALLEL
  if (SCUTIL_CommSize()>1)
    err = SCUtil_NOT_SUPPORTED;
  else
  {
    static_getGlobPtnInfo (mesh, ptnMethod, ptnApproach, imbalance_tol, debug_level);
    err = serial_migratePtnObject(mesh);
  }
#endif
  return err;
}

// *****************************************
int FMDB_Mesh_Migr (pMeshMdl mesh, pmMigrationCallbacks& migrCB,
                    map<pMeshEnt, std::pair<int/*source part*/, int/*dest part*/> >& pentToMigr,
                    map<pPartSet, std::pair<int, int> >& psetToMigr)
// *****************************************
{
  int err = SCUtil_SUCCESS;
#ifdef FMDB_PARALLEL
  std::vector<pMeshEnt> newEnt[4]; 
  std::vector<pMeshEnt> rmvEnt[4]; 
  std::vector<pPartSet> *newPSet = new std::vector<pPartSet>(mesh->getNumPart());
  std::vector<pPartSet> *rmvPSet = new std::vector<pPartSet>(mesh->getNumPart());

  if (ParUtil::Instance()->getTgtNumParts() != ParUtil::Instance()->getCurNumParts())
    FMDB_Mesh_SyncPtn(mesh); // update part boundary (remote part id) and p-classification

  err = migratePtnObject(mesh, migrCB, pentToMigr, psetToMigr, 0, newEnt, rmvEnt, newPSet, rmvPSet);
  delete newPSet;
  delete rmvPSet;
#endif
  return err;
}

int FMDB_Mesh_MigrAdv (pMeshMdl mesh,
                       pmMigrationCallbacks& migrCB,
                       map<pMeshEnt, std::pair<int, int> >& entToMigr,
                       map<pPartSet, std::pair<int, int> >& psetToMigr,
                       std::vector<pMeshEnt>* newEnt,
                       std::vector<pMeshEnt>* rmvEnt,
                       std::vector<pPartSet>* newPSet,
                       std::vector<pPartSet>* rmvPSet)
{
  int err = SCUtil_SUCCESS;
#ifdef FMDB_PARALLEL
  if (ParUtil::Instance()->getTgtNumParts() != ParUtil::Instance()->getCurNumParts())
    FMDB_Mesh_SyncPtn(mesh); // update part boundary (remote part id) and p-classification

  err = migratePtnObject(mesh, migrCB, entToMigr, psetToMigr, 1, newEnt, rmvEnt, newPSet, rmvPSet);
#endif
  return err;
}

int FMDB_Mesh_PullPartBdryToOwner (pMeshMdl mesh, 
		       pmMigrationCallbacks& migrCB,
                       std::list<pMeshEnt>& partBdryEnt,
		       std::vector<pMeshEnt>* newEnt, 
		       std::vector<pMeshEnt>* rmvEnt,
		       std::vector<pPartSet>* newPSet, 
		       std::vector<pPartSet>* rmvPSet)
{
  int err = SCUtil_SUCCESS;
#ifdef FMDB_PARALLEL
   err = pullPartBdryToOwner(mesh, migrCB, partBdryEnt, newEnt, rmvEnt, newPSet, rmvPSet);
#endif
  return err;
}

int FMDB_Mesh_PullPartBdryForSnapping (pMeshMdl mesh, 
		       pmMigrationCallbacks& migrCB,
                       std::list<pMeshEnt>& partBdryVtx,
                       std::list<std::pair<pEntity,pEntity> > & vtxEntPair,
		       std::vector<pMeshEnt>* newEnt, 
		       std::vector<pMeshEnt>* rmvEnt,
		       std::vector<pPartSet>* newPSet, 
		       std::vector<pPartSet>* rmvPSet)
{
  int err = SCUtil_SUCCESS;
#ifdef FMDB_PARALLEL
    err = pullPartBdryToOwner(mesh, migrCB, partBdryVtx, vtxEntPair, newEnt, rmvEnt, newPSet, rmvPSet);
#endif
  return err;
}

int FMDB_Ent_Migr (pMeshMdl mesh, // in
                   pmMigrationCallbacks& migrCB, // in
                   pMeshEnt ent, int source_partid, int dest_partid,
                   std::vector<pMeshEnt>* newEnts, // out - new entities on src part
                   std::vector<pMeshEnt>* oldEnts, // out - remove entities from dest part
		   std::vector<MPI_Request*>& requests) // out - mpi requests used
{
#ifdef FMDB_PARALLEL
//  return migrateEntity(mesh, ent, source_partid, dest_partid, migrCB, newEnts, oldEnts, requests);
#endif
  return SCUtil_NOT_SUPPORTED;
}

int FMDB_Mesh_Merge (pMeshMdl mesh, pmMigrationCallbacks& migrCB)
{
  int err = SCUtil_SUCCESS;
#ifdef FMDB_PARALLEL
  int mesh_dim;
  FMDB_Mesh_GetDim (mesh, &mesh_dim);

  std::vector<pMeshEnt> newEnt[4];
  std::vector<pMeshEnt> rmvEnt[4];
  std::vector<pPartSet> *newEntSet = new std::vector<pPartSet>(mesh->getNumPart());
  std::vector<pPartSet> *rmvEntSet = new std::vector<pPartSet>(mesh->getNumPart());

  std::map<pMeshEnt, std::pair<int, int> > EntToMigr;
  std::map<pPartSet, std::pair<int, int> > SetToMigr;

  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
  {
    for(mPart::psetIter pset_it=(*part_it)->beginPSet(); pset_it!=(*part_it)->endPSet(); ++pset_it)
      SetToMigr.insert(std::map<mEntityGroup*, std::pair<int, int> >::value_type((mEntityGroup*)(*pset_it), std::make_pair((*part_it)->getId(), 0)));
    for (mPart::iterall ent_it=(*part_it)->beginall(mesh_dim); ent_it!=(*part_it)->endall(mesh_dim); ++ent_it)
      EntToMigr.insert(std::map<mEntity*, std::pair<int, int> >::value_type(*ent_it, std::make_pair((*part_it)->getId(), 0)));
  }

  err = migratePtnObject(mesh, migrCB, EntToMigr, SetToMigr, 0, newEnt, rmvEnt, newEntSet, rmvEntSet);
  delete newEntSet;
  delete rmvEntSet;
#endif
  return err;
}

int FMDB_Mesh_CreateGhost (pMeshMdl mesh, pmMigrationCallbacks& migrCB, int ghostType, int brgType, int numLayer, int includeCopy)
{
#ifdef FMDB_PARALLEL
  createGhost(mesh, migrCB, ghostType, brgType, numLayer, includeCopy);
#endif
  return SCUtil_SUCCESS; 
}

int FMDB_Mesh_DelGhost (pMeshMdl mesh)
{
#ifdef FMDB_PARALLEL
  deleteGhostEntities(mesh);
#endif
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_GetGhostInfo (pMeshMdl mesh, std::vector<int>& ghostInfo)
{
#ifdef FMDB_PARALLEL
  mesh->getGhostInfo(ghostInfo);
#endif
  return SCUtil_SUCCESS; 
}

int FMDB_Topo_SetWeight (pMeshMdl mesh, int topo, double weight)
{
  int mesh_dim, type = FMDB_Topo_GetType(topo);
  FMDB_Mesh_GetDim (mesh, &mesh_dim);
  if (type!=mesh_dim) 
    return SCUtil_INVALID_ENTITY_TOPO;

  switch (topo)
  {
    case FMDB_TRI:
    case FMDB_QUAD:
    case FMDB_TET:
    case FMDB_HEX:
    case FMDB_PRISM:
    case FMDB_PYRAMID: {
                         for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
                           for (mPart::iterall ent_it=(*part_it)->beginall(type); ent_it!=(*part_it)->endall(type); ++ent_it)
                             if ((*ent_it)->getTopo()==topo)
                               (*ent_it)->attachDouble(FMDB_Util::Instance()->getWeight(), weight);
                        }
                        break;
    default: return SCUtil_NOT_SUPPORTED;
  }
  return SCUtil_SUCCESS;
}

int FMDB_Topo_DelWeight (pMeshMdl mesh, int topo)
{
  int mesh_dim, type = FMDB_Topo_GetType(topo);
  FMDB_Mesh_GetDim (mesh, &mesh_dim);
  if (type!=mesh_dim)
    return SCUtil_INVALID_ENTITY_TOPO;

  switch (topo)
  {
    case FMDB_TRI:
    case FMDB_QUAD:
    case FMDB_TET:
    case FMDB_HEX:
    case FMDB_PRISM:
    case FMDB_PYRAMID: {
                         for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
                           for (mPart::iterall ent_it=(*part_it)->beginall(type); ent_it!=(*part_it)->endall(type); ++ent_it)
                             if ((*ent_it)->getTopo()==topo)
                               (*ent_it)->deleteData(FMDB_Util::Instance()->getWeight());
                        }
                        break;
    default: return SCUtil_NOT_SUPPORTED;
  }
  return SCUtil_SUCCESS;
}
