/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "FMDB.h"
#include "mPart.h"
#include "mMesh.h"
#include "SCUtil.h"
#include "GMI.h"
#include "mEntity.h"
#include "mFMDB.h"
#include "FMDB_cint.h"

#include <vector>

using std::vector;
using std::pair;

//************************************
// 4.2 Serial Entity Information
//************************************
// return integer ID generated by FMDB_Part_SetID 
int FMDB_Ent_ID (pMeshEnt ent)
{
  if(ent->getLevel() == 0)return ent->getId();
  else return ent->getAttachedInt(FMDB_Util::Instance()->getId());
}

// return string ID consisting of vertex id's generated by FMDB_Part_SetID 
std::string FMDB_Ent_StrID(pMeshEnt ent)
{
  return ent->getUid();
}

/* Get dimension of entity ent : 0 for vertex, 1: edge, 2:face, 3:region*/
int FMDB_Ent_GetType(pMeshEnt ent, int* entType)
{
  *entType = ent->getLevel();
  return SCUtil_SUCCESS; 
}

int FMDB_Ent_GetTopo(pMeshEnt ent, int* entTopo)
{
  *entTopo = ent->getTopo();
  return SCUtil_SUCCESS;
}

/* Given mesh entity pMeshEnt, set/get its geometric classification*/
int FMDB_Ent_SetGeomClas (pMeshEnt meshEnt, pGeomEnt geomEnt)
{
  meshEnt->classify(geomEnt);
  return SCUtil_SUCCESS;
}

int FMDB_Ent_GetGeomClas (pMeshEnt meshEnt, pGeomEnt& geomEnt)
{
  if (meshEnt->getClassification()) 
   geomEnt  = meshEnt->getClassification();
  else
    return SCUtil_FAILURE;
  return SCUtil_SUCCESS;
}

int FMDB_Ent_GetGeomClasType (pMeshEnt meshEnt, int* geomType)
{
  *geomType = GEN_type(meshEnt->getClassification());
  return SCUtil_SUCCESS;
}

int FMDB_GeomEnt_GetNumRevClas (pGeomEnt geomEnt, pPart part, int type, int* numEnt)
{
  pPartEntIter iter;
  pMeshEnt ent;
  int num=0;
  int iterEnd = FMDB_PartEntIter_InitRevClas (part, geomEnt, type, iter);
  
  while (!iterEnd)
  {
    iterEnd = FMDB_PartEntIter_GetNext(iter, ent);
    if (iterEnd) break;
    ++num;
  }
  
  FMDB_PartEntIter_Del (iter);
  *numEnt = num;
  return SCUtil_SUCCESS;
}

int FMDB_Ent_GetPSet (pMeshEnt meshEnt, pEntSet& entSet)
{
   void* ptr;
   if (!EN_getDataPtr(meshEnt, FMDB_Util::Instance()->getPSet(), &ptr))
   {
     entSet = NULL; 
     return SCUtil_FAILURE;
   }
   else
   {
     entSet = (pEntSet)ptr;
     return SCUtil_SUCCESS; 
   }
}

//*******************************************
// 4.8 Parallel-specific Entity Information
//*******************************************
int FMDB_Ent_GetResPart(pMeshEnt meshEnt, pPart part, std::vector<int> resPartId)
{
#ifdef FMDB_PARALLEL
  resPartId.push_back(part->getId());
  if (meshEnt->getNumRemoteCopies())
    for (mEntity::RCIter rcIter=meshEnt->rcBegin(); rcIter!=meshEnt->rcEnd();++rcIter)
      resPartId.push_back(rcIter->first);
#endif
  return SCUtil_SUCCESS;
}

// update partition classification based on the remote part information
int FMDB_Ent_SetPtnTopo (pMeshEnt meshEnt, pPart part)
{
#ifdef FMDB_PARALLEL
  std::set<int> resPartId;
  resPartId.insert(part->getId());
  if (meshEnt->getNumRemoteCopies())
    for (mEntity::RCIter rcIter=meshEnt->rcBegin(); rcIter!=meshEnt->rcEnd();++rcIter)
      resPartId.insert(rcIter->first);
  meshEnt->setPClassification(pmModel::Instance()->getPartitionEntity(resPartId));
#endif
  return SCUtil_SUCCESS;  
}

/* Given a mesh entity, check if the entity is on the part boundary or not. If the entity is on part
boundary, isOnPrtBdry is non-zero integer, otherwise, isOnPrtBdry is 0.*/
int FMDB_Ent_IsOnPartBdry (pMeshEnt /* in */ meshEnt, int* /* out */ isOnPrtBdry)
{
  *isOnPrtBdry = 0;
#ifdef FMDB_PARALLEL
  if(meshEnt->getPClassification())
    *isOnPrtBdry = 1;
#endif
  return SCUtil_SUCCESS;
}

/* Given a mesh entity, check if the entity is a ghost copy. If the entity is a ghost copy, isGhost
is non-zero integer, otherwise, isGhost is 0.*/

int FMDB_Ent_IsGhost(pMeshEnt /* in */ meshEnt, pPart part, int* /* out */ isGhost)
{
  *isGhost=0;
 #ifdef FMDB_PARALLEL
  int gIndex=meshEnt->getGhostIndex();
  mGhost* ghost=part->getGhost(gIndex); 
  if(ghost)
   {
    if(ghost->isGhost())
      *isGhost=1;
   }
#endif
  return SCUtil_SUCCESS;
}

/*Given a mesh entity, check if the entity is ghost copied. If the entity has ghost copies on
non-self part, ghostCopied is non-zero integer, otherwise, ghostCopied is 0.*/

int FMDB_Ent_IsGhosted (pMeshEnt /* in */ meshEnt, pPart part, int* /* out */ ghostCopied)
{
  *ghostCopied=0;
 #ifdef FMDB_PARALLEL
  int gIndex=meshEnt->getGhostIndex();
  mGhost* ghost=part->getGhost(gIndex);
  if(ghost)
    {
      if(ghost->hasGhosts())
        *ghostCopied=1;
    }
#endif
  return SCUtil_SUCCESS;
}

/* Get the owning global part id of a given entity (parallel)*/
int FMDB_Ent_GetOwnPartID (pMeshEnt meshEnt, pPart part, int* partID)
{
  *partID = meshEnt->getOwnPartID(part); 
  return SCUtil_SUCCESS;
}

/*Get owner entity of a given entity (parallel)*/
int FMDB_Ent_GetOwnEnt(pMeshEnt meshEnt, pPart part, pMeshEnt& owner)
{
  owner = meshEnt; 
#ifdef FMDB_PARALLEL
  int own_pid;
  FMDB_Ent_GetOwnPartID(meshEnt, part, &own_pid);

  if (meshEnt->getPClassification()) 
  { 
    for (mEntity::RCIter rcIter=meshEnt->rcBegin(); rcIter!=meshEnt->rcEnd();++rcIter) 
    {
      if ( rcIter->first==own_pid ) 
      {
        owner = rcIter->second;
        break;
      }
    } 
  } // if boundary entity
#endif
  return SCUtil_SUCCESS;
}


/*
Given a mesh entity, get the number of remote copies (parallel).
*/
int FMDB_Ent_GetNumRmt (pMeshEnt meshEnt, int* numRmtCopy)
{
#ifdef FMDB_PARALLEL
  *numRmtCopy = meshEnt->getNumRemoteCopies();
#else
  *numRmtCopy = 0;
#endif
  return SCUtil_SUCCESS;
}

/* Given mesh entity, get its remote copies in a vector having remote part ids and remote entity address (parallel)*/
int FMDB_Ent_GetAllRmt (pMeshEnt meshEnt, std::vector<std::pair<int, pMeshEnt> >& vecRmtCopy)
{
#ifdef FMDB_PARALLEL
  if(meshEnt->getNumRemoteCopies() == 0)
    return SCUtil_FAILURE;

  for (mEntity::RCIter rciter=meshEnt->rcBegin();rciter!=meshEnt->rcEnd();++rciter)
    vecRmtCopy.push_back(std::pair<int, pMeshEnt>((int)rciter->first, (pMeshEnt)rciter->second));
#endif
  return SCUtil_SUCCESS; //success
}

/*
Given a mesh entity and a part id, get the remote copies on that part id (parallel).
*/
int FMDB_Ent_GetRmt (pMeshEnt& meshEnt, int destPart, pMeshEnt& rmtCopy)
{
#ifdef FMDB_PARALLEL
  rmtCopy = meshEnt->getRemoteCopy(destPart);   
  if(!rmtCopy)
    return SCUtil_FAILURE; // error code
#endif
  return SCUtil_SUCCESS;
}

/*Given a mesh entity, a remote part id, and an address of remote mesh entity, add the remote
copy to the entity.*/
int FMDB_Ent_SetRmt (pMeshEnt meshEnt, int partID, pMeshEnt rmtMeshEnt)
{
#ifdef FMDB_PARALLEL
  meshEnt->addRemoteCopy(partID, rmtMeshEnt);
#endif
  return SCUtil_SUCCESS;
}

/*Given a mesh entity and a remote part id, delete the remote copy of entity on the remote
part.*/
int FMDB_Ent_DelRmt (pMeshEnt /* in */ meshEnt, int /* in */ partID)
{
#ifdef FMDB_PARALLEL
  meshEnt->deleteRemoteCopy(partID);
#endif
  return SCUtil_SUCCESS;
}

/* Given a mesh entity, clear the remote copy.*/
int FMDB_Ent_ClrRmt (pMeshEnt /* in */ meshEnt)
{
#ifdef FMDB_PARALLEL
  meshEnt->clearRemoteCopies();
#endif
  return SCUtil_SUCCESS;
}

/* Given a mesh entity, get the number of ghost copies.*/
int FMDB_Ent_GetNumGhost(pMeshEnt /* in */ meshEnt, pPart part, int* /* out */ numGhostCopy)
{
  *numGhostCopy=0;
#ifdef FMDB_PARALLEL
  int gIndex=meshEnt->getGhostIndex();
  mGhost* ghost=part->getGhost(gIndex);
  if(ghost)
   {
     *numGhostCopy=ghost->getNumGhostCopies();
   }
  else
   return SCUtil_FAILURE;
#endif
  return SCUtil_SUCCESS;
}

/* Given a mesh entity, get the vector vecGhostCopy filled with pair(s) of global part id and
the memory address of ghost copies.*/
int FMDB_Ent_GetAllGhost (pMeshEnt /* in */ meshEnt, pPart part, 
                          std::vector<std::pair<int, pMeshEnt> >& /* inout */ vecGhostCopy)
{
#ifdef FMDB_PARALLEL
  int gIndex=meshEnt->getGhostIndex();
  mGhost* ghost=part->getGhost(gIndex);
  if(ghost)
   {
     if(ghost->getNumGhostCopies()==0)
       return SCUtil_FAILURE;
     std::map<int, pMeshEnt> ghostMap;
     ghost->getGhostCopies(ghostMap);
     for (mGhost::GCIter gciter=ghost->gcBegin();gciter!=ghost->gcEnd();++gciter)
      {
	vecGhostCopy.push_back(std::pair<int, pMeshEnt>((int)gciter->first, (pMeshEnt)gciter->second));
      }
   }
  else
   return SCUtil_FAILURE;
#endif
  return SCUtil_SUCCESS; //success   
}

int FMDB_Ent_GetGhost(pMeshEnt& meshEnt, int destPart, pMeshEnt& ghostEnt)
{
  cout<<__func__<<" not implemented yet\n";
  return SCUtil_NOT_SUPPORTED; 
}

//************************************
//  4.9 weight
//************************************
int FMDB_Ent_SetWeight (pMeshEnt meshEnt, double weight)
{
  meshEnt->attachDouble(FMDB_Util::Instance()->getWeight(), weight);
  return SCUtil_SUCCESS;
}

int FMDB_Ent_DelWeight (pMeshEnt meshEnt)
{
  meshEnt->deleteData(FMDB_Util::Instance()->getWeight());
  return SCUtil_SUCCESS;
}

int FMDB_Ent_GetWeight (pMeshEnt meshEnt, double* weight)
{
  if (!(meshEnt->getAttachedDouble(FMDB_Util::Instance()->getWeight(), weight)))
    *weight = 1.0; // default
  return SCUtil_SUCCESS;
}


//************************************
//  4.9 Misc
//************************************
// E.g. numDownAdj[FMDB_HEX][FMDB_EDGE] is #downward edges of hexadedron
int numDownAdj[11][3] = {/*point*/ {-1, -1, -1}, /*line*/ {-1, -1, -1}, /*polygon*/{-1, -1, -1},
                          /*tri*/ {3, 3, -1}, /*quad-1*/ {4, 4, -1}, /*polyhedron*/{-1, -1, -1},
                          /*tet*/ {4, 6, 4}, /*hex*/ {8, 12, 6}, /*prism-4*/ {6, 9, 5},
                          /*pyramid-5*/ {5, 8, 5}, /*septa*/ {-1, -1, -1} };

int FMDB_Topo_GetType(int topology)
{
  if (topology==FMDB_POINT || topology==FMDB_LINE)
    return topology;
  else if (FMDB_POLYGON<=topology && topology<=FMDB_QUAD)
    return 2;
  else if (FMDB_POLYHEDRON<=topology && topology<=FMDB_SEPTA)
    return 3;
  return -1;
}

int FMDB_Topo_GetNumDownAdj(int topo, int downType)
{
  return numDownAdj[topo][downType];
}

int FMDB_Topo_GetTopoFromFaces(int numFace, pMeshEnt* faces)
{
  switch (numFace)
  {
    case 4: return FMDB_TET;
    case 6: return FMDB_HEX;
    case 5: {
              int topo=0, countTri=0, countQuad=0;
              for (int i = 0; i < numFace; i++)
              {
                FMDB_Ent_GetTopo(faces[i], &topo);
                if (topo == FMDB_TRI)
                  countTri++;
                if(topo == FMDB_QUAD)
                  countQuad++;
              }
              if (countQuad == 3 && countTri == 2)
                return FMDB_PRISM;
              else if(countQuad == 1 && countTri ==4)
                return FMDB_PYRAMID;
            }
   default:  return -1;
  }
}
