/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "FMDB.h"
#include "mPart.h"
#include "mMesh.h"
#include "mMeshIO.h"
#include "oldFMDB.h"
#include "iUtil.h"
#include <assert.h>
#include <list>
#include "FMDB_cint.h"
#include "mMeshIO.h"

int FMDB_Mesh_GetLastError (pMeshMdl mesh, int *error_type)
{
  *error_type = mesh->last_error;
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_SetLastError (pMeshMdl mesh, int error_type)
{
  mesh->last_error = error_type;
  return SCUtil_SUCCESS;
}
/***************************************
      MESH/PART MANAGEMENT FUNCTIONS
****************************************/
int FMDB_Mesh_Create(pGeomMdl geom, pMeshMdl& mesh)
{
  int numPart=1;
#ifdef FMDB_PARALLEL
  numPart = ParUtil::Instance()->getTgtNumParts();
#endif
  mesh = new mMesh(geom);
  pPart part = new mPart(SCUTIL_CommRank()*numPart+mesh->getNumPart(), mesh->getGeomModel(), mesh);
  mesh->addPart(part);
  return SCUtil_SUCCESS;
} 

int FMDB_Mesh_Del(pMeshMdl mesh)
{
  mesh->clearPart();
  TagHolder_ClearTag (static_cast<pTagHolder>(mesh));
  delete mesh;
  return SCUtil_SUCCESS; //_Err::SUCCESS;
}

// master process load a serial mesh and return
int FMDB_Mesh_LoadSerialFromFile (pMeshMdl mesh, const char* fileName)
{
  char terminatedInputString[MAX_STRING_LENGTH];
  strncpy(terminatedInputString, fileName, MAX_STRING_LENGTH-1);
  terminatedInputString[MAX_STRING_LENGTH-1] = 0;
  char *end = terminatedInputString + strlen(terminatedInputString) - 1;
  while (end >= terminatedInputString &&
         (isspace(*end) || iscntrl(*end)) ) {
    end--;
  }
  *(end + 1) = '\0';

  int err = SCUtil_SUCCESS;
  char ext[6];
  strcpy(ext,terminatedInputString+(strlen(terminatedInputString)-3));

  // load serial mesh
#ifdef DEBUG
  double begin_time=ParUtil::Instance()->getCurTime(), begin_mem=ParUtil::Instance()->getCurMem();
#endif
  err = FMDB_Util::Instance()->import_mesh (mesh, terminatedInputString);
#ifdef DEBUG
  if (!SCUTIL_CommRank())
    cout<<"\nLOADING: Time="<<ParUtil::Instance()->getCurTime()-begin_time<<" (sec), Mem="<<ParUtil::Instance()->getCurMem() - begin_mem<<" (MB)\n";
#endif
  return err;
}

/* 
If distributed = 0, load a serial mesh and partition if mesh has multi parts
if distributed = 1, load a distributed mesh (#total parts >= # mesh files)
*/
int FMDB_Mesh_LoadFromFile (pMeshMdl mesh, const char* fileName, int distributed)
{
  // the following block will guarantee the input string is
  //  null terminated and has no trailing whitespace -  {
  char terminatedInputString[MAX_STRING_LENGTH];
  strncpy(terminatedInputString, fileName, MAX_STRING_LENGTH-1);
  terminatedInputString[MAX_STRING_LENGTH-1] = 0;
  char *end = terminatedInputString + strlen(terminatedInputString) - 1;
  while (end >= terminatedInputString && 
         (isspace(*end) || iscntrl(*end)) ) {
    end--;
  }
  *(end + 1) = '\0';
  // }

  int err = SCUtil_SUCCESS;
  char ext[6];
  strcpy(ext,terminatedInputString+(strlen(terminatedInputString)-3));
  if (distributed) // load partitioned mesh
  {
#ifdef FMDB_PARALLEL
    if(!strcmp(ext,"cdf") || !strcmp(ext, ".nc") || !strcmp(ext,"vtk"))
    {
      cout<<__func__<<" distributed mesh loading is supported with SMS only\n";
      return SCUtil_FAILURE;
    }
#ifdef DEBUG
    double begin_time=ParUtil::Instance()->getCurTime(), begin_mem=ParUtil::Instance()->getCurMem();
#endif
    if (ParUtil::Instance()->getTgtNumParts()==1)
      import_parallel_singlepart_SMS(mesh, terminatedInputString);
    else
      import_parallel_multipart_SMS(mesh, terminatedInputString);
#ifdef DEBUG
 if (!SCUTIL_CommRank())
    cout<<"\nLOADING: Time="<<ParUtil::Instance()->getCurTime()-begin_time<<" (sec), Mem="<<ParUtil::Instance()->getCurMem() - begin_mem<<" (MB)\n";
#endif
#else // serial
    cout<<__func__<<" (distributed=1) not implemented in SERIAL\n";
    return SCUtil_FAILURE;
#endif
  } // distributed=1
  else // load serial mesh and partition
  {
    if (SCUTIL_GrpSize()!=SCUTIL_CommSize())
      return SCUtil_NOT_SUPPORTED;
#ifdef DEBUG
    double begin_time=ParUtil::Instance()->getCurTime(), begin_mem=ParUtil::Instance()->getCurMem();
#endif
    err = FMDB_Util::Instance()->import_mesh (mesh, terminatedInputString);
#ifdef DEBUG
    if (!SCUTIL_CommRank())
      cout<<"\nLOADING: Time="<<ParUtil::Instance()->getCurTime()-begin_time<<" (sec), Mem="<<ParUtil::Instance()->getCurMem() - begin_mem<<" (MB)\n";
#endif
    if (err != SCUtil_SUCCESS)
      return err;
#ifdef FMDB_PARALLEL
    if (ParUtil::Instance()->size()>1)
    {
      static_getGlobPtnInfo (mesh, mesh->LB_Method, mesh->LB_Approach, 
                             mesh->imbalance_tol, mesh->debug_level);
      static_migratePtnObject(mesh);
    }
#endif
  }
  return SCUtil_SUCCESS; 
}

int FMDB_Mesh_WriteToFile (pMeshMdl mesh, const char* fileName, int distributed)
{
  char ext[6];
  strcpy(ext,fileName+(strlen(fileName)-3)); 
  
#ifdef FMDB_PARALLEL
  if (distributed)
  {
    if(!strcmp(ext,"cdf") || !strcmp(ext, ".nc"))
    {
      cout<<__func__<<" is supported only for SMS mesh\n";
      return SCUtil_FAILURE;
    }
    else if(!strcmp(ext,"vtk"))
      exportMeshToVTK(mesh, fileName);
    else
      exportMeshToSMS(mesh,fileName);
  }
  else
#endif
  {
    if (ParUtil::Instance()->rank()==0 &&mesh->getNumPart()>1)
    {
      cout<<__func__<<" with multiple parts not implemented yet\n";
      return SCUtil_FAILURE;
    }

    pPart part = *(mesh->partBegin());
    if(!strcmp(ext,"cdf"))
      exportNCDF(part, mesh->getGeomModel(), fileName, NULL);
#ifdef NCDF64
    else if(!strcmp(ext,".nc"))
      exportNCDFBin(part, mesh->getGeomModel(), fileName, NULL);
#endif/*NCDF64*/
    else if(!strcmp(ext,"vtk"))
      exportMeshToVTK(mesh, fileName);
    else
      exportSMS(mesh, fileName);
  }
  return SCUtil_SUCCESS; //_Err::SUCCESS;
}

int FMDB_Mesh_IsEmpty (pMeshMdl mesh, int* empty)
{
  int global_empty_flag=1, local_empty_flag=1;
  
  for (mMesh::partIter pit=mesh->partBegin(); pit!=mesh->partEnd(); ++pit)
    if ((*pit)->size(0))  // not empty
    {  
      global_empty_flag =0; 
      local_empty_flag=0;
      break;
    }
#ifdef FMDB_PARALLEL
  // do communication for global mesh
  MPI_Comm comm;
  SCUTIL_GetComm (&comm);
  int numProc=SCUTIL_CommSize();
  if (numProc>1) MPI_Allreduce(&local_empty_flag,&global_empty_flag,1,MPI_INT,MPI_MIN,comm);
#endif
  *empty = global_empty_flag;
  return SCUtil_SUCCESS;   
}

int FMDB_Mesh_GetDim (pMeshMdl mesh, int* dim)
{
#ifdef FMDB_PARALLEL
  *dim = ParUtil::Instance()->get_maxMeshDim();
#else
  *dim = (mesh->getPart(0)->size(3)? 3:2);
#endif
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_GetGeomMdl (pMeshMdl mesh, pGeomMdl& geom)
{
  geom = mesh->getGeomModel();
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_GetNumSet (pMeshMdl  mesh, int* numEntSet)
{
  *numEntSet = mesh->allEntSets.size();
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_GetSet (pMeshMdl  mesh, std::vector<pEntSet> &entsets)
{
  std::copy(mesh->allEntSets.begin(), mesh->allEntSets.end(), std::back_inserter(entsets));
  return SCUtil_SUCCESS;
}

// unique ent set ID gets invalid once new entity set is added (e.g. migration, modification)
int FMDB_Mesh_SetSetID (pMeshMdl  mesh)
{
  std::list<mEntitySet*>::iterator eset_it = mesh->beginEntSet();
  for (int eset_cnt=1; eset_it!=mesh->endEntSet(); ++eset_it)
    (*eset_it)->attachInt(FMDB_Util::Instance()->getId(),eset_cnt++); // attach unique ID to ent set
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_DelSetID (pMeshMdl mesh)
{
  mEntitySet* eset;
  std::list<mEntitySet*>::iterator eset_it = mesh->beginEntSet();
  for (; eset_it!=mesh->endEntSet(); ++eset_it)
  {
      eset = *eset_it;
      eset->deleteData(FMDB_Util::Instance()->getId()); // attach unique ID to ent set
  }
  return SCUtil_SUCCESS;
}

int FMDB_Mesh_GetNborProcID (pMeshMdl mesh, std::vector<int>& nbor_proc)
{
#ifdef FMDB_PARALLEL
  int nbor_rank;
  for (pmModel::PEIter peit=pmModel::Instance()->peBegin(); peit!=pmModel::Instance()->peEnd(); ++peit)
    for (set<int>::iterator partid_it= (*peit)->bpBegin(); partid_it != (*peit)->bpEnd(); ++partid_it)
    {
      nbor_rank = (*partid_it)/ParUtil::Instance()->getTgtNumParts();
      if (nbor_rank != SCUTIL_CommRank() && find(nbor_proc.begin(), nbor_proc.end(), nbor_rank)==nbor_proc.end())
        nbor_proc.push_back(nbor_rank);
    }
#endif
}

