/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "FMDB.h"
#include "mPart.h"
#include "mMesh.h"
#include "oldFMDB.h"
#include "FMDB_Iterator.h"
#include "iUtil.h"
#include "mMeshIO.h"
#include <assert.h>
#include <list>
#include <string.h>
#include "FMDB_cint.h"

//************************************
//  3.3 Part Information 
//************************************

int FMDB_Part_GlobID (pPart part)
{
  return part->getId();
}

int FMDB_Part_LocalID(pPart part)
{
#ifdef FMDB_PARALLEL
  return part->getId()%ParUtil::Instance()->getTgtNumParts();
#else
  return part->getId();
#endif
}

int FMDB_PartID_CommRank (int partID)
{
#ifdef FMDB_PARALLEL
  return partID/ParUtil::Instance()->getTgtNumParts();
#else
  return 0;
#endif
}

int FMDB_PartID_LocalID (int partID)
{
#ifdef FMDB_PARALLEL
  return partID%ParUtil::Instance()->getTgtNumParts();
#else
  return partID;
#endif
}


/* Given a part handle, get the dimension of the part. */
int FMDB_Part_GetDim (pPart part, int* dim)
{
  *dim = ((part->size(3))? 3 : ((part->size(2)) ? 2 : 0 ));
  return SCUtil_SUCCESS;
}

/* Part_Set deals with PO entity set */
int FMDB_Part_GetNumSet (pPart part, int* numEntSet)
{
  *numEntSet = part->numPSet();
  return SCUtil_SUCCESS;
}

int FMDB_Part_GetSet (pPart part, std::vector<pEntSet> &entsets)
{
  std::copy(part->beginPSet(), part->endPSet(), std::back_inserter(entsets));
  return SCUtil_FAILURE;
}

int FMDB_Part_FindEnt (pPart part, pMeshEnt ent, int* found)
{
  pMeshEnt tmpEnt=part->find(ent);
  if (tmpEnt) *found=1;
  else *found=0;
  
  return SCUtil_SUCCESS;
}
/* Given a part handle and an integer specifying the entity type (0 to 3), get the total number
of entities of type in the part. If includeGhost is 0, ghost copies are excluded, otherwise ghost
copies are included. If type is 0, get the number of vertices in the part, if type is 1, get the number
of edges in the part, if type is 2, get the number of faces in the part, if type is 3, get the number of
regions in the part. */
int FMDB_Part_GetNumEnt(pPart part, int type, int topo, int* /* out */ numEnt)
{
  pPartEntIter iter;
  pMeshEnt ent;
  int num=0, iterEnd=FMDB_PartEntIter_Init (part, type, topo, iter);
  if (!iterEnd)
  {
    while (!iterEnd)
    {
      iterEnd = FMDB_PartEntIter_GetNext(iter, ent);
      if (iterEnd) break;
      ++num;
    }
    FMDB_PartEntIter_Del (iter);
  }
  *numEnt = num;
  return SCUtil_SUCCESS;
}

/*  Get part boundary entities of a part with a specified type */
int FMDB_Part_GetNumPartBdry (pPart part, int target_part, int type, int topo, int* numEnt)
{
  pPartEntIter pent_it;
  pMeshEnt ent;
  int num=0, iterEnd =FMDB_PartEntIter_InitPartBdry (part, target_part, type, topo, pent_it);
  if (!iterEnd)
  {
    while (!iterEnd)
    {
      iterEnd = FMDB_PartEntIter_GetNext(pent_it, ent);
      if(iterEnd) break;
      ++num;
    }
    FMDB_PartEntIter_Del (pent_it); 
    *numEnt = num;
  }
  else 
    *numEnt=0;
  return SCUtil_SUCCESS;
}

int FMDB_Part_GetNumGhost (
        pPart /* in */ part, 
        int /* in */ type,
        int* /* out */ numEnt)
{
  cout<<__func__<<" not implemented yet\n";
  return SCUtil_NOT_SUPPORTED; 
}

int FMDB_Part_GetNumNborPart (pPart part, int entity_type, int* numNbor)
{
  *numNbor=0;
#ifdef FMDB_PARALLEL
  vector<int> vecPartID;
  int err = FMDB_Part_GetNborPartID (part, entity_type, vecPartID);
  if (err) return err;
  *numNbor = vecPartID.size();
#endif
  return SCUtil_SUCCESS;
}

int FMDB_Part_GetNborPartID (pPart part, int entity_type, vector<int>& nbor_part)
{
#ifdef FMDB_PARALLEL
  for (pmModel::PEIter peit=pmModel::Instance()->peBegin(); peit!=pmModel::Instance()->peEnd(); ++peit)
    for (set<int>::iterator partid_it= (*peit)->bpBegin(); partid_it != (*peit)->bpEnd(); ++partid_it)
    {
      if (*partid_it!= part->getId() && find(nbor_part.begin(), nbor_part.end(), *partid_it)==nbor_part.end())
        nbor_part.push_back(*partid_it);
    }
#endif
  return SCUtil_SUCCESS;
}

//************************************
//  3.5 ID Management
//************************************
// unique ent set ID gets invalid once new entity set is added (e.g. migration, modification)
int FMDB_Part_SetSetID (pPart part)
{
  int id=0;
  for (mPart::psetIter pset_it=part->beginPSet(); pset_it!=part->endPSet(); ++pset_it)
    ((mEntityGroup*)(*pset_it))->attachInt(FMDB_Util::Instance()->getId(),id++);
  return SCUtil_SUCCESS;
}

int FMDB_Part_DelSetID (pPart part)
{
  for (mPart::psetIter pset_it=part->beginPSet(); pset_it!=part->endPSet(); ++pset_it)
    ((mEntityGroup*)(*pset_it))->deleteData(FMDB_Util::Instance()->getId());
  return SCUtil_SUCCESS;
}

// given entity type, set unique ent id
// unique ent ID gets invalid once the mesh is modified (e.g. migration, modification)
int FMDB_Part_SetEntID (pPart part, int type)
{
  int i=1;
  switch (type)
  {
    case 0: for (mPart::iterall it=part->beginall(type); it!=part->endall(type);++it)
              (*it)->setId(i++);
            break;
    case 1:
    case 2:
    case 3: for (mPart::iterall it=part->beginall(type); it!=part->endall(type);++it)
              (*it)->attachInt(FMDB_Util::Instance()->getId(),i++);
            break;
    case 4: for (int type=0; type<4; ++type)
              FMDB_Part_SetEntID (part, type);
            break;
    default: return SCUtil_FAILURE;  // invalid type
  }
  return SCUtil_SUCCESS;
}

int FMDB_Part_DelEntID (pPart part, int type)
{
  switch (type)
  {
    case 0: break;
    case 1:
    case 2:
    case 3: for (mPart::iterall it=part->beginall(type); it!=part->endall(type);++it)
              (*it)->deleteData(FMDB_Util::Instance()->getId());
            break;
    case 4: for (int type=1; type<4; ++type)
              FMDB_Part_DelEntID (part, type);
            break;
    default: return SCUtil_FAILURE;  // invalid type
  }
  return SCUtil_SUCCESS;
}

