/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "FMDBfwd.h"
#include "ParUtil.h"
#include "mParallelUtility.h"

#include "FMDB_Internals.h"
#include "mVector.h"
#include "mPoint.h"
#include "mVertex.h"

#include "FMDBfwd.h"
#include "mEntity.h"
#include "mPart.h"
#include "mEdge.h"
#include "modeler.h"

#ifdef FMDB_PARALLEL
#include "FMDB_cint.h"
#endif

#include <iostream>
#include <list>
#include <vector>
#include <cstdio>

using std::cout;
using std::endl;
using std::vector;

#ifndef SIM
#ifdef FMDB_PARALLEL
#include "FMDB_ZoltanLoadBalancer.h"

class myZoltan_LoadBalancer2 : public FMDB_ZoltanLoadBalancer
{
public :
  myZoltan_LoadBalancer2 () 
    : FMDB_ZoltanLoadBalancer (FMDB_ZoltanLoadBalancer::Remap){}
  virtual void * getUserData (mEntity *e, int dest_proc, int &size)
  {
    int *id = new int[4];
    *id = ParUtil::Instance()->rank();
    size = 4*sizeof(int);
    return id;
  } 
  virtual void receiveUserData (mEntity *e, int pid, int tag, void *buf) 
  {
  }     
  virtual void deleteUserData (void *buf) 
  {
    int *id = (int*)buf;
    delete [] id;
  }  
  bool useWeights() const {return false;} 
  float getWeight (mEntity *) const {return 1.0;}  
};
#endif

//*******************************************
int M_dimension(pMesh m)
//*******************************************
{
  return((m->size(3))? 3 : ((m->size(2)) ? 2 : 1 ));
}
#endif

//*******************************************
mType M_GetElementType     (pEntity pE)
//*******************************************
{
#ifndef SIM  	// FMDB related
  return (mType)pE->getType();
#else
  switch(EN_type(pE))
    {
    case 0: 
      return VERTEX;
      break;
    case 1: 
      return EDGE;
      break;
    case 2:
      {
	int numEdges = F_numEdges((pFace)pE);
	if(numEdges == 3)return TRI;
	else if(numEdges == 4)return QUAD;
	else throw 1;	
      }
      break;
    case 3:
      {
	int numFaces = R_numFaces((pRegion)pE);
	if(numFaces == 4)return TET;
	else if(numFaces == 6)return HEX;
	else if(numFaces == 5)return PRISM;
	else throw 1;	
      }
      break;
    default:
      throw 1;
    }
#endif
}

//*******************************************
void  M_GetVertices (pEntity pE, std::vector<pVertex> &v)
//*******************************************
{
#ifndef SIM	// FMDB
  if(pE->getLevel() == 0)v.push_back(pE);
  else
    {
      int n = pE->size(0);
      for(int i=0;i<n;i++)v.push_back(pE->get(0,i));
    }
#else
  switch(EN_type(pE))
    {
    case 1:
      v.push_back(E_vertex((pEdge)pE,0));
      v.push_back(E_vertex((pEdge)pE,1));
      break;
    case 2:
      { 
	void *iter = 0;
	pEntity iadj; 
	pPList vtxs = F_vertices((pFace)pE,1); 
	while (iadj = (pEntity)PList_next(vtxs,&iter))v.push_back((pVertex)iadj);
	PList_delete(vtxs);
      }
      break;
    case 3:      
      { 
	void *iter = 0;
	pEntity iadj; 
	pPList vtxs = R_vertices((pRegion)pE,1);
	while (iadj = (pEntity)PList_next(vtxs,&iter))v.push_back((pVertex)iadj);
	PList_delete(vtxs);
      }
      break;
    default:
      throw 1;
    }
#endif
}


#ifdef SIM
static int TEdVt[6][2] = {{0,1},{1,2},{0,2},{0,3},
                         {1,3},{2,3}};
#endif

//*******************************************
int R_edgeDir (pRegion pf, int n)
//*******************************************
{
  int x;
#ifdef SIM
  pPList edgelist = R_edges(pf, 1);
  pEdge edge = (pEdge)PList_item(edgelist, n);
  pVertex v1 = E_vertex(edge, 0);
  pPList vertexlist = R_vertices(pf, 1);
  int indx = TEdVt[n][0];
  pVertex v1_bar = (pVertex)PList_item(vertexlist, indx);
  if(v1 == v1_bar) x = 1;
  else x = -1;
  PList_delete(edgelist);
  PList_delete(vertexlist);
#else 		// FMDB
  mEdge *ed =  (mEdge*)pf->get(1,n);
  mEdge *ed2 = (mEdge*)pf->getTemplate(n,1,0);
  mVertex *v1 = ed->vertex(0);
  mVertex *v1x = ed2->vertex(0);
  if(v1 == v1x)x = 1;
  else x = -1;
  if(ed != ed2) delete ed2; 
#endif
  return x;
}

//*******************************************
double M_SizeMin (pEntity pE)
//*******************************************
{
  SCOREC::Util::mPoint p[8];
  std::vector<pVertex> v;
  M_GetVertices (pE, v);

  for(size_t i=0;i<v.size();i++)
    {
      pPoint pp =  V_point(v[i]); 
      p[i](0) = P_x(pp);
      p[i](1) = P_y(pp);
      p[i](2) = P_z(pp);
    }

  switch (M_GetElementType(pE))
    {
    case VERTEX : return 1.0;    
    case EDGE   : 
      {
	SCOREC::Util::mVector v1(p[1]-p[0]);
	return sqrt (v1*v1);
      }
    case TRI :
      {
	SCOREC::Util::mVector v1(p[1]-p[0]);
	SCOREC::Util::mVector v2(p[2]-p[1]);
	SCOREC::Util::mVector v3(p[0]-p[2]);
	SCOREC::Util::mVector v1v2 = v1 % v2;
	double V = sqrt(v1v2 * v1v2);
	double A1 = sqrt (v1 * v1);
	double A2 = sqrt (v2 * v2);
	double A3 = sqrt (v3 * v3);
	return fabs(V) / (A1+A2+A3);	
      }
    case QUAD :
      {
	SCOREC::Util::mVector v1(p[1]-p[0]);
	SCOREC::Util::mVector v2(p[2]-p[1]);
	SCOREC::Util::mVector v3(p[0]-p[2]);
	SCOREC::Util::mVector v1v2 = v1 % v2;
	double V = sqrt(v1v2 * v1v2);
	double A1 = sqrt (v1 * v1);
	double A2 = sqrt (v2 * v2);
	double A3 = sqrt (v3 * v3);
	return fabs(V) / (A1+A2+A3);	
      }
    case TET :
      {
	SCOREC::Util::mVector v1(p[0]-p[1]);
	SCOREC::Util::mVector v2(p[0]-p[2]);
	SCOREC::Util::mVector v3(p[0]-p[3]);
	SCOREC::Util::mVector v4(p[2]-p[1]);
	SCOREC::Util::mVector v5(p[3]-p[1]);
	SCOREC::Util::mVector v1v2 = v1 % v2;
	SCOREC::Util::mVector v1v3 = v1 % v3;
	SCOREC::Util::mVector v2v3 = v2 % v3;
	SCOREC::Util::mVector v4v5 = v4 % v5;
	double V = v1v2 * v3;
	double A1 = sqrt (v1v2 * v1v2);
	double A2 = sqrt (v1v3 * v1v3);
	double A3 = sqrt (v2v3 * v2v3);
	double A4 = sqrt (v4v5 * v4v5);
	return fabs(V) / (A1+A2+A3+A4);
      }
    default : 
      break;
    }
  return 0.0;
}

#ifndef SIM
int M_LoadBalance3(pMesh pm)
{
#ifdef FMDB_PARALLEL
  myZoltan_LoadBalancer2 lb;
#endif
  return 0;
}
#endif

// it returns minimum pid
int M_OwnerProc (pMesh pm, pEntity pe)
{
#ifndef SIM
  return pe->getOwner();
#else
  return 0;
#endif  
}

int M_numRemoteCopies (pMesh pm, pEntity pe)
{
#ifdef FMDB_PARALLEL
  size_t ncopies = pe->getNumRemoteCopies();
#ifdef DEBUG
  //this is the wrong way to do this function!
  vector<std::pair<pEntity,int> > vec;
  EN_getCopies(pe, vec);
  if(ncopies != vec.size())
     throw 1;
#endif
  return ncopies;
#else
  return 0;
#endif
}

pPList  M_getRemoteCopies (pMesh, pEntity)
{  
  return 0;
}

int M_getRemotePID (void *)
{
  return 0;
}

int M_MaximumVertexId (pMesh pm)
{
#ifndef SIM
  int localMaxId = pm->getIdGenerator().getMaxValue();
  int globalMaxId = localMaxId;
#ifdef FMDB_PARALLEL
  MPI_Allreduce(&localMaxId,&globalMaxId,1,MPI_INT,MPI_MAX,
		ParUtil::Instance()->getComm());
#endif
  return globalMaxId;
#else
  return 0;
#endif
}

int E_getWhatInTag (pEntity e)
{
#ifndef SIM	
  return GEN_tag(e->getClassification());
#else
  return 0;
#endif
}

int E_getWhatInType (pEntity e)
{
#ifndef SIM	
  return GEN_type(e->getClassification());
#else
  return 0;
#endif
}

void V_setWhatInType (pMesh pm, pVertex e, int x)//  categorize a mesh vertex to a model entity type
{
  throw;
#ifndef SIM	
  e->classify (pm->getGEntity (x, 1));	
#else
  return;
#endif
}

void E_setWhatInType (pMesh pm, pEdge e, int x) //  categorize a mesh edge to a model entity type
{
  throw;
#ifndef SIM
  e->classify(pm->getGEntity (x,1));
#else
  return;
#endif
}

void F_setWhatInType (pMesh pm, pFace e, int x)  //  categorize a mesh face to a model entity type
{
  throw;
#ifndef SIM
  e->classify(pm->getGEntity (x,1));	
  printf ("a face is classified on %d %d %p\n",x,1,e->getClassification());
#else
  return;
#endif
}
