/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/ 
#ifndef SIM
#include "FMDB_cint.h"
#include "FMDB_Internals.h"
#include "FMDBfwd.h"
#include "ParUtil.h"
#include "pmMigration.h"
#include "mParallelUtility.h"
#include "mEntity.h"
#include "mPart.h"
#include "mEdge.h"
#include "mVector.h"
#include "mPoint.h"
#include "mVertex.h"
#include "mException.h"
#include "mFMDB.h"
#include "SCUtil.h"
#ifdef FMDB_PARALLEL
#include "pmModel.h"
#include "pmMigrationCallbacks.h"
#include "pmZoltanCallbacks.h"
#endif
#ifdef DMUM
#include "mDmum.h"
#endif
#include "FMDB_EntGrp.h"
#include "FMDB.h"
#include "mMeshIO.h"
#include <iostream>
#include <list>
#include <vector>
#include <algorithm>
#include <iterator>
#include <string>

using std::cout;
using std::endl;
using std::vector;
using std::copy;
using std::string;
using std::list; 

#define MAX_NUM_PARTS_PER_PROCESS 1000

// **********************************
// General mesh Info
// **********************************
int M_getMaxVId(pMesh mesh)
{
#ifndef FMDB_PARALLEL
  return mesh->getIdGenerator().getMaxValue(); 
#else
  return P_getMaxInt(mesh->getIdGenerator().getMaxValue());  
#endif   
}

int M_pid()
{
  return P_pid();
}

int M_size()
{
  return P_size();
}

void M_getPOEntities(pMesh mesh, std::vector<pEntity>* poEntities)
{
  for (int i=0; i<3; ++i)
    for (mPart::iterall it=mesh->beginall(i); it!=mesh->endall(i); ++it)
    {
      if ((*it)->size(i+1)==0)
        poEntities[i].push_back(*it);
    }
  for (mPart::iterall it=mesh->beginall(3); it!=mesh->endall(3); ++it)
    poEntities[3].push_back(*it);
}

void M_print(pMesh pm)
{
  ParUtil::Instance()->Msg(ParUtil::INFO,"\n***  M_print(mesh)  ***\n");	  
  pm->printAll();
  ParUtil::Instance()->Barrier(__LINE__,__FILE__);
  ParUtil::Instance()->Msg(ParUtil::INFO,"\n***  M_print(mesh)  ***\n");	  
}

void M_printNumEntities(pMesh mesh)
{
  ParUtil::Instance()->Msg(ParUtil::INFO,"\n****************************\n");
  ParUtil::Instance()->Msg(ParUtil::INFO,"    M_printNumEntities\n\n");
  mesh->printNumEntities();  
  ParUtil::Instance()->Msg(ParUtil::INFO,"\n****************************\n\n");	  
}

int PM_merge(pMesh mesh)
{
#ifdef FMDB_PARALLEL
  if (P_size()==1) return 0;
#ifdef DEBUG
  ParUtil::Instance()->Msg(ParUtil::INFO,"\n****************************\n");
  ParUtil::Instance()->Msg(ParUtil::INFO,"    PM_merge - ");
  double t1 = ParUtil::Instance()->wTime();
#endif
  zoltanCB zlb;
  int ret = pmMerge(mesh,zlb);
#ifdef DEBUG
  double t2 = ParUtil::Instance()->wTime(); 
  ParUtil::Instance()->Msg(ParUtil::INFO,
          "      (t = %f sec)\n",t2-t1);
  ParUtil::Instance()->Msg(ParUtil::INFO,"****************************\n\n");	  
#endif
  return ret;
#else
  return 0;
#endif
}


#ifdef FMDB_PARALLEL
void PM_setMaxNumParts(int num)                 // for multiple parts per process
{
  if(num>=MAX_NUM_PARTS_PER_PROCESS)  
     num=MAX_NUM_PARTS_PER_PROCESS;
  ParUtil::Instance()->setTgtNumParts(num); 
#ifdef DEBUG
  if (ParUtil::Instance()->master()) 
    cout<<__func__<<": maxNumParts="<<num<<endl;
#endif
}

void M_loadbalance(pPart part,pmMigrationCallbacks &cb)
{
  if (P_size()==1) return;
  ParUtil::Instance()->Barrier(__LINE__,__FILE__);
  ParUtil::Instance()->Msg(ParUtil::INFO,"\n****************************\n");
  ParUtil::Instance()->Msg(ParUtil::INFO,"    M_loadbalance\n");
//  FMDB_Mesh_GlobPtn(part->getMesh(), cb);
  pmLoadBalance(part, cb);
  ParUtil::Instance()->Barrier(__LINE__,__FILE__);
  ParUtil::Instance()->Msg(ParUtil::INFO,"****************************\n\n");	  
  return;
}


void M_loadbalance2(vector<pPart>& parts,pmMigrationCallbacks &cb)   
{
  ParUtil::Instance()->Barrier(__LINE__,__FILE__);
  ParUtil::Instance()->Msg(ParUtil::INFO,"\n****************************\n");
  ParUtil::Instance()->Msg(ParUtil::INFO,"    M_loadbalance\n");
  pmLoadBalance(parts, cb);
  ParUtil::Instance()->Barrier(__LINE__,__FILE__);
  ParUtil::Instance()->Msg(ParUtil::INFO,"****************************\n\n");
  return;
}

int M_partMigration(std::vector<pMesh>&meshes,                             // migrate a whole part
		     int source_pid,
		     int target_rank,
		     pmMigrationCallbacks &cb)
{
   return pmPartMigration(meshes, source_pid, target_rank, cb);
}

int M_migration(pMesh part, int dimToMove, std::list<pEntity>& part_bdry_ents,
                pmMigrationCallbacks &cb,
                int dimToKnow1, std::vector<pEntity>& rmE1, std::vector<pEntity>& newE1,
		int dimToKnow2, std::vector<pEntity>& rmE2, std::vector<pEntity>& newE2,
                std::vector<pEntityGroup>* rmEG, std::vector<pEntityGroup>* newEG)
{
  if (P_size()==1) return 0;
#ifdef DEBUG
  ParUtil::Instance()->Msg(ParUtil::INFO,"\n****************************\n");
  if (dimToMove==1)
  {
    if (dimToKnow1 == 3)
      ParUtil::Instance()->Msg(ParUtil::INFO,"    M_migration_Realign\n");
    else if (dimToKnow1==1)      
      ParUtil::Instance()->Msg(ParUtil::INFO,"    M_migration_Swap\n");
  }
  else
  {
    if (dimToKnow2==-1)
      ParUtil::Instance()->Msg(ParUtil::INFO,"    M_migration_Coarsen\n");
    else 
      ParUtil::Instance()->Msg(ParUtil::INFO,"    M_migration_Snap\n");
  }
#endif
  
  std::vector<pEntity> rmE[4];
  std::vector<pEntity> newE[4];
  int ret = FMDB_Mesh_PullPartBdryToOwner(part->getMesh(), cb, part_bdry_ents, newE, rmE, newEG,rmEG);

  if (dimToKnow1>=0)
  {  
    std::copy(rmE[dimToKnow1].begin(), rmE[dimToKnow1].end(), 
              std::back_inserter(rmE1));
    std::copy(newE[dimToKnow1].begin(), newE[dimToKnow1].end(), 
              std::back_inserter(newE1));
  }

  if (dimToKnow2>=0)
  {  
    std::copy(rmE[dimToKnow2].begin(), rmE[dimToKnow2].end(), 
              std::back_inserter(rmE2));
    std::copy(newE[dimToKnow2].begin(), newE[dimToKnow2].end(), 
              std::back_inserter(newE2));
  }
  #ifdef DEBUG
  ParUtil::Instance()->Msg(ParUtil::INFO,"****************************\n\n");	  
#endif
  return ret;
}


int M_migrationForSnap(pMesh part, pmMigrationCallbacks &cb,
		std::list<pEntity>& part_bdry_vtx, 
  	        std::list<std::pair<pEntity,pEntity> > &vtx_ent_pair,
                int dimToKnow1, std::vector<pEntity>& rmE1, std::vector<pEntity>& newE1,
                std::vector<pEntityGroup>* rmEG, std::vector<pEntityGroup>* newEG)
{
  if (P_size()==1) return 0;
  ParUtil::Instance()->Barrier(__LINE__,__FILE__);
#ifdef DEBUG
  ParUtil::Instance()->Msg(ParUtil::INFO,"\n****************************\n");
  ParUtil::Instance()->Msg(ParUtil::INFO,"    M_migration_Snap\n");
#endif
  
  std::vector<pEntity> rmE[4];
  std::vector<pEntity> newE[4];
  int ret = FMDB_Mesh_PullPartBdryForSnapping(part->getMesh(), cb, part_bdry_vtx, vtx_ent_pair, newE, rmE, newEG,rmEG);

//  int ret = pullPartBdryToOwner(part->getMesh(), cb, vtsOnCB, vtfcPairs,newE,rmE, newEG,rmEG);

  if (dimToKnow1>=0)
  {  
    std::copy(rmE[dimToKnow1].begin(), rmE[dimToKnow1].end(), 
              std::back_inserter(rmE1));
    std::copy(newE[dimToKnow1].begin(), newE[dimToKnow1].end(), 
              std::back_inserter(newE1));
  }
  ParUtil::Instance()->Barrier(__LINE__,__FILE__);
#ifdef DEBUG
  ParUtil::Instance()->Msg(ParUtil::INFO,"****************************\n\n");	  
#endif
  return ret;
}
			   
#endif

void M_set_maxMeshDim(int dim)
{
#ifdef FMDB_PARALLEL
  ParUtil::Instance()->set_maxMeshDim(dim);
#endif
}

int M_maxMeshDim()
{
#ifdef FMDB_PARALLEL
  return ParUtil::Instance()->get_maxMeshDim();
#endif
}

int M_globalMaxDim(pMesh mesh)
{
  int dim = mesh->getDimension();
  int maxDim = P_getMaxInt(dim);
  return maxDim;
}

int M_globalMinDim(pMesh mesh)
{
  int dim = mesh->getDimension();
  int maxDim = P_getMinInt(dim);
  return maxDim;
}

void M_numEntitiesOwned(pMesh mesh, int dim, vector<int>& output)
{
#ifdef FMDB_PARALLEL
  unsigned int numPE = ParUtil::Instance()->size();
  output.resize(numPE);
  int *in = new int;
  int *out = new int[numPE];
 
  mEntity* ent;
  int numEnt=0;
  for (mPart::iterall it=mesh->beginall(dim);it!=mesh->endall(dim);++it)
  {
    ent = *it;
    if (ent->getOwner()==ParUtil::Instance()->rank())
      numEnt++;
  }

  *in = numEnt;
  MPI_Allgather(in, 1, MPI_INT, out, 1, MPI_INT, ParUtil::Instance()->getComm());
  for (unsigned i = 0; i < numPE; i++)
    output[i] = out[i];

  delete in;
  delete [] out;
#else
  output.push_back(mesh->size(dim));
#endif
}

void M_updateOwnership(pMesh mesh)
{ 
#ifdef FMDB_PARALLEL
  if (M_NumPE()==1)
    return;
#ifdef _DEBUG
  double t1 = ParUtil::Instance()->wTime();
  ParUtil::Instance()->Msg(ParUtil::INFO,"\n****************************\n");
  ParUtil::Instance()->Msg(ParUtil::INFO,"    M_updateOwnership...\n");
#endif

  pmModel::Instance()->updateOwnership(mesh->getMesh());

#ifdef _DEBUG
  double t2 = ParUtil::Instance()->wTime();      
  ParUtil::Instance()->Msg(ParUtil::INFO,
          "      (t = %f sec)\n",t2-t1);
  ParUtil::Instance()->Msg(ParUtil::INFO,"****************************\n\n");	  
#endif
#endif
  return;
}

void DMUM_startMonitoring(pMesh mesh)
{ 
#ifdef DMUM
  mDMUM::Instance()->startMonitor((mPart*)mesh); 
  if (!P_pid()) system("rm mRep.dat");
#else 
  if (!P_pid())   cout<<"FMDB ERROR: Compile FMDB with DMUM=1\n";
#endif  
}

void DMUM_stopMonitoring(pMesh mesh)
{
#ifdef DMUM
  mDMUM::Instance()->stopMonitor((mPart*)mesh); 
  mDMUM::Instance()->write(mesh, "mRep.dat");
#else 
  if (!P_pid())   cout<<"FMDB ERROR: Compile FMDB with DMUM=1\n";
#endif
}

void DMUM_resumeMonitoring(pMesh mesh)
{  
#ifdef DMUM
  mDMUM::Instance()->resumeMonitor((mPart*)mesh); 
#else 
  if (!P_pid())   cout<<"FMDB ERROR: Compile FMDB with DMUM=1\n";
#endif
}

void DMUM_pauseMonitoring(pMesh mesh)
{  
#ifdef DMUM
  mDMUM::Instance()->pauseMonitor((mPart*)mesh); 
#else 
  if (!P_pid())   cout<<"FMDB ERROR: Compile FMDB with DMUM=1\n";
#endif
}

void DMUM_print(pMesh mesh)
{
#ifdef DMUM
  mDMUM::Instance()->print((mPart*)mesh); 
#else 
  if (!P_pid()) cout<<"FMDB ERROR: Compile FMDB with DMUM=1\n";
#endif
}

bool DMUM_isOn(pMesh mesh)
{
#ifdef DMUM
  return (mPart*)mesh->DMUM_on; 
#else
  return false;
#endif
}

#endif   /* ifndef SIM */
