/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "iBase.h"
#include "iMesh.h"
#include "FMDB.h"
#include <stdint.h>

// Ensure that a string will be null terminated, even if it wasn't before.
#define NULL_TERMINATE(a_in_, len_, a_)         \
  do {                                          \
    snprintf(a_, 1023, "%s", a_in_);            \
    a_[len_] = 0;                               \
    int i__ = len_-1;                           \
    while (i__ >= 0 && a_[i__] == ' ') {        \
      a_[i__--] = 0;                            \
    };                                          \
  } while(false);


void iMesh_createTag(iMesh_Instance instance,
                     /*in*/ const char* tag_name_in,
                     /*in*/ const int tag_size,
                     /*in*/ const int tag_type,
                     /*out*/ iBase_TagHandle* tag_handle, int *err,
                     /*in*/ const int tag_name_len )
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  if(tag_size<0)
  {
   *err = iBase_INVALID_ARGUMENT;
   FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
   *tag_handle = 0;
   return;
  }
  char tag_name[1024];
  NULL_TERMINATE(tag_name_in, tag_name_len, tag_name);

  if(tag_type<iBase_TagValueType_MIN || tag_type>iBase_TagValueType_MAX)
  {
    *err = iBase_INVALID_ARGUMENT;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    *tag_handle = 0;
    return;
  }

  pTag tag;
  if (FMDB_Mesh_CreateTag((pMeshMdl)instance, tag_name, tag_type, tag_size, tag))
  {
    *err = iBase_TAG_ALREADY_EXISTS;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    *tag_handle = (iBase_TagHandle)tag;
    return;
  }

  *tag_handle = (iBase_TagHandle)tag;
}

void iMesh_destroyTag(iMesh_Instance instance,
                             /*in*/ iBase_TagHandle tag_handle,
                             /*in*/ const int forced, int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  int exist;
  FMDB_Mesh_HasTag ((pMeshMdl)instance, (pTag)tag_handle, &exist);

  // tag validity check
  if (!exist)  {
    *err = iBase_INVALID_ARGUMENT;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  int in_use=0;
  FMDB_Tag_InUse ((pMeshMdl)instance, (pTag)tag_handle, &in_use);
  if (in_use && !forced)
  {
    *err = iBase_TAG_IN_USE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  *err = FMDB_Mesh_DelTag ((pMeshMdl)instance, (pTag)tag_handle, forced);
  return;
}

void iMesh_getTagHandle(iMesh_Instance instance,
                        /*in*/ const char* tag_name_in,
                        iBase_TagHandle *tag_handle, int *err,
                        int tag_name_len)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  char tag_name[1024];
  NULL_TERMINATE(tag_name_in, tag_name_len, tag_name);

  pTag tag;
  if (FMDB_Mesh_FindTag((pMeshMdl)instance, tag_name, tag)) { // no tag found
    *err = iBase_INVALID_ARGUMENT;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  *tag_handle = (iBase_TagHandle)tag;
}

void iMesh_getTagType(iMesh_Instance instance,
                             /*in*/ const iBase_TagHandle tag_handle,
                             int *tag_type, int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  int exist;
  FMDB_Mesh_HasTag((pMeshMdl)instance, (pTag)tag_handle, &exist);
  // (pTag)tag_handle validity check
  if (!exist)
  {
    *err = iBase_INVALID_ARGUMENT;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  FMDB_Tag_GetType((pTag)tag_handle, tag_type);
}

void iMesh_getTagSizeBytes(iMesh_Instance instance,
                                  /*in*/ const iBase_TagHandle tag_handle,
                                  int *byte_tag_size, int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  int exist;
  FMDB_Mesh_HasTag((pMeshMdl)instance, (pTag)tag_handle, &exist);
  // (pTag)tag_handle validity check
  if (!exist)
  {
    *err = iBase_INVALID_ARGUMENT;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  int tag_type, tag_size;
  FMDB_Tag_GetType ((pTag)tag_handle, &tag_type);
  FMDB_Tag_GetSize ((pTag)tag_handle, &tag_size);
    
  switch(tag_type)
  {
  case iBase_INTEGER:
    *byte_tag_size = (int)(sizeof(int)*tag_size);
    break; 
  case iBase_DOUBLE:           
    *byte_tag_size = (int)(sizeof(double)*tag_size);
    break;                     
  case iBase_ENTITY_HANDLE:    
    *byte_tag_size = (int)(sizeof(void*)*tag_size);
    break; 
  default:
    *byte_tag_size = (int)(sizeof(char)*tag_size);
    break;
  }     
}

void iMesh_getTagSizeValues(iMesh_Instance instance,
                                   /*in*/ const iBase_TagHandle tag_handle,
                                   int *tag_size, int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  int exist;
  FMDB_Mesh_HasTag((pMeshMdl) instance, (pTag)tag_handle, &exist);
  // (pTag)tag_handle validity check
  if (!exist)
  {
    *err = iBase_INVALID_ARGUMENT;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  FMDB_Tag_GetSize((pTag)tag_handle, tag_size);
}

void iMesh_getTagName(iMesh_Instance instance,
                             /*in*/ const iBase_TagHandle tag_handle,
                             char *name, int *err,
                             int name_len)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  int exist;
  FMDB_Mesh_HasTag((pMeshMdl) instance, (pTag)tag_handle, &exist);
  // (pTag)tag_handle validity check
  if (!exist)
  {
    *err = iBase_INVALID_ARGUMENT;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  FMDB_Tag_GetName ((pTag)tag_handle, name, name_len);
}

