/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "iMeshP.h"
#include "FMDB.h"

void iMeshP_getNumGlobalParts(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            int *num_global_part, 
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  int localNumPart=0;
  FMDB_Mesh_GetNumPart((pMeshMdl)instance, &localNumPart);
  MPI_Allreduce(&localNumPart,num_global_part,1,MPI_INT,MPI_SUM, ParUtil::Instance()->getComm()); 
  return;
}

void iMeshP_getNumLocalParts(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            int *num_local_part, 
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  FMDB_Mesh_GetNumPart((pMeshMdl)instance, num_local_part);
  return;
}


void iMeshP_getLocalParts(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            iMeshP_PartHandle **parts,
            int *parts_allocated,
            int *parts_size,
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  int procId=SCUTIL_CommRank();
  std::vector<pPart> part_vec;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, procId, part_vec);
  int numParts = part_vec.size();
  if( *parts ==0 || *parts_allocated==0 ) {
    *parts = (iMeshP_PartHandle*)(calloc(numParts, sizeof(iMeshP_PartHandle) )); 
    if (*parts==0 ) 
    { 
       *err = iBase_MEMORY_ALLOCATION_FAILED; 
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
       return;
    }
     
    *parts_allocated = numParts; 
  }
  else if ( *parts_allocated < numParts) 
  {
    *err = iBase_BAD_ARRAY_SIZE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }
   
  int count = 0; 
  for(vector<pPart>::iterator pit=part_vec.begin(); pit!=part_vec.end(); ++pit)
  { 
    (*parts)[count] = (iMeshP_PartHandle)(*pit); 
    count++; 
  }
   
  *parts_size = numParts; 
  return;
}

void iMeshP_getRankOfPart(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_Part part_id,
            int *rank,
            int *err)
{
  *err = iBase_SUCCESS; 
  *rank = FMDB_PartID_CommRank(part_id);
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  return;
}

void iMeshP_getRankOfPartArr(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_Part *part_ids,
            const int part_ids_size,
            int **ranks, 
            int *ranks_allocated, 
            int *ranks_size,
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  if( *ranks ==0 || *ranks_allocated==0 ) {
    *ranks = (int*)(calloc(part_ids_size, sizeof(int))); 
    if (*ranks==0 ) { 
      *err = iBase_MEMORY_ALLOCATION_FAILED; 
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
     
    *ranks_allocated = part_ids_size; 
  }
  else if ( *ranks_allocated < part_ids_size) 
  {
    *err = iBase_BAD_ARRAY_SIZE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }
   
  for(int i=0; i<part_ids_size; ++i) 
    *(*ranks + i*sizeof(int)) = FMDB_PartID_CommRank(part_ids[i]);
  *ranks_size = part_ids_size; 
  return;
}


void iMeshP_createPart(
            iMesh_Instance instance,
            iMeshP_PartitionHandle partition,
            iMeshP_PartHandle *part,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  int numPart;
  FMDB_Mesh_GetNumPart((pMeshMdl)instance, &numPart);
  FMDB_Mesh_SetNumPart((pMeshMdl)instance, numPart+1);

  pPart tempPart;
  FMDB_Mesh_GetPart ((pMeshMdl)instance, numPart, tempPart);
  *part = (iMeshP_PartHandle)tempPart;
  return;
}

void iMeshP_destroyPart(
            iMesh_Instance instance,
            iMeshP_PartitionHandle partition,
            iMeshP_PartHandle part,
            int *err)
{
  *err = FMDB_Mesh_DelPart ((pMeshMdl)instance, (pPart)part);
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  return;
}


void iMeshP_getPartIdFromPartHandle(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle part,
            iMeshP_Part *part_id,
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  *part_id = (iMeshP_Part)(FMDB_Part_GlobID ((pPart) part));
  return;
}

void iMeshP_getPartIdsFromPartHandlesArr(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle *parts,
            const int parts_size,
            iMeshP_Part **part_ids,
            int *part_ids_allocated,
            int *part_ids_size,
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  for(int i=0; i<parts_size; ++i)
    if (!((pPart)parts[i]))
    {
      *err = iBase_FAILURE;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
  int procId=SCUTIL_CommRank();
  std::vector<pPart> part_vec;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, procId, part_vec);
  int numParts = part_vec.size();
  if( *part_ids ==0 || *part_ids_allocated==0 ) {
    *part_ids = (iMeshP_Part*)(calloc(parts_size, sizeof(iMeshP_Part) )); 
    if (*part_ids==0 ) { 
       *err = iBase_FAILURE;
       FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
       return;
    }
     
    *part_ids_allocated = parts_size; 
  }
  else if ( *part_ids_allocated < parts_size) {
    *err = iBase_BAD_ARRAY_SIZE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }
   
  for(int i=0; i<parts_size; ++i)
    *(*part_ids + i*sizeof(iMeshP_Part)) = (iMeshP_Part)(FMDB_Part_GlobID((pPart)(parts[i]))); 
  *part_ids_size = parts_size; 
  return;
}


void iMeshP_getPartHandleFromPartId(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            iMeshP_Part part_id,
            iMeshP_PartHandle *part,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  pPart part_hdl;
  int maxNumPart=0;

  FMDB_Mesh_GetNumPart((pMeshMdl)instance, &maxNumPart);

  if (part_id < maxNumPart*SCUTIL_CommRank() || part_id >= maxNumPart*(SCUTIL_CommRank()+1)){
    *err = iBase_FAILURE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  *err = FMDB_Mesh_GetPart((pMeshMdl)instance, part_id%maxNumPart, part_hdl);
  
  if (*err != iBase_SUCCESS) 
  {
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  *part = (iMeshP_PartHandle)part_hdl;
  return;
}

void iMeshP_getPartHandlesFromPartsIdsArr(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_Part *part_ids,
            const int part_ids_size,
            iMeshP_PartHandle **parts,
            int *parts_allocated,
            int *parts_size,
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  pPart part_hdl;
  int maxNumPart=0;
  FMDB_Mesh_GetNumPart((pMeshMdl)instance, &maxNumPart);

  std::vector<pPart> part_vec;
  for (int i=0; i<part_ids_size; ++i)
  {
    if (part_ids[i] < maxNumPart*SCUTIL_CommRank() || part_ids[i] >= maxNumPart*(SCUTIL_CommRank()+1)) 
    {
      *err = iBase_FAILURE;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      break;
    }
    *err = FMDB_Mesh_GetPart((pMeshMdl)instance, part_ids[i]%maxNumPart, part_hdl);
    if (*err != iBase_SUCCESS) 
    {
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
    part_vec.push_back(part_hdl);
  }  


  if( *parts ==0 || *parts_allocated==0 ) 
  {
    *parts = (iMeshP_PartHandle*)(calloc(part_ids_size, sizeof(iMeshP_PartHandle) )); 
    if (*parts==0 ) 
    { 
       *err = iBase_MEMORY_ALLOCATION_FAILED; 
       FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
       return;
    }
    *parts_allocated = part_ids_size; 
  }
  else if ( *parts_allocated < part_ids_size) 
  {
    *err = iBase_BAD_ARRAY_SIZE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }
   
  int count = 0; 
  for(vector<pPart>::iterator pit=part_vec.begin(); pit!=part_vec.end(); ++pit)
  { 
    (*parts)[count]  = (iMeshP_PartHandle)(*pit); 
    count++; 
  }
   
  *parts_size = part_ids_size; 
  
  return;
}

/*
 * NOTE:
 * This assumes the count should consist ONLY of entities that are OWNED by the given
 * part.  It does NOT include remote copies in the count.
 */
void iMeshP_getNumOfType(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle part,
            const iBase_EntitySetHandle entity_set,
            int entity_type, 
            int *num_type, 
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  // check validity of topology
  if(entity_type < iBase_VERTEX || entity_type > iBase_ALL_TYPES)
  {
    *err = iBase_INVALID_ENTITY_TYPE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }         
            
  int counter=0, entEnd;
  pMeshEnt ent;
  
  if ((iBase_EntitySetHandle)instance!=entity_set) // entity set
  {
    pSetEntIter iter;
    entEnd = FMDB_SetEntIter_Init((pEntSet)entity_set, entity_type, iMesh_ALL_TOPOLOGIES, iter);
    while (!entEnd)
    {
      entEnd = FMDB_SetEntIter_GetNext(iter, ent);
      if (entEnd) break;
      ++counter;
    }
    FMDB_SetEntIter_Del(iter);
  }
  else // part
  {
    pPartEntIter iter;
    entEnd = FMDB_PartEntIter_Init((pPart)part, entity_type, iMesh_ALL_TOPOLOGIES, iter);
    *num_type=0;
    if (entEnd) return;
    while (!entEnd)
    {
      entEnd = FMDB_PartEntIter_GetNext(iter, ent);
      if (entEnd) break;
        ++counter;
    }
    FMDB_PartEntIter_Del(iter);
    *num_type=counter;
  } // part
}

/*
 * NOTE:
 * This assumes the count should consist ONLY of entities that are OWNED by the given
 * part.  It does NOT include remote copies in the count.
 */
void iMeshP_getNumOfTopo(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle part,
            const iBase_EntitySetHandle entity_set,
            int entity_topology, 
            int *num_topo, 
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  // check validity of topology
  if(entity_topology < iMesh_POINT || entity_topology > iMesh_ALL_TOPOLOGIES)
  {
    *err = iBase_INVALID_ENTITY_TOPOLOGY;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }

  int counter=0, entEnd;
  pMeshEnt ent;

  if ((iBase_EntitySetHandle)instance!=entity_set) // entity set
  {
    pSetEntIter iter;
    entEnd = FMDB_SetEntIter_Init((pEntSet)entity_set, FMDB_ALLTYPE, entity_topology, iter);
    while (!entEnd)
    {
      entEnd = FMDB_SetEntIter_GetNext(iter, ent);
      if (entEnd) break;
      ++counter;
    }
    FMDB_SetEntIter_Del(iter);
  }
  else // part
    counter=((pPart)part)->getNumTopo(entity_topology); 
  *num_topo = counter;
  return;
}

void iMeshP_getAdjEntIndices(
            iMesh_Instance instance,
            iMeshP_PartitionHandle partition,
            iMeshP_PartHandle part,
            iBase_EntitySetHandle entity_set_handle,
            int entity_type_requester,
            int entity_topology_requester,
            int entity_type_requested,
            iBase_EntityHandle** entity_handles,
            int* entity_handles_allocated,
            int* entity_handles_size,
            iBase_EntityHandle** adj_entity_handles,
            int* adj_entity_handles_allocated,
            int* adj_entity_handles_size,
            int** adj_entity_indices,
            int* adj_entity_indices_allocated,
            int* adj_entity_indices_size,
            int** offset,
            int* offset_allocated,
            int* offset_size,
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  // check validity of topology
  if(entity_topology_requester < iMesh_POINT || entity_topology_requester > iMesh_ALL_TOPOLOGIES)
  {
    *err = iBase_INVALID_ENTITY_TOPOLOGY;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }
  // check validity of type
  if(entity_type_requester < FMDB_VERTEX || entity_type_requester > FMDB_ALLTYPE
    || entity_type_requested < FMDB_VERTEX || entity_type_requested > FMDB_ALLTYPE)
  {
    *err = iBase_INVALID_ENTITY_TOPOLOGY;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }

  pPartEntIter iter;
  pMeshEnt ent;
  int inSet = 0, partID, ent_type;
  std::vector<pMeshEnt> ents;

  // looping through entities with entity_topology and entity_type on part and see if they are in set
  int entEnd = FMDB_PartEntIter_Init((pPart)part, entity_type_requester, entity_topology_requester, iter);

  while (!entEnd)
  {
    entEnd = FMDB_PartEntIter_GetNext(iter, ent);
    if (entEnd) break;

    inSet=1;
    if ((iBase_EntitySetHandle)instance!=entity_set_handle) // root set    
      FMDB_Set_FindEnt((pEntSet)entity_set_handle, ent, &inSet);
    if (inSet) 
    {
      // need to check to make sure the part owns this entity
      FMDB_Ent_GetOwnPartID(ent, (pPart)part, &partID);
      if (!(partID == FMDB_Part_GlobID((pPart)part)))
        continue;
      
      // add to ents vector
      ents.push_back(ent);
    }
  }

  FMDB_PartEntIter_Del(iter);
  
  // allocate memory for entity_handles
  if (*entity_handles_allocated == 0)  
  {
    *entity_handles = (iBase_EntityHandle*)(calloc(ents.size(), sizeof(iBase_EntityHandle))); 
    if(*entity_handles == 0) 
    {
      *err = iBase_MEMORY_ALLOCATION_FAILED;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
    *entity_handles_allocated = ents.size();
  }
  else if(*entity_handles_allocated < ents.size()) 
  {
    *err = iBase_BAD_ARRAY_SIZE; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  // allocate memory for offset (same length as entity_handles)
  if (*offset_allocated == 0)  
  {
    *offset = (int*)(calloc(ents.size(), sizeof(int))); 
    if(*offset == 0) 
    {
      *err = iBase_MEMORY_ALLOCATION_FAILED;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
    *offset_allocated = ents.size();
  }
  else if(*offset_allocated < ents.size()) 
  {
    *err = iBase_BAD_ARRAY_SIZE; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  // add entities to entity_handles
  for (int i = 0; i < ents.size(); i++)
    *entity_handles[i] = (iBase_EntityHandle)ents[i];
  *entity_handles_size = ents.size();

  // find all adjacent entities.  add to set to remove duplicates
  // keep track of count (that includes duplicates)
  int num_ents = 0;
  std::vector<pMeshEnt> vecAdj;
  std::set<pMeshEnt> adjEnts;
  for (int i = 0; i < ents.size(); i++)
  {
    FMDB_Ent_GetAdj(ent, entity_type_requested, 1, vecAdj);
    num_ents += vecAdj.size();

    for (int j = 0; j < vecAdj.size(); j++)
      vecAdj.push_back((pMeshEnt)(vecAdj[j]));
  }

  // allocate memory for adj_entity_handles
  if (*adj_entity_handles_allocated == 0)  
  {
    *adj_entity_handles = (iBase_EntityHandle*)(calloc(adjEnts.size(), sizeof(iBase_EntityHandle))); 
    if(*adj_entity_handles == 0) 
    {
      *err = iBase_MEMORY_ALLOCATION_FAILED;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
    *adj_entity_handles_allocated = adjEnts.size();
  }
  else if(*adj_entity_handles_allocated < adjEnts.size()) 
  {
    *err = iBase_BAD_ARRAY_SIZE; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  // allocate memory for adj_entity_indices
  if (*adj_entity_indices_allocated == 0)  
  {
    *adj_entity_indices = (int*)(calloc(num_ents, sizeof(int))); 
    if(*adj_entity_indices == 0) 
    {
      *err = iBase_MEMORY_ALLOCATION_FAILED;
     FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
     return;
    }
    *adj_entity_indices_allocated = num_ents;
  }
  else if(*adj_entity_indices_allocated < num_ents) 
  {
    *err = iBase_BAD_ARRAY_SIZE; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  // add adj entities to adj_entity_handles
  std::set<pMeshEnt>::iterator it;
  int i;
  for (i = 0, it = adjEnts.begin(); it != adjEnts.end(); it++, i++)
    (*adj_entity_handles)[i] = (iBase_EntityHandle)*it;
  *adj_entity_handles_size = adjEnts.size();

  // populate adj_entity_indices.  also update values in offset
  int index = 0, loc;
  for (int i = 0; i < ents.size(); i++)
  {
    *offset[i] = index;
    
    FMDB_Ent_GetAdj(ent, entity_type_requested, 1, vecAdj); // loop through all adj ents for this entity
    for (int j = 0; j < vecAdj.size(); j++, index++)
    {
      for (int k = 0; k < *adj_entity_handles_size; k++) // find index in adj_entity_handles
      {
        if ((pMeshEnt)(vecAdj[j]) == (pMeshEnt)(*adj_entity_handles[k]))
        {
          (*adj_entity_indices)[index] = k; 
          break;
        }

      }
    }
  }

  *adj_entity_indices_size = num_ents;
  *offset_size = ents.size();

  return;
}

void iMeshP_getEntities(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle part,
            const iBase_EntitySetHandle entity_set,
            int entity_type,
            int entity_topology,
            iBase_EntityHandle **entity_handles,
            int *entity_handles_allocated,
            int *entity_handles_size,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  pPart part_hdl = (pPart)part;
  
  bool check = checkTypeTopo((int)entity_type,(int)entity_topology);
			   
  if (!check) {
    *err = iBase_BAD_TYPE_AND_TOPO; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    *entity_handles_size = 0;
    return;
  }
  
  int numEnt=0, counter=0;
  
  // compute numEnt for part
  if (entity_topology == iMesh_ALL_TOPOLOGIES)  
  {
    if (entity_type<4)
      numEnt = part_hdl->size(entity_type);
    else 
      numEnt = part_hdl->size(0)+part_hdl->size(1)+part_hdl->size(2)+part_hdl->size(3);
   }		  
  else // topology specified
    numEnt=part_hdl->getNumTopo(entity_topology);
  
  if (numEnt == 0) {
    *entity_handles_size = 0;
    return;
  }
  
  // allocate memory  
  if (*entity_handles_allocated == 0)  
  {  // should check *entity_handles, not entity_handles. 0826 
    *entity_handles = (iBase_EntityHandle*)(calloc(numEnt, sizeof(iBase_EntityHandle))); 
    if(*entity_handles == 0) 
    {
      *err = iBase_MEMORY_ALLOCATION_FAILED;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
    *entity_handles_allocated = numEnt;
  }
  else 
    if(*entity_handles_allocated<numEnt) 
    {
      *err = iBase_BAD_ARRAY_SIZE; 
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
  
  *entity_handles_size = numEnt;

  switch (entity_type)
  {
    case iBase_VERTEX: 
    case iBase_EDGE: 
      for (mPart::iterall it=part_hdl->beginall(entity_type); it!=part_hdl->endall(entity_type);++it)
          (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 

      break;	     
    case iBase_FACE:
    {
      if (entity_topology == iMesh_ALL_TOPOLOGIES)      
      {
        for (mPart::iterall it=part_hdl->beginall(entity_type); it!=part_hdl->endall(entity_type);++it)
          (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
      }
      if(entity_topology == iMesh_TRIANGLE || entity_topology == iMesh_QUADRILATERAL)   
      {
        for (mPart::iterall it=part_hdl->beginall(entity_type); it!=part_hdl->endall(entity_type);++it)
          if(entity_topology == (iMesh_EntityTopology)(*it)->getTopo())  
            (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
      }
      break;	 
    }
    case iBase_REGION:
    {
      if (entity_topology == iMesh_ALL_TOPOLOGIES)   
      {
        for (mPart::iterall it=part_hdl->beginall(entity_type); it!=part_hdl->endall(entity_type);++it)
	       (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
      }
      if( entity_topology == iMesh_TETRAHEDRON || entity_topology == iMesh_HEXAHEDRON || entity_topology == iMesh_PYRAMID || entity_topology == iMesh_PRISM )  
      {
        for (mPart::iterall it=part_hdl->beginall(entity_type); it!=part_hdl->endall(entity_type);++it)
          if(entity_topology == (iMesh_EntityTopology)(*it)->getTopo())    
            (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
      }
      break;	 
    }
    case iBase_ALL_TYPES:
    {
      switch (entity_topology)
      {
        case iMesh_POINT:
	      {
          for (mPart::iterall it=part_hdl->beginall(0); it!=part_hdl->endall(0);++it)
            (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
        }
	      break;	 
        case iMesh_LINE_SEGMENT:
	      {
          for (mPart::iterall it=part_hdl->beginall(1); it!=part_hdl->endall(1);++it)
            (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
          break;	 
	      }
        case iMesh_TRIANGLE:
	      {
          for (mPart::iterall it=part_hdl->beginall(2); it!=part_hdl->endall(2);++it)
            if(entity_topology == (iMesh_EntityTopology)((*it)->getTopo()))
              (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
	      }
	      break;	 
	       
        case iMesh_QUADRILATERAL:      // QUAD
	      {
          for (mPart::iterall it=part_hdl->beginall(2); it!=part_hdl->endall(2);++it)
            if(entity_topology == (iMesh_EntityTopology)((*it)->getTopo()))
              (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
	      }
	      break;	 
	      
        case iMesh_TETRAHEDRON: 
	      {
          for (mPart::iterall it=part_hdl->beginall(3); it!=part_hdl->endall(3);++it)
            if(entity_topology == (iMesh_EntityTopology)((*it)->getTopo()))
              (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
	      }
	      break;	 
	      
        case iMesh_HEXAHEDRON:       // HEX
	      {
          for (mPart::iterall it=part_hdl->beginall(3); it!=part_hdl->endall(3);++it)
              if(entity_topology == (iMesh_EntityTopology)((*it)->getTopo()))
                (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
	      }
	      break;	 
	      
        case iMesh_PYRAMID:       //  PYR
	      {
          for (mPart::iterall it=part_hdl->beginall(3); it!=part_hdl->endall(3);++it)
            if(entity_topology == (iMesh_EntityTopology)((*it)->getTopo()))
              (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
	      }
	      break;	 
	      
        case iMesh_PRISM :       // WED
	      {
          for (mPart::iterall it=part_hdl->beginall(3); it!=part_hdl->endall(3);++it)
            if(entity_topology == (iMesh_EntityTopology)((*it)->getTopo()))
              (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
	      }
	      break;	
        case iMesh_ALL_TOPOLOGIES:
	      {
          for (int dim=0; dim<4; ++dim)
            for (mPart::iterall it=part_hdl->beginall(dim); it!=part_hdl->endall(dim);++it)
              (*entity_handles)[counter++]= (iBase_EntityHandle)(*it); 
	      }
	      break;	 
        default: assert(numEnt==0);
          break;	
	    }
    } // case all_types
  } //switch
}

void iMeshP_initEntIter(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle part,
            const iBase_EntitySetHandle entity_set,
            const int entity_type,
            const int entity_topology,
            const int resilient,
            iBase_EntityIterator* entity_iterator,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  if (resilient)
  {
    *err = iBase_NOT_SUPPORTED;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  bool check = checkTypeTopo((int)entity_type, (int)entity_topology);
  if (!check)
  {
    *err = iBase_BAD_TYPE_AND_TOPO;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  std::vector<pPart> part_vec;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, SCUTIL_CommRank(), part_vec);

  // check if the part_handle is valid (local part handle)
  std::vector<pPart>::iterator iter = std::find(part_vec.begin(), part_vec.end(), (pPart)part);
  if (iter == part_vec.end() )
  {    *err = iBase_FAILURE; // iBase_INVALID_PART_HANDLE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  pIterator pIter;
  createIterator((pPart)part, pIter, entity_type, entity_topology);
  *entity_iterator = (iBase_EntityIterator)pIter;

  return;
}

void iMeshP_initEntArrIter(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle part,
            const iBase_EntitySetHandle entity_set,
            const int entity_type,
            const int entity_topology,
            const int array_size,
            const int resilient,  
            iBase_EntityArrIterator* entArr_iterator,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  if (resilient)
  {
    *err = iBase_NOT_SUPPORTED;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }         
            
  bool check = checkTypeTopo((int)entity_type, (int)entity_topology);
  if (!check)
  {         
    *err = iBase_BAD_TYPE_AND_TOPO;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }
  
  std::vector<pPart> part_vec;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, SCUTIL_CommRank(), part_vec);

  // check if the part_handle is valid (local part handle)
  std::vector<pPart>::iterator iter = std::find(part_vec.begin(), part_vec.end(), (pPart)part);
  if (iter == part_vec.end() )
  {    *err = iBase_FAILURE; // iBase_INVALID_PART_HANDLE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }
  pIterator pIter;
  createIterator((pPart)part, pIter, entity_type, entity_topology, array_size);
  *entArr_iterator = (iBase_EntityArrIterator)pIter;
  return;
}
