/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "iMeshP.h"
#include "FMDB.h"

void iMeshP_getNumPartNbors(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle part,
            int entity_type,
            int *num_part_nbors,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

 // constraint: single part per mesh instance only
  int numPart=0;
  FMDB_Mesh_GetNumPart ((pMeshMdl)instance, &numPart);
  if (numPart > 1)
  {
    *err = iBase_NOT_SUPPORTED;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  std::vector<pPart> part_vec;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, SCUTIL_CommRank(), part_vec);

  // check if the part_handle is valid (local part handle)
  std::vector<pPart>::iterator iter = std::find(part_vec.begin(), part_vec.end(), (pPart)part);
  if (iter == part_vec.end() ) 
  {
    *err = iBase_FAILURE; // iBase_INVALID_PART_HANDLE; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }

  *err = FMDB_Part_GetNumNborPart ((pPart)part, entity_type, num_part_nbors);
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  return;
}


void iMeshP_getNumPartNborsArr(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle *parts,
            int parts_size,
            int entity_type,
            int **num_part_nbors,
            int *num_part_nbors_allocated,
            int *num_part_nbors_size,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  // constraint: single part per mesh instance only
  int numPart=0;
  FMDB_Mesh_GetNumPart ((pMeshMdl)instance, &numPart);
  if (numPart > 1)
  {
    *err = iBase_NOT_SUPPORTED;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  if ( *num_part_nbors==0||*num_part_nbors_allocated==0 ) 
  {
    *num_part_nbors = (int*) calloc( parts_size, sizeof(int));
    *num_part_nbors_allocated = parts_size;
    if( *num_part_nbors==0 ) 
    {
      *err = iBase_MEMORY_ALLOCATION_FAILED;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
  }
  else if( *num_part_nbors_allocated < parts_size ) {
      *err = iBase_BAD_ARRAY_SIZE;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return; 
  }
  
  int num_part_nbor;
  for( int i=0; i<parts_size; ++i)
  {
    iMeshP_getNumPartNbors(instance, partition, parts[i], entity_type, &num_part_nbor, err);  
    (*num_part_nbors)[i]= num_part_nbor;
    if (*err!= iBase_SUCCESS )
    {
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return; 
    }
  }
  *num_part_nbors_size=parts_size;
}


void iMeshP_getPartNbors(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle part,
            int entity_type,
            int *num_part_nbors,
            iMeshP_Part **nbor_part_ids,
            int *nbor_part_ids_allocated,
            int *nbor_part_ids_size,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

 // constraint: single part per mesh instance only
  int numPart=0;
  FMDB_Mesh_GetNumPart ((pMeshMdl)instance, &numPart);
  if (numPart > 1)
  {
    *err = iBase_NOT_SUPPORTED;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  std::vector<pPart> part_vec;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, SCUTIL_CommRank(), part_vec);

  // check if the part_handle is local 
  std::vector<pPart>::iterator iter = std::find(part_vec.begin(), part_vec.end(), (pPart)part);
  if( iter == part_vec.end() ) {
    *err = iBase_FAILURE; // iBase_INVALID_PART_HANDLE; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }
  vector<int> part_nbors;
  FMDB_Part_GetNborPartID ((pPart)part, entity_type, part_nbors);

  if (*nbor_part_ids==0 || *nbor_part_ids_allocated==0)
  {      
    *nbor_part_ids = (iMeshP_Part*)(calloc(part_nbors.size(), sizeof(iMeshP_Part))); 
    if( *nbor_part_ids==0 ){ 
      *err = iBase_MEMORY_ALLOCATION_FAILED;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
    *nbor_part_ids_allocated = part_nbors.size(); 
  }
  else if (*nbor_part_ids_allocated < part_nbors.size())
  {
    *err = iBase_BAD_ARRAY_SIZE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  *num_part_nbors = part_nbors.size(); 
  *nbor_part_ids_size = part_nbors.size(); 

  int icount=0; 
  for(std::vector<int>::iterator nbor_it = part_nbors.begin(); nbor_it!= part_nbors.end(); ++nbor_it) 
    (*nbor_part_ids)[icount++]= (iMeshP_Part)(*nbor_it);       
  return;
}

void iMeshP_getPartNborsArr(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle *parts,
            const int parts_size,
            int entity_type,
            int **num_part_nbors,
            int *num_part_nbors_allocated,
            int *num_part_nbors_size,
            iMeshP_Part **nbor_part_ids,
            int *nbor_part_ids_allocated,
            int *nbor_part_ids_size,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  // constraint: single part per mesh instance only
  int numPart=0, icount;
  FMDB_Mesh_GetNumPart ((pMeshMdl)instance, &numPart);
  if (numPart > 1)
  {
    *err = iBase_NOT_SUPPORTED;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  /// allocate memory for num_part_nbors  
  if ( *num_part_nbors==0 || *num_part_nbors_allocated==0 ) 
  { 
    *num_part_nbors = (int*)(calloc(parts_size, sizeof(int)));  
    if (*num_part_nbors ==0 )
    { 
      *err = iBase_MEMORY_ALLOCATION_FAILED;  
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
    *num_part_nbors_allocated = parts_size;     
  }
  else if ( *num_part_nbors_allocated<parts_size ) 
  {
    *err = iBase_BAD_ARRAY_SIZE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }

  std::vector<pPart> part_vec;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, SCUTIL_CommRank(), part_vec);

  vector<int> nbor_pids; 

  for(icount=0; icount<parts_size; ++icount) // for each part
  {
    pPart part_hdl = (pPart)(parts[icount]); 
    
    // check if the part_handle is valid (local)
    std::vector<pPart>::iterator iter = std::find(part_vec.begin(), part_vec.end(), part_hdl);
    if( iter == part_vec.end() ) {
      *err = iBase_FAILURE; // iBase_INVALID_PART_HANDLE; 
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return; 
    }
 
    vector<int> part_nbors;
    *err = FMDB_Part_GetNborPartID (part_hdl, entity_type, part_nbors);
    if (*err) 
    {
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
    (*num_part_nbors)[icount] = part_nbors.size();    
    
    for(vector<int>::iterator pid_it=part_nbors.begin(); pid_it!=part_nbors.end(); ++pid_it) 
      nbor_pids.push_back(*pid_it);
  }  
  *num_part_nbors_size = parts_size;

  // allocate memory for nbor_part_ids  
  int size=nbor_pids.size(); 
  if( *nbor_part_ids_allocated==0 || *nbor_part_ids==0  ) 
  {   
    *nbor_part_ids = (iMeshP_Part*)(calloc(size, sizeof(iMeshP_Part))); 
    if(*nbor_part_ids ==0 )
    {
      *err = iBase_MEMORY_ALLOCATION_FAILED;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
    *nbor_part_ids_allocated=nbor_pids.size();
  }
  else if ( *nbor_part_ids_allocated < size ) 
  {        
    *err = iBase_BAD_ARRAY_SIZE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }

  icount=0; 
  for(vector<int>::iterator pid_it=nbor_pids.begin(); pid_it!=nbor_pids.end(); ++pid_it) 
    (*nbor_part_ids)[icount++] = (iMeshP_Part)(*pid_it); 
    
  *nbor_part_ids_size = size; 
  return;
} 

void iMeshP_getNumPartBdryEnts(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle part, 
            int entity_type, 
            int entity_topology, 
            iMeshP_Part target_part_id, 
            int *num_entities, 
            int *err)
{
  *err = FMDB_Part_GetNumPartBdry ((pPart)part, target_part_id, entity_type, 
                             entity_topology, num_entities);
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  return;
}

void iMeshP_getPartBdryEnts(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iMeshP_PartHandle part, 
            int entity_type, 
            int entity_topology, 
            iMeshP_Part target_part_id, 
            iBase_EntityHandle **entity_handles,
            int *entity_handles_allocated,
            int *entity_handles_size, 
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
 
  pPart part_hdl = (pPart)part; 
  pMeshEnt ent;
  bool ok = checkTypeTopo((int)entity_type,(int)entity_topology);
  if (!ok) {
    *err = iBase_BAD_TYPE_AND_TOPO; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }
  
  std::vector<pPart> part_vec;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, SCUTIL_CommRank(), part_vec);

  // check if the part_handle is valid (local part handle)
  std::vector<pPart>::iterator iter = std::find(part_vec.begin(), part_vec.end(), (pPart)part);
  if (iter == part_vec.end() ) 
  {
    *err = iBase_FAILURE; // iBase_INVALID_PART_HANDLE; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }

  int pid = target_part_id; 
  if( pid==FMDB_Part_GlobID(part_hdl))
  {                  
    *err = iBase_FAILURE;                // iBase_INVALID_PART_ID
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }    

  std::vector<pMeshEnt> entities;
  int isOnPartBdry;  
  pPartEntIter pent_it;
  int iterEnd =FMDB_PartEntIter_InitPartBdry (part_hdl, (int)target_part_id, entity_type, entity_topology, pent_it);
  if (!iterEnd)
  {
    while (!iterEnd)
    {
      iterEnd = FMDB_PartEntIter_GetNext(pent_it, ent);
      if(iterEnd) break;
      entities.push_back(ent); 
    }
    FMDB_PartEntIter_Del(pent_it);
  }
 
  if( *entity_handles==0 || *entity_handles_allocated==0 ) { 
    *entity_handles = (iBase_EntityHandle*)(calloc( entities.size(), sizeof(iBase_EntityHandle))); 
    if( *entity_handles==0 ) {
      *err = iBase_MEMORY_ALLOCATION_FAILED;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }
    
    std::vector<pEntity>::iterator iter=entities.begin(), itend=entities.end(); 
    int icount=0; 
    for (; iter!=itend; iter++){
      (*entity_handles)[icount++] = (iBase_EntityHandle)(*iter);  
    }
    
    *entity_handles_allocated = entities.size(); 
  }
  else if ( (unsigned int)(*entity_handles_allocated)<entities.size() ) 
  {      
    *err = iBase_BAD_ARRAY_SIZE;  
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }
  *entity_handles_size = entities.size();
}

void iMeshP_initPartBdryEntIter(
    iMesh_Instance instance, 
    const iMeshP_PartitionHandle partition, 
    const iMeshP_PartHandle part, 
    int entity_type, 
    int entity_topology, 
    iMeshP_Part target_part_id, 
    const int resilient,
    iBase_EntityIterator* entity_iterator, 
    int* err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  if (resilient)
  {
    *err = iBase_NOT_SUPPORTED;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  bool check = checkTypeTopo((int)entity_type, (int)entity_topology);
  if (!check) 
  {
    *err = iBase_BAD_TYPE_AND_TOPO; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }
    
  std::vector<pPart> part_vec;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, SCUTIL_CommRank(), part_vec);

  // check if the part_handle is valid (local part handle)
  std::vector<pPart>::iterator iter = std::find(part_vec.begin(), part_vec.end(), (pPart)part);
  if (iter == part_vec.end() ) 
  {
    *err = iBase_FAILURE; // iBase_INVALID_PART_HANDLE; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }
    
  // make sure nbor_part_ids is not current_part_id
  if( SCUTIL_CommRank()==(int)target_part_id) 
  {
    *err=iBase_INVALID_ARGUMENT; 
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return; 
  }
  pIterator pIter;
  createPartBdryIterator((pPart)part, pIter, (int)target_part_id, entity_type, entity_topology);
  *entity_iterator = (iBase_EntityIterator)pIter;
}

void iMeshP_initPartBdryEntArrIter(
    iMesh_Instance instance, 
    const iMeshP_PartitionHandle partition, 
    const iMeshP_PartHandle part, 
    int entity_type, 
    int entity_topology, 
    int array_size, 
    iMeshP_Part target_part_id, 
    const int resilient,
    iBase_EntityArrIterator* entity_iterator, 
    int* err)

{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  if (resilient)
  {
    *err = iBase_NOT_SUPPORTED;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  bool check = checkTypeTopo((int)entity_type, (int)entity_topology);
  if (!check)
  {
    *err = iBase_BAD_TYPE_AND_TOPO;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  std::vector<pPart> part_vec;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, SCUTIL_CommRank(), part_vec);

  // check if the part_handle is valid (local part handle)
  std::vector<pPart>::iterator iter = std::find(part_vec.begin(), part_vec.end(), (pPart)part);
  if (iter == part_vec.end() )
  {    *err = iBase_FAILURE; // iBase_INVALID_PART_HANDLE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }
  pIterator pIter;
  int partid=static_cast<int>(target_part_id);
  createPartBdryIterator((pPart)part, pIter,partid, entity_type,entity_topology, array_size);
  *entity_iterator = (iBase_EntityArrIterator)pIter;
  return;

}


