/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "iMeshP.h"
#include "FMDB.h"
#include "mpi.h"
#include "mMesh.h"

void iMeshP_createPartitionAll(
            iMesh_Instance instance,
            MPI_Comm communicator,
            iMeshP_PartitionHandle *partition,
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  SCUTIL_Init(communicator);
  *partition = (iMeshP_PartitionHandle)instance;

  // re-assign part id since FMDB_Init is called before SCUTIL_Init(communicator);
  int counter=0, maxNumPart;
  vector<pPart> part_handles;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, SCUTIL_CommRank(), part_handles);
  FMDB_Mesh_GetNumPart ((pMeshMdl)instance, &maxNumPart);
  for (std::vector<pPart>::iterator pit=part_handles.begin(); pit!=part_handles.end(); ++pit)
  {
    (*pit)->setId(SCUTIL_CommRank()*maxNumPart+counter);
    counter++;
  }
  return;
}


void iMeshP_destroyPartitionAll(
            iMesh_Instance instance,
            iMeshP_PartitionHandle partition,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  ParUtil::Instance()->Finalize(0); // this doesn't call MPI_Finalize
  return;
}

void iMeshP_loadAll(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iBase_EntitySetHandle entity_set,
            const char *name, 
            const char *options,
            int *err, 
            int name_len, 
            int options_len)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
  
  // load a partitioned mesh by default
  int distributed=0;
  // FMDB takes digits for options. 
  // If zero is given, a serial mesh is loaded and partitioned
  // if non-zero is given, a distributed mesh is loaded from multiple files
  if (!options_len)
    distributed = 1;

  *err = FMDB_Mesh_LoadFromFile ((pMeshMdl)instance, name, distributed); 
  if (*err) 
  {
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }
#ifdef DEBUG
  FMDB_Mesh_DspNumEnt((pMeshMdl)instance);
#endif
  // initialize adj table
  for (int i=0; i<4; ++i)
    for (int j=0; j<4; ++j)
      if (i==j) // diagonal
        ((pMeshMdl)instance)->MRM[i][j] = iBase_AVAILABLE;
      else // non-diagonal
        ((pMeshMdl)instance)->MRM[i][j] = iBase_ALL_ORDER_1;

  int mesh_dim;
  FMDB_Mesh_GetDim ((pMeshMdl)instance, &mesh_dim);
  if (mesh_dim!=3)
  {
    ((pMeshMdl)instance)->MRM[3][3] = iBase_UNAVAILABLE; 
    for (int i=0; i<3; ++i)
      ((pMeshMdl)instance)->MRM[3][i] = ((pMeshMdl)instance)->MRM[i][3] = iBase_UNAVAILABLE; 
  }
}

void iMeshP_saveAll(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iBase_EntitySetHandle entity_set,
            const char *name, 
            const char *options,
            int *err, 
            const int name_len, 
            int options_len)
{
  *err = FMDB_Mesh_WriteToFile((pMeshMdl)instance, name, 1);
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  return;
}


void iMeshP_getPartitionComm(
            iMesh_Instance instance,
            iMeshP_PartitionHandle partition,
            MPI_Comm *communicator,
            int *err)
{
  *err = SCUTIL_GetComm (communicator);
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  return;
}


void iMeshP_syncPartitionAll(
            iMesh_Instance instance,
            iMeshP_PartitionHandle partition,
            int *err)
{
  *err = FMDB_Mesh_SyncPtn ((pMeshMdl)instance);
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  return;
}


void iMeshP_getNumPartitions(
            iMesh_Instance instance,
            int *num_partitions,
            int *err)
{
  *err = iBase_SUCCESS;
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  *num_partitions=1;
  return;
}



void iMeshP_getPartitions(
            iMesh_Instance instance,
            iMeshP_PartitionHandle **partitions,
            int *partitions_allocated, 
            int *partitions_size, 
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  int size=1; // assumption: single partition
  if ( *partitions==0 || *partitions_allocated==0 ) {
    *partitions = (iMeshP_PartitionHandle*)( calloc(size, sizeof(iMeshP_PartitionHandle)) );
    if(*partitions==0) {
      *err = iBase_MEMORY_ALLOCATION_FAILED;
      FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
      return;
    }

    *partitions_allocated = size;
  }
  else if( *partitions_allocated < size) {
    *err = iBase_BAD_ARRAY_SIZE;
    FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
    return;
  }

  (*partitions)[0]= (iMeshP_PartitionHandle)instance;
  *partitions_size=size;
 
  return;
}


/* Note:
 * This assumes that iMeshP_getNumOfType only returns the number of entities on the 
 * specified part that are OWNED by that part.
 */
void iMeshP_getNumOfTypeAll(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iBase_EntitySetHandle entity_set,
            int entity_type, 
            int *num_type, 
            int *err)
{
 *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);
 
  int local_num_type, sum_num_type=0, result, isList, isPO;

  // we need to loop over all parts on this partition
  std::vector<pPart> parts;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, SCUTIL_CommRank(), parts);

  for (int i = 0; i < parts.size(); i++) 
  {
    iMeshP_getNumOfType(instance, partition, (iMeshP_PartHandle)parts[i], entity_set, entity_type, &local_num_type, err);
    sum_num_type  += local_num_type;
  }

  if ((iBase_EntitySetHandle)instance==entity_set) // root set
    MPI_Allreduce(&sum_num_type, &result, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
  else // entity set
    result = sum_num_type;

  *num_type = result;

  return;
}


/* Note:
 * This assumes that iMeshP_getNumOfTopo only returns the number of entities on the 
 * specified part that are OWNED by that part.
 */
void iMeshP_getNumOfTopoAll(
            iMesh_Instance instance,
            const iMeshP_PartitionHandle partition,
            const iBase_EntitySetHandle entity_set,
            int entity_topology, 
            int *num_topo, 
            int *err)
{
  *err = iBase_SUCCESS; 
  FMDB_Mesh_SetLastError ((pMeshMdl)instance, *err);

  int local_num_topo, sum_num_topo=0, result, isList, isPO;

  // we need to loop over all parts on this partition
  std::vector<pPart> parts;
  FMDB_Mesh_GetAllPart((pMeshMdl)instance, SCUTIL_CommRank(), parts);

  for (int i = 0; i < parts.size(); i++) 
  {
    local_num_topo=0;
    iMeshP_getNumOfTopo(instance, partition, (iMeshP_PartHandle)parts[i], entity_set, entity_topology, &local_num_topo, err);
    sum_num_topo += local_num_topo;
  }
  
  if ((iBase_EntitySetHandle)instance==entity_set) // root set
    MPI_Allreduce(&sum_num_topo, &result, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
  else 
    result = sum_num_topo;

  *num_topo = result;
  return;
}


