/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#ifndef _H_ParUtil
#define _H_ParUtil
#include <string>
#include <memory>
#include <iostream>
#include <vector>
#include <sys/utsname.h>
#include <sys/resource.h>
#include <malloc.h>
#ifdef FMDB_PARALLEL
#include "mpi.h" 
#include "IPComMan.h"
#include <list>
#include "pmUtility.h"
#else 
typedef void* MPI_Comm;
typedef void* MPI_Request;
#endif // end of #ifdef FMDB_PARALLEL
/**
   ParUtil is a Singleton. It gives some
   general services for parallel implementation.
*/

class ParUtil {
  ParUtil();
public:
    /// Constructor and destructors should be private but I'm fed up of this
    /// c++ compiler always shouting
  ~ParUtil();
  /// Message severity level
  enum MessageLevel {DEBUG1,DEBUG2,INFO,WARNING,ERROR};
  /// returne the only instance
  static ParUtil* Instance();
  /// initialization, needed for mpi and IPComMan
  void init(MPI_Comm); 

  /// adds a barrier
  void Barrier(int, const char*);
  /// close communications
  void Finalize(bool do_mpi_finalize);

// time/mem resource 
  void reset() { resetTime(); resetMem(); }
  void resetTime() { getrusage(RUSAGE_SELF, &ruse_start); }
  void resetMem() { meminfo_start = mallinfo(); }

  double getTimeDiff();
  double getMemDiff();
  double getCurMem();
  double getCurTime();

  void sysinfo();
  void report(char const* title, double time, double memory);
// end of resource

  /// compute wall time
  double wTime () const;
  /// set the verbosity level (0,1,2,3)
  inline void setVertbosityLevel(int i){vl = i;}
  /// prints a message, same format as printf
  void Msg(MessageLevel lev, const char *fmt, ...);
  static void setBufferSize (int b) {buffer_size = b;}
  int splitComm (int numProc);
  void setComm(MPI_Comm comm);
  inline MPI_Comm getComm() { return mycomm; }
  inline void set_maxMeshDim(int dim) { maxMeshDim = dim; }
  inline int get_maxMeshDim() { return maxMeshDim; }
#ifdef FMDB_PARALLEL
  inline int group() { return split_key; }
  inline int rank() { return myrank; }
  inline int size() { return mysize; }
  inline int master() { return myrank==0; }
  inline int getCurNumParts() {return curNumParts; }  
  void setCurNumParts(int num);        
  void setTgtNumParts(int num);   
  inline int getTgtNumParts()  {return tgtNumParts; }
  void setUsrNumParts(int num);
  inline int getUsrNumParts()  {return usrNumParts; }  
  /// Communication Manager
  inline IPComMan *ComMan() { return CM; }
#else
  inline int group() { return 0; }
  inline int rank() { return 0; }
  inline int size() { return 1; }
  inline int master() { return 1; }
#endif
  inline void   resetBarrierSensor() { timeSpentOnBarriers = 0; }
  inline double getBarrierSensor()   { return timeSpentOnBarriers; }
private:
  static std::auto_ptr<ParUtil> instance;
  int vl;

  MPI_Comm mycomm; 
  int maxMeshDim;
  double timeSpentOnBarriers;
  struct rusage ruse_start;   // <sys/resource.h>
  struct mallinfo meminfo_start;   // <malloc.h>
#ifdef FMDB_PARALLEL
  int initialized;
  int finalized;
  int myrank;
  int mysize; 
  IPComMan *CM;
  int curNumParts;                   // current_local_num_parts_per_process, max_num 
  int tgtNumParts;                   // target_local_num_parts_per_process, max_num
  int usrNumParts; 
  int split_key;
public: 
  std::vector<int> localPids;             // local part ids for on-process parts 
#endif
  static int buffer_size; 
};

#endif
