/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#ifndef _mPart_H_
#define _mPart_H_
#include "GenTag.h"
#include "mEntityContainer.h"
#include "mIdGenerator.h"
#include "mIterator.h"
#include "mEntity.h"
#include "modeler.h"
#include "mEntityGroup.h"

#include <iosfwd>
#include <list>
#include "mAttachableDataContainer.h"
#include "mGhost.h"
  class mEntity;
  class mEntityGroup;
  class mFace;
  class mEdge;
  class mVertex;
  class mRegion;
  class mTet;
  class mHex;
  class mPrism;
  class mPyramid;
  class FMDB_LoadBalancerCallbacks;
  class FMDB_DataExchanger;
  class mGhost;
  
  struct rp_int 
  {
    int i;
    mEntity *entity;
  };
 
  struct rp_int2
  {
    int i;
    int j;
    mEntity *entity;
  };

  /**
   mPart class provides an interface to
   a complete mesh that can be distributed.
  */
class mPart : public Taggable, public mAttachableDataContainer
{
protected:
// ********************************************
//		MESH + MODEL
// ********************************************
    mMesh* mesh_instance;
    pGModel theSolidModel;
    
// ********************************************
//		ID
// ********************************************
    // part id
    unsigned int iD; 
    /// an iD generator for vertices
    mIdGenerator theIdGenerator;

// ********************************************
//		CONTAINER
// ********************************************
    /// container with 4 hash tables, all mesh entities
    mPartEntityContainer allEntities;

    std::vector<EntitySet<mEntity>* > allPSet;
    std::vector<mGhost*> ghosts; 

// ********************************************
//           ENTITY-GROUP-RELATED
// ********************************************
public:    
   typedef std::vector<EntitySet<mEntity>* >::iterator psetIter;
   
   EntitySet<mEntity>* createPSet(int);
   void deletePSet(EntitySet<mEntity>* PSet, int force=0);    // force=1, delete attached data on entities in the group
   
   psetIter beginPSet() {return allPSet.begin(); }
   psetIter endPSet() {return allPSet.end(); }
   int numPSet() {return allPSet.size();  } 
   void clearPSet(); 

   conType getContainerType() const  {return PART; }
   int dimension;
   void setDimension(int dim) { if(dim>0) dimension=dim; }
   int getDimension() const;

// ********************************************
//		typedef
// ********************************************
    /// iterators on all entities (the whole graph)
    typedef mPartEntityContainer::iter       iterall;
    /// iterators on leaves
    typedef mLeavesIterator                  iter;

// ********************************************
//		CONSTRUCTOR/DESTRUCTOR
// ********************************************
    /// constructor
    mPart(int id = 1, pGModel model = 0, mMesh* mesh=0);
    virtual ~mPart(); 

// ********************************************
//		PART ID
// ********************************************
    inline void setId(int id) { iD = id; }
    inline int getId(){return iD;}

// ********************************************
//		MESH LOAD RELATED
// ********************************************
    int representationFlag; // 0: non-one level representation
                            // 1: one-level representation    
    void setRepresentationFlag(int i) { representationFlag=i; }
    int getRepresentationFlag() { return representationFlag; }
    
// ********************************************
//		MESH MODEL
// ********************************************
    mMesh* getMesh() {return mesh_instance; }
    pGModel  getSolidModel () const {return theSolidModel;}
    void setSolidModel (pGModel m);
    
// ********************************************
//		VERTEX ID
// ********************************************
    /// we have one id generator for vertices, that function returns it
    inline const mIdGenerator & getIdGenerator () const 
    {return theIdGenerator;} 
    virtual int getNewVertexId();
    virtual void setNewVertexId(int);

#ifdef FMDB_PARALLEL
   void pmSetNewVertexId(int);
   void updateId() { theIdGenerator.update(); }
#endif

// ********************************************
//		GENTITY
// ********************************************
    typedef pGEntity  (*GEntity_FP)(pGModel,int,int);
    GEntity_FP classif_fp;    
    GEntity_FP getGEntity_FP() { return classif_fp; }
    void setGEntity_FP(GEntity_FP fp) { classif_fp = fp; }
    pGEntity getGEntity (int tag,int type);
    pGEntity getGEntity(int tag,int type,pGEntity(*)(pGModel,int,int));    

// ********************************************
//		ENTITY QUERY
// ********************************************
    /// tells if a mesh entity exists
    virtual mEntity *find (mEntity *);
    /// a mesh has a container with all mesh entities
    mPartEntityContainer & getEntities () {return allEntities;}
    /// get a vertex with id iD, returns 0 if not found
    mVertex *getVertex	(int iD);
    /// get an edge with vertices v1 and v2, returns 0 if not found
    mEdge   *getEdge	(mVertex *v1, mVertex *v2);

    /// get an triangle with vertices v1, v2 and v3, returns 0 if not found
    mFace   *getTri	(mVertex *v1, mVertex *v2, mVertex *v3);
    /// get an quad with vertices v1, v2, v3 and v4, returns 0 if not found
    mFace   *getQuad	(mVertex *v1, mVertex *v2, mVertex *v3, mVertex *v4);
    /// get a tet   with vertices v1, v2, v3 and v4, returns 0 if not found
    mTet    *getTet	(mVertex *v1, mVertex *v2, mVertex *v3, mVertex *v4);
    /// get an hex with vertices v1, v2, ... and v8, returns 0 if not found
    mHex    *getHex	(mVertex *v1, mVertex *v2, mVertex *v3, mVertex *v4,
			 mVertex *v5, mVertex *v6, mVertex *v7, mVertex *v8);

// ********************************************
//		ENTITY CREATION
// ********************************************
    /// add a mesh entity
    virtual void add (mEntity *);
    mEntity* getEntity (mEntity::mType type, mVertex *v[], pGEntity g);

    /// create a new vertex
    mVertex *createVertex (int iD, const double x, const double y, const double z, 
    			pGEntity classif);

    mVertex *createVertex_noUpdateId(int iD, const double x, const double y, const double z, 
    			pGEntity classif);
			
    /// create a new vertex
    mVertex *createVertex (const double x, const double y, const double z , 
    			pGEntity classif);
    /// create a new edge
    mEdge   *createEdge (int iV1 , int iV2, pGEntity classif);			
   	
    /// create a new edge
    mEdge   *createEdge (mVertex *v1 , mVertex *v2, pGEntity classif);

    /// create a new face
    mFace   *createFaceWithVertices (int iV1 , int iV2, int iV3, pGEntity classif);
    /// create a new face
    mFace   *createFaceWithVertices (int iV1 , int iV2, int iV3, int iV4, 
    			pGEntity classif);
    /// create a new face
    mFace   *createFaceWithVertices (mVertex *v1 , mVertex *v2, mVertex *v3, 
    			pGEntity classif);
    /// create a new face
    mFace   *createFaceWithVertices (mVertex *v1 , mVertex *v2, mVertex *v3, mVertex *v4, 
    			pGEntity classif);
    /// create a new face
    mFace   *createFaceWithEdges (mEdge *e1 , mEdge *e2, mEdge *e3, 
    			pGEntity classif, int *dir = 0);
    /// create a new face
    mFace   *createFaceWithEdges (mEdge *e1 , mEdge *e2, mEdge *e3, mEdge *e4, 
    			pGEntity classif, int *dir = 0);

   /// create a new tet
    mTet    *createTetWithVertices (int iV1 , int iV2, int iV3, int iV4, 
    			pGEntity classif);
    /// create a new tet
    mTet    *createTetWithVertices (mVertex*v1, mVertex*v2, mVertex*v3, mVertex*v4,
			pGEntity classif);
    /// create a new tet
    mTet    *createTetWithFaces (mFace *f1 , mFace *f2, mFace *f3, mFace *f4, 
    			pGEntity classif);
    /// create a new hex
    mHex    *createHexWithVertices (int iV1 , int iV2, int iV3, int iV4, 
		    	int iV5 , int iV6, int iV7, int iV8, pGEntity classif);
    /// create a new hex
    mHex    *createHexWithVertices (mVertex*v1, mVertex*v2, mVertex*v3, mVertex*v4,
		    	mVertex*v5, mVertex*v6, mVertex*v7, mVertex*v8, pGEntity
			classif);
    /// create a new hex
    mHex    *createHexWithFaces (mFace *f1 , mFace *f2, mFace *f3, mFace *f4, mFace *f5, mFace *f6,
    			pGEntity classif);

    /// create a new prism
    mPrism    *createPrismWithVertices (int iV1 , int iV2, int iV3, int iV4, int iV5 , int iV6, 
			pGEntity classif);
    /// create a new prism
    mPrism    *createPrismWithVertices (mVertex*v1, mVertex*v2, mVertex*v3, mVertex*v4,
			mVertex*v5, mVertex*v6,	pGEntity classif);
    /// create a new hex
    mPrism    *createPrismWithFaces (mFace *f1 , mFace *f2, mFace *f3, mFace *f4, mFace *f5,
    			pGEntity classif);
				
    /// create a new pyramid
    mPyramid    *createPyramidWithVertices (int iV1 , int iV2, int iV3, int iV4, int iV5,
			pGEntity classif);
    /// create a new pyramid
    mPyramid    *createPyramidWithVertices (mVertex*v1, mVertex*v2, mVertex*v3, mVertex*v4,
			mVertex*v5, pGEntity classif);
    /// create a new pyramid
    mPyramid    *createPyramidWithFaces (mFace *f1 , mFace *f2, mFace *f3, mFace *f4, mFace *f5,
    			pGEntity classif);					
    
// ********************************************
//		ENTITY DELETION
// ********************************************
    /// delete from database but does not delete the pointer
    virtual void del (mEntity *);
    /// deletes form database then deletes the pointer 
    virtual void DEL (mEntity *);

// *******************************************
//             ENTITY TOPOLOGY INFO
// *******************************************
    unsigned long numTopo[12]; 
    void incTopo(int topo);
    void decTopo(int topo);
    int getNumTopo(int topo);
// ********************************************
//		MESH INFO/ITERATOR
// ********************************************
    /// gives the TOTAL size of entities of dimension what
    virtual int size(int what) const;
    int getRefinementLevel( mEntity *e ) {return 0;}  // added for backward compatibility
    /// iterator on all entities
    virtual iterall beginall(int what) const;
    /// end iterator on all entities
    virtual iterall endall(int what) const;
    /// iterator on leaves of the mesh (otctree leaves traversal)
    virtual iter begin(int what) const;
    /// end iterator on leaves of the mesh (otctree leaves traversal)
    virtual iter end  (int what) const;
    /// iterator on entities of dimension what classified on which
    virtual mClassIterator begin(int what, int which, int onwhat) const; 
    /// end iterator on entities of dimension what classified on which
    virtual mClassIterator end(int what, int which, int onwhat) const;

// ********************************************
//		COMPUTATION
// ********************************************
    SCOREC::Util::mPoint COG(mEntity*);
    SCOREC::Util::mPoint closestPoint (const SCOREC::Util::mPoint & pt , int dim , int tag ) const;

// ********************************************
//		GHOSTING
// ********************************************
    // Functions used in ghost creation/deletion
    void addGhost(mGhost* ghost);
    int getGhostVecSize();
    mGhost* getGhost(int index);
    std::vector<mGhost*> getGhostVec();
    void clearGhostVec();

// ********************************************
//		DEBUG - mDebugUtil.cc
// ********************************************
    int numUniqueEntities(int);
    void printPSet();
    void printAll();
    void printNumEntities();

// ********************************************
//		DMUM-related
// ********************************************
#ifdef DMUM
    bool DMUM_on;
    int count_trvs[4];
#endif

// ********************************************
//		FLEXIBLE M-REP 
// ********************************************

  short int MRM[4][4];

#ifdef FLEXDB
  // Mesh Representation Matrix
  // type definitions
  typedef mEdge* (*t_createE_FP)(mPart*, mVertex*, mVertex*,pGEntity);

  typedef mFace* (*t_createTri_E_FP)(mPart*, mEdge*, mEdge*,mEdge*,pGEntity,int*);
  typedef mFace* (*t_createQuad_E_FP)(mPart*, mEdge*, mEdge*,mEdge*,mEdge*,pGEntity,int*);

  typedef mTet* (*t_createTet_F_FP)(mPart*,mFace*,mFace*,mFace*,mFace*,pGEntity);
  typedef mHex* (*t_createHex_F_FP)(mPart*,mFace*,mFace*,mFace*,mFace*,mFace*,mFace*,pGEntity);
  typedef mPrism* (*t_createPrism_F_FP)(mPart*,mFace*,mFace*,mFace*,mFace*,mFace*,pGEntity);
  typedef mPyramid* (*t_createPyramid_F_FP)(mPart*,mFace*,mFace*,mFace*,mFace*,mFace*,pGEntity);
  
  typedef mFace* (*t_createTri_V_FP)(mPart*, mVertex*, mVertex*,mVertex*,pGEntity,int*);
  typedef mFace* (*t_createQuad_V_FP)(mPart*, mVertex*, mVertex*,mVertex*,mVertex*,pGEntity,int*);

  typedef mTet* (*t_createTet_V_FP)(mPart*,mVertex*,mVertex*,mVertex*,mVertex*,pGEntity);
  typedef mHex* (*t_createHex_V_FP)(mPart*,mVertex*,mVertex*,mVertex*,mVertex*
                                         ,mVertex*,mVertex*,mVertex*,mVertex*,pGEntity);
  typedef mPrism* (*t_createPrism_V_FP)(mPart*,mVertex*,mVertex*,mVertex*,mVertex*
                                         ,mVertex*,mVertex*,pGEntity);
  typedef mPyramid* (*t_createPyramid_V_FP)(mPart*,mVertex*,mVertex*,mVertex*,mVertex*
                                         ,mVertex*,pGEntity);
    
  typedef void (*t_deleteR_FP)(mPart*,mEntity*);
  typedef void (*t_deleteF_FP)(mPart*,mEntity*);
  typedef void (*t_deleteE_FP)(mPart*,mEntity*);

// mesh mod functions    
  t_createE_FP createE_FP;

  t_createTri_E_FP createTri_E_FP;
  t_createQuad_E_FP createQuad_E_FP;
  t_createTet_F_FP createTet_F_FP;
  t_createHex_F_FP createHex_F_FP;
  t_createPrism_F_FP createPrism_F_FP;
  t_createPyramid_F_FP createPyramid_F_FP;
    
  t_createTri_V_FP createTri_V_FP;
  t_createQuad_V_FP createQuad_V_FP;
  t_createTet_V_FP createTet_V_FP;
  t_createHex_V_FP createHex_V_FP;
  t_createPrism_V_FP createPrism_V_FP;
  t_createPyramid_V_FP createPyramid_V_FP;
  
  t_deleteR_FP deleteR_FP;
  t_deleteF_FP deleteF_FP;
  t_deleteE_FP deleteE_FP;

  void setMeshModFunctors(bool oneLevel=false);
#endif  
};

class mPartLessThanKey
{
  public :
    bool operator() (mPart* mesh1, mPart* mesh2) const
    {
      return mesh1->getId() < mesh2->getId();
    }
};
#endif 
