/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#ifdef FMDB_PARALLEL
#include <iomanip>
#include <fstream>
#include "mMeshIO.h"
#include "FMDB.h"

struct msg_struct
{
  mEntity* ent;
  int index;
  int owner_partid; 
  msg_struct(mEntity* e, int i, int j): ent(e), index(i),owner_partid(j){}
};

//********************************************************
int import_serial_and_partition_SMS(const char *filename, mMesh *mesh,
				    pmMigrationCallbacks &lb)
//********************************************************				    
{
  int err = FMDB_Util::Instance()->import_mesh(mesh, filename);
  if (err) return err;

  if (ParUtil::Instance()->size()>1) 
  { 
    ParUtil::Instance()->Barrier(__LINE__,__FILE__);  
    pPart part;
    FMDB_Mesh_GetPart (mesh, 0, part);
    pmLoadBalance(part, lb);
#ifdef PERIODIC2D
    MPBC_update();
#endif 
  }
  return SCUtil_SUCCESS; 
}

//********************************************************
int import_parallel_singlepart_SMS (mMesh* mesh, const char* fName)
//********************************************************
{
  char line[256];
  int i;
  int NbRegions=0,NbFaces=0,NbEdges,NbVertices,NbPoints;
  int GEntityType,GEntityId,EntityNbConnections;
  int Dummy;
  int Version=3;                // default value 
  int VertexId1,VertexId2;
  int Edge1;
  int Face1;
  int nbPts,NbEdgesOnFace,NbFacesOnRegion;
  double x,y,z;
  double u,v;
  int patch;
  int pclassid, ucbid, cbpid;
  pmEntity* pe;
  mEntity* ent;

  char inmesh[256]; 
  char without_extension[256];
        
  map<int,mEntity*> CBEntities;                            // entity and its local unique id 
  list<msg_struct> entitiesToBounce; 

  // test mesh file availability
  snprintf(without_extension,strlen(fName)-3,"%s",fName);
  sprintf(inmesh,"%s0.sms",without_extension); 
  FILE* test_in = fopen (inmesh,"r");
  if(!test_in)
  {
    cout<<"["<<SCUTIL_CommRank()<<"] FMDB ERROR: unable to open file \""<<inmesh<<"\"\n";
    return SCUtil_FILE_NOT_FOUND;  
  }
  fclose (test_in);
    
  int maxNumPart=ParUtil::Instance()->getTgtNumParts();
  pPart part;

  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)  
  {
    part = *part_it;
    int partid_offset = SCUTIL_CommGrp()*SCUTIL_GrpSize()*maxNumPart;
    int mypartid=FMDB_Part_GlobID(part)-partid_offset;
    sprintf(inmesh,"%s%d.sms",without_extension,mypartid);   
    FILE* in = fopen (inmesh,"r");

    if (in)
    { 

// *********************************
// STEP 1: setup format
// *********************************
    fscanf(in,"%s %d",line,&Version);
    typedef GEntity* (*entityBy_FP)(SGModel*,int,int);
    entityBy_FP fp = GM_entityByTag; 

// *********************************
// STEP 2: check number of partitions
// *********************************
    int pid, total, numPEs, totalCBEntities;                         // local totalCBEntities 
    fscanf(in,"%d %d %d %d",&total, &pid, &numPEs, &totalCBEntities);
    if (total>SCUTIL_GrpSize()*maxNumPart || pid+partid_offset != FMDB_Part_GlobID(part))
    {
#ifdef DEBUG
      cout<<"["<<SCUTIL_CommRank()<<"] total #parts="<<total
          <<", SCUTIL_GrpSize()*maxNumPart="<<SCUTIL_GrpSize()*maxNumPart
          <<", fileid+partid_offset="<<pid+partid_offset
          <<", local partid="<< FMDB_Part_GlobID(part)<<endl;
#endif
      return SCUtil_FILE_NOT_FOUND;
    }

// *********************************
// STEP 3: setup partition model
// *********************************
    int peid, pedim, peowner, penumBPs;
    pmModel* pmodel=pmModel::Instance();
    pmEntity* newPE;
  
    for (i=0; i<numPEs;++i)
    {
      fscanf(in,"%d %d %d %d",&peid, &pedim, &peowner, &penumBPs);
      set<int> bps;
    
      int bp;
      for (int j=0; j<penumBPs;++j)
      {  
        fscanf(in,"%d",&bp);
        bps.insert(bp+partid_offset);
      }
      newPE = new pmEntity(peid, bps, pedim);   
      pmodel->addPEntity(newPE); 
      newPE->setOwner(peowner+partid_offset);
    } 

// *********************************
// STEP 4: read # entities per type
// *********************************
  
    fscanf(in,"%d %d %d %d %d",&NbRegions,&NbFaces,&NbEdges,
                               &NbVertices,&NbPoints);

// *********************************
// STEP 5: read vertices
// *********************************
    std::vector<mVertex*> vertices;
    mVertex* vv;        
    int vid;
    for(i=0;i<NbVertices;i++)
    {
      fscanf(in,"%d",&vid);       
      fscanf(in,"%d",&GEntityId);       
      if (!GEntityId) continue;

      fscanf(in,"%d %d %lf %lf %lf",&GEntityType,&EntityNbConnections,
  	                                  &x,&y,&z); 
      vv = part->createVertex(vid,x,y,z,
                part->getGEntity(GEntityId,GEntityType,fp)); 
      vertices.push_back(vv);
	    
      switch(GEntityType)
      {
        case 0:
            break;
        case 1:
            fscanf(in,"%le",&u);
            vv->attachVector(FMDB_Util::Instance()->getParametric(),SCOREC::Util::mVector (u,0,0) );
            break;
        case 2:
	    fscanf(in,"%le %le %d",&u,&v,&patch);		
	    vv->attachVector(FMDB_Util::Instance()->getParametric(),SCOREC::Util::mVector (u,v,patch) );
   	    break;
	case 3:
	    break;
 	default :
  	    if (!in)
	      return SCUtil_TAG_NOT_FOUND;	      
      } // switch(GEntityType)

      if(Version==4) {
  	fscanf(in,"%d %d %d", &pclassid, &ucbid, &cbpid);
	if (pclassid!=-1)
	{
       	  pe = pmodel->getPartitionEntity(pclassid);
          assert(pe!=0);
	  vv->setPClassification(pe);
          if(cbpid==mypartid)                                // the owner 
	    CBEntities[ucbid]=(mEntity*)vv;
	  else
	    entitiesToBounce.push_back(msg_struct((mEntity*)vv, ucbid, cbpid+partid_offset));
	}
      } // if(Version==4)
      
      if(Version==3) {
        fscanf(in,"%d %d", &pclassid, &ucbid);
        if (pclassid!=-1)
        {
          pe = pmodel->getPartitionEntity(pclassid);
          assert(pe!=0);
          vv->setPClassification(pe);
          CBEntities[ucbid]=(mEntity*)vv;
        }
      } // if(Version==3)
    } // for(i=0;i<NbVertices;i++)
   
// *********************************
// STEP 6: read edges
// *********************************
    std::vector<mEntity*> edges;
    mVertex* v1;
    mVertex* v2;
    mEntity* e;
    
    for(i=0;i<NbEdges;i++)
    {
      fscanf(in,"%d",&GEntityId);
      if (!GEntityId) continue;
      
      fscanf(in,"%d %d %d %d %d",&GEntityType, &VertexId1,&VertexId2,
	             &EntityNbConnections,&nbPts); 
      v1 = vertices[VertexId1-1];
      v2 = vertices[VertexId2-1];
      e = M_createE(part, v1, v2, part->getGEntity(GEntityId,GEntityType,fp));
      edges.push_back(e);
 
      for(int j=0;j<nbPts;j++)
      {
        fscanf(in,"%lf %lf %lf ", &x, &y, &z);
        pPoint pt = P_new();
        P_setPos(pt, x, y, z);
        switch(GEntityType)
        {
          case 1: {fscanf(in,"%le",&u);FMDB_P_setParametricPos(pt, u, 0, 0);}
               break;
          case 2: {fscanf(in,"%le %le %d",&u,&v,&patch); FMDB_P_setParametricPos(pt, u, v, 0);}
               break;
          default: break;
        } // switch(GEntityType)
        E_setPoint(e, pt);
      } // for(int j=0;j<nbPts;j++)

      if(Version==4) 
      {
        fscanf(in,"%d %d %d", &pclassid, &ucbid, &cbpid);
        if (pclassid!=-1)
        {
          pe = pmodel->getPartitionEntity(pclassid);
   	  assert(pe!=0);
    	  e->setPClassification(pe);
	  if(cbpid==mypartid)                                // the owner 
	    CBEntities[ucbid]=e;
 	  else
	    entitiesToBounce.push_back(msg_struct(e, ucbid, cbpid+partid_offset));
        }
      } //  if(Version==4) 

      if(Version==3) 
      {
        fscanf(in,"%d %d", &pclassid, &ucbid);
        if (pclassid!=-1)
        {
          pe = pmodel->getPartitionEntity(pclassid);
          assert(pe!=0);
          e->setPClassification(pe);
          CBEntities[ucbid]=e;
        }
      }
    }  // for(i=0;i<NbEdges;i++)

// *********************************
// STEP 7: read faces
// *********************************

    std::vector<mEntity*> faces;
    mEntity* f;
    mEntity* eg[4];
  
    for(i=0;i<NbFaces;i++)
    {
      fscanf(in,"%d",&GEntityId);
      if(!GEntityId) continue;

      fscanf(in,"%d %d",&GEntityType, &NbEdgesOnFace);
      for(int j=0; j<NbEdgesOnFace; j++)
      {
        fscanf(in, "%d", &Edge1);
        eg[j]=edges[abs(Edge1)-1]; 
      }

      fscanf(in, "%d", &nbPts);
  
      f = M_createF(part, NbEdgesOnFace, eg, 0, part->getGEntity(GEntityId,GEntityType,fp)); 
      
      faces.push_back(f);

      for(int j=0;j<nbPts;j++)
      {
	switch(GEntityType)
	{
	  case 0: break;
	  case 1: fscanf(in,"%le",&u);
		 break;
	  case 2: fscanf(in,"%le %le %d",&u,&v,&patch);
		 break;
	  case 3: break;
	  default: cout<<__LINE__<<","<<__FILE__<<": unknown tag in sms loadSmsFile";
                   return SCUtil_TAG_NOT_FOUND;
	} // switch(GEntityType)
      } // for(int j=0;j<nbPts;j++)

      if(Version==4) 
      {      
        fscanf(in,"%d %d %d", &pclassid, &ucbid, &cbpid);
        if (pclassid!=-1)
        {
          pe = pmodel->getPartitionEntity(pclassid);
	  assert(pe!=0);
	  f->setPClassification(pe);
	
	  if(cbpid==mypartid)                                // the owner 
	    CBEntities[ucbid]=f;
  	  else
	    entitiesToBounce.push_back(msg_struct(f, ucbid, cbpid+partid_offset));
        } // if (pclassid!=-1)
      } // if(Version==4) 

      if(Version==3) 
      {
        fscanf(in,"%d %d", &pclassid, &ucbid);
        if (pclassid!=-1)
        {
          pe = pmodel->getPartitionEntity(pclassid);
          assert(pe!=0);
          f->setPClassification(pe);
          CBEntities[ucbid]=f;
        } // if (pclassid!=-1)
      } // if(Version==3) 
    } // for(i=0;i<NbFaces;i++)
  
// *********************************
// STEP 8: read regions
// *********************************

    std::vector<mEntity*> regions;
    mEntity* fc[6];
    mEntity* r;
    for(i=0;i<NbRegions;i++)
    {
      fscanf(in,"%d",&GEntityId);
      if(!GEntityId) continue;
      fscanf(in,"%d",&NbFacesOnRegion);
     
      for(int j=0; j<NbFacesOnRegion; j++)
      {
        fscanf(in,"%d", &Face1);
        fc[j]=faces[abs(Face1)-1];
      }
      fscanf(in,"%d", &Dummy);
      
      r=M_createR(part,NbFacesOnRegion,fc,0,part->getGEntity(GEntityId,3,fp));
      regions.push_back(r);
    } // for(i=0;i<NbRegions;i++)
 
// ************************************
//    STEP 9: Read Entity Sets in Part Level
// ************************************
    int NbEntSets=0, MATCHING=0;
    pEntSet eset;
    if (fscanf(in, "%s", line)!= EOF)
      MATCHING=1;
    if (MATCHING==1 && !strcmp(line,"PartSet"))
    {
      fscanf(in, "%d", &NbEntSets);
      int isList, numEnts, entType, entId;
 
      int eset_cnt=0;
      for (; eset_cnt<NbEntSets; ++eset_cnt)
      {
        fscanf(in, "%d %d", &isList, &numEnts);
        if (!numEnts) continue;
        eset = part->createPSet(part->getDimension() );
        eset->attachInt(FMDB_Util::Instance()->getId(), eset_cnt); // attach ID starting from NbMeshSets
        for (int ent_cnt=0; ent_cnt<numEnts; ++ent_cnt)
        {
          fscanf(in, "%d %d", &entType, &entId);
          switch(entType)
          {
            case 0: ((mEntityGroup*)eset)->addEntity(vertices[entId-1]);
                  break;
            case 1: ((mEntityGroup*)eset)->addEntity(edges[entId-1]);
                  break;
            case 2: ((mEntityGroup*)eset)->addEntity(faces[entId-1]);
                  break;
            case 3: ((mEntityGroup*)eset)->addEntity(regions[entId-1]);
                  break;
            default: break;
          }  // switch (entType)
        } // for ent_cnt
      } // for (; eset_cnt<NbEntSets; ++eset_cnt)

      if (fscanf(in, "%s", line) == EOF)
        MATCHING=0;
    }  // if (MATCHING==1 && strcmp(line,"PartSet")) 
    fclose (in);
  } // else


// *********************************
// STEP 9: setup part boundary
// *********************************

  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::All_Reduce);
  int msg_size = sizeof(mEntity*)+sizeof(int)*2; 
  CM->set_fixed_msg_size(msg_size);
  void* msg_send = CM->alloc_msg(msg_size);
  int* s_id; 
  mEntity** s_ent; 

  CM->set_tag(0);

  if (Version==4) 
  {
    list<rc_struct_2> entitiesToBroadcast;

    for(list<msg_struct>::iterator lit=entitiesToBounce.begin(); lit!=entitiesToBounce.end(); lit++) 
    {
      s_ent = (mEntity**)msg_send; 
      *s_ent = (*lit).ent; 
      s_id =  (int*)((char*)msg_send + sizeof(mEntity*)); 
      s_id[0] = (*lit).index;
      s_id[1] = mypartid+partid_offset;
      CM->send((*lit).owner_partid/maxNumPart, (void*)msg_send);
    } // for(list<rc_struct_4>::iterator lit

    CM->finalize_send();
    CM->free_msg(msg_send);

    void* msg_recv; 
    int pid_from; 

    // receive phase
    while(CM->receive(msg_recv, &pid_from))
    {
      mEntity* ent = *(mEntity**)msg_recv;
      s_id = (int*)((char*)msg_recv + sizeof(mEntity*));
      CBEntities[s_id[0]]->addRemoteCopy(s_id[1], ent);
      entitiesToBroadcast.push_back(rc_struct_2(CBEntities[*s_id], pid_from, ent));
      CM->free_msg(msg_recv);
    } //  while(int rc = CM->receive(msg_recv, &pid_from))
  
    for(map<int, mEntity*>::iterator mapit=CBEntities.begin(); mapit!=CBEntities.end(); mapit++) 
      entitiesToBroadcast.push_back(rc_struct_2(mapit->second, mypartid+partid_offset, mapit->second)); 

    exchangeEntitiesToBroadcast(entitiesToBroadcast);
  } // if version==4

  if(Version==3) 
  {
    for(map<int, mEntity*>::iterator CBEntitiesIter=CBEntities.begin();
                       CBEntitiesIter!=CBEntities.end();CBEntitiesIter++)
    {
      int idx =  CBEntitiesIter->first;
      ent = CBEntitiesIter->second;
      if (!ent) continue;
      pe = ent->getPClassification();
      for (pmEntity::BPIter bpiter=pe->bpBegin(); bpiter!=pe->bpEnd();++bpiter)
      {
	if (*bpiter==mypartid+partid_offset) continue;
        s_ent = (mEntity**)msg_send;
        *s_ent = ent;
        s_id = (int*)((char*)msg_send + sizeof(mEntity*));
        s_id[0] = idx;
        s_id[1] = mypartid + partid_offset;
        CM->send((*bpiter)/maxNumPart, (void*)msg_send);
      } // for pmEntity::BPIter 
    } // for map<int, mEntity*>::iterator
    CM->finalize_send();
    CM->free_msg(msg_send);

    // receive phase begins
    void *msg_recv;
    int pid_from;

    while(CM->receive(msg_recv, &pid_from))
    {
      s_ent = (mEntity**)msg_recv;
      s_id = (int*)((char*)msg_recv + sizeof(mEntity*));
      CBEntities[s_id[0]]->addRemoteCopy(s_id[1], *s_ent);
      CM->free_msg(msg_recv);
    } // while(int rc = CM->receive(msg_recv, &pid_from))
  } // if version==3
} // for each part

// *********************************
// STEP 10 : SET mesh dimension
// ********************************* 
  int local_dim = (mesh->getPart(0)->size(3)? 3: 2), global_dim=0;
  MPI_Allreduce(&local_dim, &global_dim, 1, MPI_INT, MPI_MAX, ParUtil::Instance()->getComm());
  ParUtil::Instance()->set_maxMeshDim(global_dim);

// ******************************************************************
// STEP 11: remove pmodel entity with NO classified mesh entities 
//          and update neighbor proc info
// ******************************************************************
  vector<pmEntity*> pmEntityToRemove;
  for(pmModel::PEIter peit=pmModel::Instance()->peBegin(); peit!=pmModel::Instance()->peEnd(); ++peit)
    if ((*peit)->getNumRClasEnt()==0)
      pmEntityToRemove.push_back(*peit);

  for (vector<pmEntity*>::iterator peit=pmEntityToRemove.begin(); peit!=pmEntityToRemove.end();++peit)
    pmModel::Instance()->delPEntity(*peit);

  vector<int> NborProc;
  FMDB_Mesh_GetNborProcID (mesh, NborProc);

  set<int>SetNborProc;
  std::copy(NborProc.begin(), NborProc.end(), inserter(SetNborProc, SetNborProc.end()));
  ParUtil::Instance()->ComMan()->set_Nbrs(SetNborProc);

  return SCUtil_SUCCESS; 
}
#endif
