/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "ParUtil.h"
#include "FMDB.h"
#ifdef FMDB_PARALLEL
#else
#ifndef MSC
#include <sys/time.h>
#endif
#endif

std::auto_ptr<ParUtil> ParUtil::instance;
int ParUtil::buffer_size = 1024;

ParUtil* ParUtil::Instance()
{
  if(!instance.get())
    {
      instance.reset(new ParUtil);
      instance->setVertbosityLevel(0);
    }
  return instance.get();
}

ParUtil::~ParUtil() 
{}

ParUtil::ParUtil() 
{
#ifdef FMDB_PARALLEL
  initialized = 0;
  finalized = 0; 
#endif
}

void ParUtil::init(MPI_Comm communicator) 
{
#ifdef FMDB_PARALLEL
  if(initialized) 
    return;

  int init_flag=0;
  MPI_Initialized(&init_flag);
  if(!init_flag)
    MPI_Init(0,0);

  MPI_Comm_dup(communicator, &mycomm);
  MPI_Comm_rank(mycomm, &myrank);
  MPI_Comm_size(mycomm, &mysize);

  MPI_Errhandler_set(mycomm, MPI_ERRORS_RETURN);
  CM = new IPComMan(mycomm, 0, 10240, IPComMan::Neighbors);
  vl = 1;
  resetBarrierSensor();
  initialized = 1;    
  
  curNumParts = 1;           // current_local_num_parts_per_process
  tgtNumParts = 1;           // target_local_num_parts_per_process
  usrNumParts = 1;
  split_key=0;  // group ID 
#endif
}

double ParUtil::wTime() const
{

#ifdef FMDB_PARALLEL
  return MPI_Wtime(); 
#else
  struct timeval tp;
  struct timezone tzp;
  double timeval;
  
  gettimeofday(&tp,&tzp);
  
  timeval = (double) tp.tv_sec;
  timeval = timeval + (double) ((double) .000001 * (double) tp.tv_usec);
  
  return(timeval);
#endif
}

void ParUtil::Msg(ParUtil::MessageLevel level, const char *fmt, ...)
{ 
  char buff[1024];
  va_list  args;
  va_start (args, fmt);
  vsprintf(buff, fmt, args);
  va_end (args);

  switch(level)
  {
    case INFO:
      if(vl >= 0 && master())
	  fprintf(stdout,"%s",buff);
      break;
    case WARNING:
      fprintf(stdout,"Proc %d FMDB WARNING : %s",rank(),buff);
      fflush(stdout);
      break;
    case ERROR:
      fprintf(stdout,"FMDB FATAL ERROR : %s",buff);
      fflush(stdout);
      break;
    default: fprintf(stdout,"[P%d]: %s",rank(),buff);
      fflush(stdout);
      break;
  }
}

void ParUtil::Barrier(int line, const char *fn)
{
#ifdef FMDB_PARALLEL
    double t1 = wTime();
    MPI_Barrier(mycomm);
    timeSpentOnBarriers += (wTime()-t1);
#endif
}

void ParUtil::Finalize(bool do_mpi_finalize)
{
#ifdef FMDB_PARALLEL
   if(initialized && !finalized) 
   {
     delete CM;
     if (do_mpi_finalize)
       MPI_Finalize();
     finalized = 1;
   }
#endif
}
int ParUtil::splitComm (int numProc)
{
#ifdef FMDB_PARALLEL
  int world_size;
  MPI_Comm_size(MPI_COMM_WORLD, &world_size);
  if (world_size%numProc!=0)
  {
    cout<<__func__<<" failed: #processes % #procPerGroup != 0";
    return 1; 
  }
  MPI_Comm my_comm;
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  int my_row = world_rank/numProc;
  split_key=my_row;
  MPI_Comm_split(MPI_COMM_WORLD, my_row, world_rank, &my_comm);
  setComm(my_comm); 
#endif
  return 0;
}

void ParUtil::setComm(MPI_Comm communicator) 
{ 
#ifdef FMDB_PARALLEL
  MPI_Comm_rank(communicator, &myrank); // recompute myrank
  MPI_Comm_size(communicator, &mysize);  // recompute mysize
  MPI_Comm_dup(communicator, &mycomm); // set ParUtil's communicator

// Note: potential bug with Zoltan/Partitining ito Non-COMM_WORKD communicator
// without MPI_COMM_WORLD, Zoltan partitioner crashes in initial partitioning
// in serial mesh loading: do CM->set_comm after zoltan interaction
// in distributed mesh loading: do CM->set_comm in the beginning
//  CM->set_comm(communicator);  // set IPComMan's communicator
#endif
}

double ParUtil::getTimeDiff()
{
  struct rusage ruse_now;
  getrusage(RUSAGE_SELF, &ruse_now);
  struct timeval res_time;
  res_time.tv_sec = ruse_now.ru_utime.tv_sec - ruse_start.ru_utime.tv_sec;
  res_time.tv_usec = ruse_now.ru_utime.tv_usec - ruse_start.ru_utime.tv_usec;
  return double(res_time.tv_sec) + double(res_time.tv_usec)/1000000.0;
}

double ParUtil::getCurTime()
{
  struct rusage ruse_now;
  getrusage(RUSAGE_SELF, &ruse_now);
  return double(ruse_now.ru_utime.tv_sec) + double(ruse_now.ru_utime.tv_usec)/1000000.0;
}

double ParUtil::getCurMem()
{
  struct mallinfo meminfo_now = mallinfo();
  return ((double)meminfo_now.arena)/1048576.0;
}

double ParUtil::getMemDiff()
{
  struct mallinfo meminfo_now = mallinfo();
  return ((double)(meminfo_now.arena-meminfo_start.arena))/1048576.0;
}

void ParUtil::sysinfo()
{
  if (!master()) return;
  struct utsname u;
  if (uname(&u) == 0)
    printf("\n%s %s %s %s %s\n",
           u.sysname, u.nodename, u.release, u.version, u.machine);
  fflush(stdout);
}

void ParUtil::report(char const* title, double time, double memory)
{
  if (master())
  {
    printf("[grp%d|rank%d] %-20s %6.3f sec %7.3f MB \n", SCUTIL_CommGrp(), SCUTIL_CommRank(), title, time, memory);
    fflush(stdout);
  }
}

#ifdef FMDB_PARALLEL
void ParUtil::setCurNumParts(int num) 
{ 
#ifdef DEBUG_
  if (master())
    cout<<"* ParUtil->curNumParts is set to "<<num<<"\n";
#endif
  curNumParts = num; 
}

void ParUtil::setTgtNumParts(int num) 
{
#ifdef DEBUG_
  if (master())  
    cout<<"* ParUtil->tgtNumParts is set to "<<num<<"\n";  
#endif 
  tgtNumParts = num; 
}

void ParUtil::setUsrNumParts(int num) 
{
#ifdef DEBUG_
  if (master())
    cout<<"* ParUtil->usrNumParts is set to "<<num<<"\n";
#endif

  usrNumParts = num; 
}
#endif


