/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/ 
#ifdef FMDB_PARALLEL
#include <set>
#include "pmUtility.h"
#include "FMDB.h"

void printPidList (std::list<int>& poor_to_rich_pids)
{
  cout<<"["<<SCUTIL_CommRank()<<"] Part ID in Poor To Rich Order: ";
  for (std::list<int>::iterator partid_it=poor_to_rich_pids.begin();
                 partid_it!=poor_to_rich_pids.end();++partid_it)
    std::cout<<*partid_it<<" ";
  std::cout<<std::endl;
}

void broadcastAscendingPartID(int num_part, std::list<int>& poor_to_rich_pids)
{
  IPComMan *CM = ParUtil::Instance()->ComMan();
  CM->set_comm_validation(IPComMan::All_Reduce);
  CM->set_tag(0);

  int msg_size = sizeof(int)*num_part*SCUTIL_CommSize();
  CM->set_fixed_msg_size(msg_size);
  int* msg_send = (int*)CM->alloc_msg(msg_size);

  if (!SCUTIL_CommRank())  // master
  {
    int index=0;
    for (std::list<int>::iterator rank_it=poor_to_rich_pids.begin();rank_it!=poor_to_rich_pids.end();++rank_it)
      msg_send[index++]=*rank_it;  // part id in sorted order
    for (int dest_rank=1; dest_rank<SCUTIL_CommSize(); ++dest_rank)
      CM->send(dest_rank, (void*)msg_send);
  }
  CM->finalize_send();
  CM->free_msg(msg_send);

  // receive phase begins -- on non-master
  void* msg_recv;
  int rank_from;
  while (int rc=CM->receive(msg_recv, &rank_from))
  {
    for (int index = 0; index < num_part*SCUTIL_CommSize(); ++index)
      poor_to_rich_pids.push_back(((int*)msg_recv)[index]);
    CM->free_msg(msg_recv);
  }  // if(rc)
}

// ***********************************************************
// poor_to_rich_pids[0] is the part id with the fewest # ptn object entities
void getAscendingPartIDWithNumPtnObj(mMesh* mesh, std::list<int>& poor_to_rich_pids)
// ***********************************************************
{
  IPComMan *CM = ParUtil::Instance()->ComMan();
  int mesh_dim = ParUtil::Instance()->get_maxMeshDim();

  // one round communication to get #regions for all processors
  std::vector<std::pair<int,int> > numPtnObj_partId;

  CM->set_comm_validation(IPComMan::All_Reduce);
  CM->set_tag(0);

  int num_part = ParUtil::Instance()->getTgtNumParts();
  int msg_size = sizeof(int)*num_part;
  CM->set_fixed_msg_size(msg_size);
  int* msg_send = (int*)CM->alloc_msg(msg_size);

  for (int i=0; i<num_part; ++i)
  {
    if (SCUTIL_CommRank())
      msg_send[i]=mesh->getPart(i)->size(mesh_dim);  // # partition objects in local process
    else // master
      numPtnObj_partId.push_back(std::pair<int,int>(mesh->getPart(i)->size(mesh_dim), i));
  }

  if (SCUTIL_CommRank())
    CM->send(0/*master*/, (void*)msg_send);
  CM->finalize_send();
  CM->free_msg(msg_send);

  // receive phase begins
  void* msg_recv;
  int rank_from;

  while(int rc=CM->receive(msg_recv, &rank_from))
  {
    int* msg_int = (int*)((char*)msg_recv);
    for (int i=0; i<num_part; ++i)
      numPtnObj_partId.push_back(std::pair<int,int>(msg_int[i], rank_from*num_part+i));
    CM->free_msg(msg_recv);
  }  // if(rc)

  if (!SCUTIL_CommRank())
  {
    std::sort(numPtnObj_partId.begin(),numPtnObj_partId.end()); // valid for master only
    // store sorted ranks in output container
    for (std::vector<std::pair<int,int> >::iterator vecIt=numPtnObj_partId.begin();vecIt!=numPtnObj_partId.end();++vecIt)
      poor_to_rich_pids.push_back((*vecIt).second);
  }
  // do one more round of communications to push the master's result to non-master
  broadcastAscendingPartID(num_part, poor_to_rich_pids);
}

// ***********************************************************
void serial_getSortedPids_poor_to_rich(pMeshMdl mesh, list<int>& poor_to_rich_pids)
// ***********************************************************
{
  int dim = ParUtil::Instance()->get_maxMeshDim();

  vector<pair<int,int> > numPOEnt_pid;
  for (mMesh::partIter part_it=mesh->partBegin(); part_it!=mesh->partEnd(); ++part_it)
    numPOEnt_pid.push_back(pair<int,int>((*part_it)->size(dim), (*part_it)->getId()));
  std::sort(numPOEnt_pid.begin(), numPOEnt_pid.end());
  for (std::vector<pair<int,int> >::iterator vecIt=numPOEnt_pid.begin();vecIt!=numPOEnt_pid.end();++vecIt)
    poor_to_rich_pids.push_back((*vecIt).second);
}

// *****************************************
void set_subtract(std::set<int> A, std::set<int> B, std::set<int>& C)
// *****************************************
{
  std::set<int>::iterator aiter, biter;
  for (aiter=A.begin(); aiter!=A.end();++aiter)
  {
    biter = B.find(*aiter);
    if (biter==B.end()) C.insert(*aiter);
  }
} 
#endif
