/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include "mFMDB.h"
#include "mEntityGroup.h"
#include "mEntity.h"
#include "FMDBInternals.h"
#include <algorithm>
#include "ParUtil.h"
#include "FMDB.h"

mEntityGroup::mEntityGroup(int d)
{
#ifdef FMDB_PARALLEL
   bdryEnts=NULL; 
#endif
}

mEntityGroup::~mEntityGroup()
{   
  container.clear();
#ifdef FMDB_PARALLEL
  if(bdryEnts!=NULL)
      SAFE_DELETE(bdryEnts); 
#endif
}

void mEntityGroup::addEntity(mEntity* ent) 
{
   container.push_back(ent);
   EN_attachDataPtr(ent, FMDB_Util::Instance()->getPSet(), this); 
   return;  
}

void mEntityGroup::removeEntity(mEntity* ent)
{
   iter entit = std::find(container.begin(), container.end(), ent); 
   if(entit==container.end())
      return; 
      
   container.erase(entit);
   ent->deleteData(FMDB_Util::Instance()->getPSet());
   return; 
}

void mEntityGroup::addEntity(mEntity* ent, int pos)
{
  EN_attachDataPtr(ent, FMDB_Util::Instance()->getPSet(), this); 

  if (pos >= container.size())
     container.resize(pos+1);
  list<mEntity*>::iterator it_pos=container.begin();
  std::advance(it_pos, pos);
  std::list<mEntity*>::iterator ent_it = container.erase(it_pos); 
  container.insert(ent_it, ent); 
}

void mEntityGroup::clear()
{
// delete attached data on interior entities
  for (iter ent_it=container.begin(); ent_it!=container.end(); ++ent_it) 
    (*ent_it)->deleteData(FMDB_Util::Instance()->getPSet());
  container.clear(); 
}

mEntity* mEntityGroup::getEntity(int pos)
{
  list<mEntity*>::iterator it_pos=container.begin();
  std::advance(it_pos, pos);
  return *it_pos;   
}

int mEntityGroup::getEntityOrder(mEntity* ent)
{
  return std::distance(container.begin(), std::find(container.begin(), container.end(), ent));
}

void mEntityGroup::getNumEntity(int* numV, int* numE, int* numF, int* numR)
{
  int numVtx=0, numEdge=0, numFace=0, numRgn=0;
  for (iter ent_it=container.begin(); ent_it!=container.end(); ++ent_it) 
    switch ((*ent_it)->getLevel())
    {
      case 0: ++numVtx; break;
      case 1: ++numEdge; break;
      case 2: ++numFace; break;
      case 3: ++numRgn; break;
      default: break;
    }
  *numV = numVtx;
  *numE = numEdge;
  *numF = numFace;
  *numR = numRgn;
}

void mEntityGroup::print()
{
  if (container.size()==0)
    std::cout<<"["<<SCUTIL_CommRank()<<"] pset "<<FMDB_Set_ID((pEntSet)this)<<": empty\n";
  else
    for (iter ent_it = container.begin(); ent_it!=container.end(); ++ent_it)
      std::cout<<"["<<SCUTIL_CommRank()<<"] pset "<<FMDB_Set_ID((pEntSet)this)<<": " <<(*ent_it)->getUid()<<endl; 
}

#ifdef FMDB_PARALLEL
void mEntityGroup::computeBdryEntities() // called in FMDB_Mesh_GlobPtn
{
  if (bdryEnts==NULL)
    bdryEnts = new set<void*>; 
  if (container.size()==0) return;
  mEntity* down_ent;             
  void* tmp;       
  int dim=(*begin())->getLevel();
  for(iter ent_it=begin(); ent_it!=end(); ++ent_it) 
  {   
    for(int i=0; i<(*ent_it)->size(dim-1); ++i)
    {
      // get the downward adjacency
      down_ent = (*ent_it)->get(dim-1, i);

      // get the upward adjacency for down
      if (down_ent->getPClassification()) 
	bdryEnts->insert(down_ent); 
      else
      {	
        for(int j=0; j<down_ent->size(dim); ++j) 
        {
           if(!EN_getDataPtr(down_ent->get(dim, j), FMDB_Util::Instance()->getPSet(), &tmp))        // not in the entity-group  
             bdryEnts->insert(down_ent);  	   
	   else 
	     if(tmp!= (void*)this)                     // no overlap between entity-groups 
	       bdryEnts->insert(down_ent); 
         }      
      }
    } // for i
  } // iter ent_it
}
void mEntityGroup::removeBdryEntities() 
{
  if (bdryEnts==NULL) return;  
  bdryEnts->clear();
  SAFE_DELETE(bdryEnts); 
}; 
#endif
