/****************************************************************************** 

  (c) 2004-2012 Scientific Computation Research Center, 
      Rensselaer Polytechnic Institute. All rights reserved.
  
  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.
 
*******************************************************************************/
#include <vector>
#include "FMDB.h"
#include "mMesh.h"
#include "mPart.h"

mMesh::mMesh(pGModel geom) 
: TagHolder()
{
  parts = new vector<mPart*>;
  theGeomModel = geom;
#ifdef FMDB_PARALLEL
  LB_Method=FMDB_PARMETIS; 
  LB_Approach=PartKway;
  imbalance_tol=1.03; 
  debug_level=0;
#endif
}

mMesh::~mMesh() 
{
  clearEntSet(); 
  clearPart();
  delete parts;
}

void mMesh::addPart (mPart* part) 
{
  parts->push_back(part);
}

void mMesh::delPart (mPart* part)
{
  partIter pit = std::find(parts->begin(), parts->end(), part);
  if (pit != parts->end())
    parts->erase(pit); 
  delete part;
}
 
void mMesh::clearPart()
{
  for (partIter pit=partBegin(); pit!=partEnd(); ++pit)
    delete *pit;
  parts->clear();
}

mPart* mMesh::getPart(int ith)
{  
  if (parts->size()<=((size_t)ith))
    return NULL;
  else
    return (*parts)[ith];
}


void mMesh::removeEntSet(EntitySet<mEntity> * entSet)
{
   std::list<EntitySet<mEntity>* >::iterator loc = std::find(allEntSets.begin(), allEntSets.end(),entSet);
   allEntSets.erase(loc++);
   if (entSet->isList())
      delete static_cast<EntitySetUnordered<mEntity>*>(entSet);
   else
      delete static_cast<EntitySetOrdered<mEntity>*>(entSet);
}


void mMesh::clearEntSet()
{
  EntitySet<mEntity> * entSet; 

  std::list<EntitySet<mEntity>* >::iterator loc = allEntSets.begin(); 
  while (loc!= allEntSets.end())
  {
    entSet = *loc; 
    if (entSet->isList())
      delete static_cast<EntitySetUnordered<mEntity>*>(entSet);
    else
      delete static_cast<EntitySetOrdered<mEntity>*>(entSet);
    ++loc; 
  }
}

bool mMesh::findTagID (unsigned int tag_id)
{
  if(tag_att_map.find(tag_id)==tag_att_map.end()) 
    return false; // tag_id not found
  return true; 
}

int mMesh::getTagID (const char *tag_name, unsigned int* tag_id)
{
  std::map <char*, unsigned int>::const_iterator it = name_tag_map.begin();
  for (; it!=name_tag_map.end(); ++it)
  {
    if (!strncmp((*it).first, tag_name, strlen(tag_name)))  // tag found
    {
      *tag_id=(*it).second;
      return 0;  // success
    }
  }
  return 1;  // failure
}

int mMesh::nameTagID (unsigned int tag_id, char* name)
{
  std::map <char*, unsigned int>::const_iterator it = name_tag_map.begin();
  for (; it!=name_tag_map.end(); ++it)
    if (it->second==tag_id)
    {
      strcpy (name, it->first);
      return 0; // sucess
    }
  return 1; // tag_id not found
}

int mMesh::typeTagID (unsigned int tag_id, int* type)
{
  if(tag_att_map.find(tag_id)==tag_att_map.end()) // tag_id not found
    return 1; // failure
  *type=tag_att_map[tag_id].first;
  return 0; // sucess
}

int mMesh::sizeTagID (unsigned int tag_id, int* size)
{
  if(tag_att_map.find(tag_id)==tag_att_map.end()) // tag_id not found
    return 1; // failure
  *size=tag_att_map[tag_id].second;
  return 0; // sucess
}

int mMesh::createTagID (const char *name, int type, int size, unsigned int* tag_id)
{
  std::map <char*, unsigned int>::const_iterator it = name_tag_map.begin();
  for (; it!=name_tag_map.end(); ++it)
    if (!strncmp((*it).first, name, strlen(name)))
      return 1; // the tag exists  

  char* name_str = new char[strlen(name)+1];
  strcpy (name_str, name);
  unsigned int new_id = (unsigned int) name_tag_map.size()+1000; 
  name_tag_map[name_str] =new_id;
  tag_att_map[new_id] = std::pair<int, int>(type, size);
  *tag_id=new_id;
  return 0;
}

int mMesh::deleteTagID (unsigned int tag_id)
{
  std::map<unsigned int, std::pair<int, int> >::iterator it_att=tag_att_map.find(tag_id);
  if (it_att==tag_att_map.end()) // tag_id not found
    return 1; // failure

  std::map <char*, unsigned int>::iterator it_name = name_tag_map.begin();
  for (; it_name!=name_tag_map.end(); ++it_name)
    if (it_name->second==tag_id)
      name_tag_map.erase(it_name);  
  tag_att_map.erase(it_att);
  return 0; // sucess
}

void mMesh::getAllTagID (std::vector<unsigned int>& ids)
{
  std::map <char*, unsigned int>::const_iterator it = name_tag_map.begin();
  for (; it!=name_tag_map.end(); ++it)
    ids.push_back((*it).second);
}

#ifdef FMDB_PARALLEL
// ghosting history interrogation
void mMesh::addGhostInfo(int gType, int bType, int numLayer)
{
  vecGhostRule.push_back(gType);
  vecGhostRule.push_back(bType);
  vecGhostRule.push_back(numLayer);
}

void mMesh::clearGhostInfo()
{
   vecGhostRule.clear();
}

void mMesh::getGhostInfo(std::vector<int>& ghost_info)
{
  std::copy(vecGhostRule.begin(), vecGhostRule.end(), back_inserter(ghost_info));
}
#endif

