/******************************************************************************

  (c) 2004-2012 Scientific Computation Research Center,
      Rensselaer Polytechnic Institute. All rights reserved.

  The LICENSE-SCOREC file included with this distribution describes the terms
  of the SCOREC Non-Commercial License this program is distributed under.

*******************************************************************************/

#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <malloc.h>
#include "FMDB.h"
#include "SCUtil.h"
#include <unistd.h>
#include <sys/syscall.h>
#include <errno.h>

using std::cout;
using std::endl;
using std::vector;

void CHECK(const char *msg, int oK)
{     
  if (oK)
    fprintf(stdout,"%s OK \n",msg);
  else
    fprintf(stdout,"%s FAILED\n", msg);
  fflush(stdout);
}

int TEST_MESH(int argc, char *argv[], const char* mesh_file, const char* out_file);

void CLEAR_MESH (pMeshMdl mesh);
int TEST_PART_SETGET_TAG (pMeshMdl mesh, pPart part);
int TEST_PART_DEL_TAG (pMeshMdl mesh, pPart part);
int TEST_SET_SETGET_TAG (pMeshMdl mesh, pEntSet eset);
int TEST_SET_DEL_TAG (pMeshMdl mesh, pEntSet eset);
int TEST_ENT_SETGET_TAG (pMeshMdl mesh, pMeshEnt ent);
int TEST_ENT_DEL_TAG (pMeshMdl mesh, pMeshEnt ent);
int TEST_MESH_SET(pMeshMdl mesh, pPart part);
int TEST_PART_SET (pMeshMdl mesh, pPart part);

template <class T>
void TEST_TAG (pMeshMdl mesh, pTag tag, const char* in_name, int name_len, int in_type, int in_size)
{
  char tag_name[name_len];
  int tag_type, tag_size, tag_byte;
  // verifying byte tag info
  FMDB_Tag_GetName (tag, tag_name, name_len);
  FMDB_Tag_GetType (tag, &tag_type);
  FMDB_Tag_GetSize (tag, &tag_size);
  FMDB_Tag_GetByte (tag, &tag_byte);
  assert(!strncmp(tag_name, in_name, name_len));
  assert(tag_type == in_type);
  assert(tag_size == in_size);
  assert(((size_t)tag_byte)==sizeof(T)*tag_size);
  assert(!strncmp(tag_name, in_name, name_len)
          && tag_type == in_type && tag_size == in_size && ((size_t)tag_byte)==sizeof(T)*tag_size);
}

int main (int argc, char *argv[])
{
  char mesh_file[1024], out_file [1024];
  if (argc==1)
  {
    strcpy(mesh_file, "../meshes/test3d_1Kr.sms");
    strcpy(out_file, "out_temp");
    TEST_MESH(argc, argv, mesh_file, out_file);
  }
  else if (argc==2)
  { 
    strcpy(mesh_file, argv[1]);
    strcpy(out_file, "out_temp");
    TEST_MESH(argc, argv, mesh_file, out_file);
  }
  else
  {
    strcpy(mesh_file, argv[1]);
    strcpy(out_file, argv[2]);
    TEST_MESH(argc, argv, mesh_file, out_file);
  }

  return 0;
}

int TEST_MESH(int argc, char *argv[], const char* mesh_file, const char* out_file)
{
// variable declaration
  int err, iType, iTopo, partDim, numEnt, numType, numTopo;
  int forceDel=0, exist;
  size_t size_tag;
  int i, j, k, failed;
  int type_topo[] = {0, 1, 2, 5, 11};
  int num_type[] = {0, 0, 0, 0};
  pTag byte_tag, int_tag, dbl_tag, ent_tag,  entset_tag;
  pTag intarr_tag, dblarr_tag, entarr_tag,  entsetarr_tag;
  pTag cloneTag;
  vector<pTag> tags, tags_before;
  pPart part;
  pPart newPart;
  pMeshEnt ent;
  vector<pMeshEnt> edges;
  pPartEntIter pent_it;

// end of variable declaration
#ifndef FMDB_PARALLEL
  MPI_Comm MPI_COMM_WORLD=0;
#endif
/*************************************************************************
@  Initialization and system/resource info:
@
@  SCUTIL_Init, SCUTIL_DspSysInfo, SCUTIL_GetNumProc, SCUTIL_GetProcID
@  SCUTIL_GetRsrcDiff
@ *************************************************************************/
  SCUTIL_DspSysInfo();
  CHECK("\n* Starting FMDB service...", !SCUTIL_Init(MPI_COMM_WORLD));

  ParUtil::Instance()->report("Initial cost: ", 0, ParUtil::Instance()->getCurMem());

  int numProc=SCUTIL_CommSize(), procId=SCUTIL_CommRank();
  CHECK("\n* Checking Rank/Size Functions...", numProc+procId);

/*************************************************************************
@  MESH/PART FUNCTIONS:
@
@  FMDB_Mesh_Create, FMDB_Mesh_Del, FMDB_Mesh_LoadFromFile
@  FMDB_Mesh_WriteToFile, FMDB_Mesh_DspNumEnt, FMDB_Mesh_Verify
@  FMDB_Part_Create, FMDB_Part_Del, FMDB_Mesh_GetParts, 
@  FMDB_Mesh_GetPart, FMDB_Part_GetID, FMDB_Mesh_GetNumPart
@ *************************************************************************/

  pMeshMdl mesh; 
  CHECK ("\n* Creating FMDB Mesh Instance...", !FMDB_Mesh_Create (NULL, mesh));
  
  SCUTIL_ResetRsrc();

  err = FMDB_Mesh_LoadFromFile(mesh, mesh_file, 0);

  double load_time=0.0, load_mem=0.0;
  SCUTIL_GetRsrcDiff (&load_time, &load_mem);
  printf("\t(mesh loading)         %7.3f sec  %7.3f MB\n", load_time, load_mem);

  if(err != SCUtil_SUCCESS)
    {
      cout << "\n Unable to load file " << mesh_file << endl;
      return SCUtil_FILE_NOT_FOUND;
    }

  SCUTIL_ResetRsrc();

  CHECK ("\n* Displaying Mesh Info...", !FMDB_Mesh_DspNumEnt (mesh));
  vector<pPart> parts;
  FMDB_Mesh_GetAllPart (mesh, procId, parts);
  part = parts.at(0);
  FMDB_Mesh_GetPart(mesh, 0, newPart);
  FMDB_Mesh_GetNumPart(mesh, &i);
  CHECK ("* Checking Part Retrival...", parts.size()==1 && newPart==part && i==1);

/*************************************************************************
@  TAG FUNCTIONS:
@
@  FMDB_Mesh_CreateTag, FMDB_Tag_GetName, FMDB_Tag_GetType, FMDB_Tag_GetSize, 
@  FMDB_Tag_GetByte, FMDB_Mesh_FindTag, FMDB_Mesh_HasTag, FMDB_Mesh_GetTag
@ *************************************************************************/
  FMDB_Mesh_GetTag (mesh, tags_before);
  size_tag=tags_before.size();
  tags_before.clear();

  CHECK("\n* Creating Single Data Tags...", 1);
  FMDB_Mesh_CreateTag (mesh, "byte", SCUtil_BYTE, 7, byte_tag);
  FMDB_Mesh_CreateTag (mesh, "integer", SCUtil_INT, 1, int_tag);
  FMDB_Mesh_CreateTag (mesh, "double", SCUtil_DBL, 1, dbl_tag);
  FMDB_Mesh_CreateTag (mesh, "entity", SCUtil_ENT, 1, ent_tag);
  FMDB_Mesh_CreateTag (mesh, "entity set", SCUtil_SET, 1, entset_tag);

  CHECK("* Checking Aborting Duplicate Tags...",
       FMDB_Mesh_CreateTag (mesh, "byte", SCUtil_BYTE, 7, byte_tag) 
       && FMDB_Mesh_CreateTag (mesh, "integer", SCUtil_INT, 1, int_tag) 
       && FMDB_Mesh_CreateTag (mesh, "double", SCUtil_DBL, 1, dbl_tag) 
       && FMDB_Mesh_CreateTag (mesh, "entity", SCUtil_ENT, 1, ent_tag) 
       && FMDB_Mesh_CreateTag (mesh, "entity set", SCUtil_SET, 1, entset_tag));

  CHECK("\n* Creating Multiple Data (Array) Tags...", 1);
  FMDB_Mesh_CreateTag (mesh, "integer array", SCUtil_INT, 3, intarr_tag);
  FMDB_Mesh_CreateTag (mesh, "double array", SCUtil_DBL, 3, dblarr_tag);
  FMDB_Mesh_CreateTag (mesh, "entity array", SCUtil_ENT, 3, entarr_tag);
  FMDB_Mesh_CreateTag (mesh, "entity set array", SCUtil_SET, 3, entsetarr_tag);

  FMDB_Mesh_GetTag (mesh, tags_before);
  // In this case, no new tags are being created. Instead, old tags are being re-used from the file.
  if(tags_before.size() == size_tag)
   size_tag=0;

  // verifying tag info
  TEST_TAG<char>(mesh, byte_tag, "byte", 100, SCUtil_BYTE, 7);
  TEST_TAG<int>(mesh, int_tag, "integer", 100, SCUtil_INT, 1);
  TEST_TAG<double>(mesh, dbl_tag, "double", 100, SCUtil_DBL, 1);
  TEST_TAG<pMeshEnt>(mesh, ent_tag, "entity", 100, SCUtil_ENT, 1);
  TEST_TAG<pEntSet>(mesh, entset_tag, "entity set", 100, SCUtil_SET, 1);

  TEST_TAG<int>(mesh, intarr_tag, "integer array", 100, SCUtil_INT, 3);
  TEST_TAG<double>(mesh, dblarr_tag, "double array", 100, SCUtil_DBL, 3);
  TEST_TAG<pMeshEnt>(mesh, entarr_tag, "entity array", 100, SCUtil_ENT, 3);
  TEST_TAG<pEntSet>(mesh, entsetarr_tag, "entity set array", 100, SCUtil_SET, 3);

  FMDB_Mesh_HasTag(mesh, int_tag, &exist);
  CHECK ("\n* Checking Tag Existence...", exist);
  FMDB_Mesh_FindTag (mesh, "byte", cloneTag);
  CHECK ("* Checking Tag Handle Retrieval by Name..", cloneTag == byte_tag);
  FMDB_Mesh_GetTag (mesh, tags);
  CHECK ("* Checking All Tag Retrieval..", tags.size()==size_tag+9);

/*************************************************************************
@  MESH/MODEL INFO FUNCTIONS:
@
@  FMDB_Mesh_IsEmpty, FMDB_Mesh_GetGeomMdl, FMDB_Mesh_GetDim
@ *************************************************************************/
  FMDB_Mesh_IsEmpty(mesh, &i);
  CHECK ("* Checking Mesh Emptiness..", i==0);
  pGeomMdl geom;
  FMDB_Mesh_GetGeomMdl (mesh, geom);
  CHECK ("* Retrieving Geom Model...", geom==NULL);
  FMDB_Mesh_GetDim (mesh, &i);
  CHECK ("* Retrieving Mesh Dimension Information...", i > 0);    
  
/*************************************************************************
@  PART INFO FUNCTIONS:
@
@  FMDB_Part_GetDim, FMDB_Part_GetNumEnt, FMDB_Part_GetNumPartBdyEnt, 
@  
@ *************************************************************************/

  for (iType=FMDB_VERTEX; iType<FMDB_ALLTYPE; ++iType)
  {
    numType=0;
    FMDB_Part_GetNumEnt(part, iType, FMDB_ALLTOPO, &numType);
    num_type[iType] = numType;
    numTopo=0;
    for (iTopo=type_topo[iType]; iTopo<type_topo[iType+1]; ++iTopo)
    {
      numEnt=0;
      FMDB_Part_GetNumEnt(part, iType, iTopo, &numEnt);
      numTopo += numEnt;
    }
    assert (numType==numTopo);
  }
  FMDB_Part_GetDim (part, &partDim);
  CHECK ("* Testing Part Dimension Information...", num_type[partDim] > 0);

  FMDB_Part_GetNumEnt(part, FMDB_ALLTYPE,  FMDB_ALLTOPO, &numType);
  FMDB_Part_GetNumEnt(part, FMDB_ALLTYPE, FMDB_ALLTOPO, &numTopo);
  CHECK ("* Checking Part Ent Type/Topo Functions...", numType==numTopo);

/*************************************************************************
@  TAG FUNCTIONS:
@ *************************************************************************/
  SCUTIL_ResetRsrc();

  int nRepeat=1;
  if (argc==4)
    nRepeat=atoi(argv[3]);
  for (int i=0; i<nRepeat; ++i)
{
  // attach tag to part handle
  TEST_PART_SETGET_TAG(mesh, part);

  int iterEnd = FMDB_PartEntIter_Init(part, FMDB_ALLTYPE, FMDB_ALLTOPO, pent_it);

  while (!iterEnd)
  {
    iterEnd = FMDB_PartEntIter_GetNext(pent_it, ent);
    if(iterEnd) break; 
    
    TEST_ENT_SETGET_TAG(mesh, ent);
  }
  FMDB_PartEntIter_Del (pent_it);
  CHECK("* Checking Entity Set/Get Tag...", 1);
}
  double test_time=0.0, test_mem=0.0;
  SCUTIL_GetRsrcDiff(&test_time, &test_mem);

/*************************************************************************
@  ENTITY ADJ FUNCTIONS:
@
@  FMDB_Ent_GetNumAdj, FMDB_Ent_GetAdj, FMDB_Ent_Get2ndAdj
@ *************************************************************************/
  pPartEntIter entIter; 
  vector<pMeshEnt> adjEnts;
  for (i=0; i<4; ++i)
  {
    int iterEnd = FMDB_PartEntIter_Init (part, i,  FMDB_ALLTOPO, entIter);
    if (iterEnd) continue;
    failed=0;
    while (!iterEnd)
    {
      iterEnd = FMDB_PartEntIter_GetNext(entIter, ent);
      if(iterEnd) break; 
      for (j=0; j<=4; ++j)
      {
        if (i==j) continue;
        adjEnts.clear();
        failed += FMDB_Ent_GetAdj(ent, j, 1, adjEnts);
        FMDB_Ent_GetNumAdj (ent, j, &k);
        if (adjEnts.size()!=((size_t)k)) ++failed;
      }
    }
    FMDB_PartEntIter_Del (entIter);
    switch (i)
    {
      case 0: CHECK ("* Checking Vertex Adjacency...", !failed); break;
      case 1: CHECK ("* Checking Edge Adjacency...", !failed); break;
      case 2: CHECK ("* Checking Face Adjacency...", !failed); break;
      case 3: CHECK ("* Checking Region Adjacency...", !failed); break;
    }
  }  // for i
/*
  for (int i=0; i<1; ++i)
  {
    int iterEnd = FMDB_PartEntIter_Init (part, i, FMDB_ALLTOPO, entIter);
    failed=0;
    while (!iterEnd)
    {
      iterEnd = FMDB_PartEntIter_GetNext(entIter, ent);
      if(iterEnd) break; 
      for (int brgType=0; brgType<=4; ++brgType)
        for (int tgtType=0; tgtType<=4; ++tgtType)
        {
          if (brgType==tgtType && brgType!=4) continue;
          adjEnts.clear();
          failed += FMDB_Ent_Get2ndAdj(ent, brgType, tgtType, adjEnts);
        }
    } // while
    FMDB_PartEntIter_Del (entIter);
    switch (i)
    {
      case 0: CHECK ("* Checking Vertex 2nd Adjacency...", !failed); break;
      case 1: CHECK ("* Checking Edge 2nd Adjacency...", !failed); break;
      case 2: CHECK ("* Checking Face 2nd Adjacency...", !failed); break;
      case 3: CHECK ("* Checking Region 2nd Adjacency...", !failed); break;
    }
  }  // for i
*/

/*************************************************************************
@  Mesh SET FUNCTIONS IN MESH LEVEL
@ *************************************************************************/

//  TEST_MESH_SET (mesh, part); // create three mesh-level sets
//  TEST_PART_SET (mesh, part);

/*************************************************************************
@  SET TAG FUNCTIONS:
@ *************************************************************************/
  pMeshSetIter mset_it;
  pEntSet eset;
  int iterEnd = FMDB_MeshSetIter_Init (mesh, mset_it);
  while(!iterEnd)
  {
    iterEnd = FMDB_MeshSetIter_GetNext(mesh, mset_it, eset);
    if(iterEnd) break;
    TEST_SET_SETGET_TAG(mesh, eset);
  }

/*************************************************************************
@  SET/TAG FILE I/O
@ *************************************************************************/
  SCUTIL_ResetRsrc();
  double write_time=0.0, write_mem=0.0;
//  for (int i=0; i<10; ++i)
  {
  // write out mesh 
  CHECK("* Checking Tag File Output...", !FMDB_Mesh_WriteToFile (mesh, out_file, 0));
  SCUTIL_GetRsrcDiff(&write_time, &write_mem);

  printf("\t(mesh loading)         %7.3f sec  %7.3f MB\n", load_time, load_mem);
  printf("\t(testing)              %7.3f sec  %7.3f MB\n", test_time, test_mem);
  printf("\t(mesh writing * 10)    %7.3f sec  %7.3f MB\n", write_time, write_mem);
        
  // remove all mesh info and restore it from file
  CLEAR_MESH(mesh);
        
  FMDB_Mesh_LoadFromFile(mesh, out_file , 0);
  }
  char ext[6];
  strcpy(ext,out_file+(strlen(out_file)-3));
if (strcmp(ext,"cdf") && strcmp(ext, ".nc") && strcmp(ext,"vtk"))
{
    tags.clear();
    FMDB_Mesh_GetTag (mesh, tags);
    CHECK ("* Checking Restored Tag..", tags.size()==9+size_tag);
    
  //iterEnd = FMDB_PartEntIter_Reset(pent_it);
  iterEnd = FMDB_PartEntIter_Init (part, i, FMDB_ALLTOPO, pent_it);
  while (!iterEnd)
  {
    iterEnd = FMDB_PartEntIter_GetNext(pent_it, ent);
    if(iterEnd) break;
    
    tags.clear();
    FMDB_Ent_GetTag (ent, tags);
    assert(tags.size()>=7 && tags.size()<=7+size_tag);
  }
  FMDB_PartEntIter_Del (pent_it);

  tags.clear();
  FMDB_Part_GetTag (part, tags);
  CHECK ("* Checking Restored Tag Data..", tags.size()==7);


  iterEnd = FMDB_MeshSetIter_Reset (mesh, mset_it);
  while(!iterEnd)
  {
    iterEnd = FMDB_MeshSetIter_GetNext(mesh, mset_it, eset);
    if(iterEnd) break; 
    tags.clear();
    FMDB_Set_GetTag (eset, tags);
    assert(tags.size()==7);
  }
}

  for (iType=FMDB_VERTEX; iType<FMDB_REGION; ++iType)
  {
    FMDB_Part_GetNumEnt(part, iType, FMDB_ALLTOPO, &numType);
    assert(num_type[iType] = numType);
  }

/*************************************************************************
@  CLEAN-UP
@ *************************************************************************/
  forceDel=1;
  if (!FMDB_Mesh_FindTag(mesh, "byte", byte_tag))
    CHECK ("\n* Deleting Byte Tag...", !FMDB_Mesh_DelTag (mesh, byte_tag, forceDel));
  if (!FMDB_Mesh_FindTag(mesh, "integer", int_tag))
    CHECK ("* Deleting Integer Tag...", !FMDB_Mesh_DelTag (mesh, int_tag, forceDel));
  if (!FMDB_Mesh_FindTag(mesh, "double", dbl_tag))
    CHECK ("* Deleting Double Tag...", !FMDB_Mesh_DelTag (mesh, dbl_tag, forceDel));
  if (!FMDB_Mesh_FindTag(mesh, "entity", ent_tag))
    CHECK ("* Deleting Entity Tag...", !FMDB_Mesh_DelTag (mesh, ent_tag, forceDel));
  if (!FMDB_Mesh_FindTag(mesh, "entity", ent_tag))
    CHECK ("* Deleting Set Tag...", !FMDB_Mesh_DelTag (mesh, entset_tag, forceDel));

  if (!FMDB_Mesh_FindTag(mesh, "integer array", intarr_tag))
    CHECK ("* Deleting Integer Array Tag...", !FMDB_Mesh_DelTag (mesh, intarr_tag, forceDel));
  if (!FMDB_Mesh_FindTag(mesh, "double array", dblarr_tag))
    CHECK ("* Deleting Double Array Tag...", !FMDB_Mesh_DelTag (mesh, dblarr_tag, forceDel));
  if (!FMDB_Mesh_FindTag(mesh, "entity array", entarr_tag))
    CHECK ("* Deleting Entity Array Tag...", !FMDB_Mesh_DelTag (mesh, entarr_tag, forceDel));
  if (!FMDB_Mesh_FindTag(mesh, "entity set array", entsetarr_tag))
    CHECK ("* Deleting Set Array Tag...", !FMDB_Mesh_DelTag (mesh, entsetarr_tag, forceDel));

  CHECK ("\n* Deleting FMDB Mesh Instance...", !FMDB_Mesh_Del (mesh));
  CHECK ("\n* Finalizing FMDB Service...", !SCUTIL_Finalize());

/*************************************************************************
@  PRINTING COST
@ *************************************************************************/

  printf("\n[Cost] %7.3f sec   %7.3f MB \n",  load_time+test_time+write_time, load_mem+test_mem+write_mem);
  printf("\t(mesh loading)     %7.3f sec  %7.3f MB\n", load_time, load_mem);
  printf("\t(testing)          %7.3f sec  %7.3f MB\n", test_time, test_mem);
  printf("\t(mesh writing)     %7.3f sec  %7.3f MB\n", write_time, write_mem);
  return 0;
}

int TEST_PART_SET (pMeshMdl mesh, pPart part)
{
/*************************************************************************
@  Part SET FUNCTIONS IN PART LEVEL
@
@  FMDB_Ent_GetNumAdj, FMDB_Ent_GetAdj, FMDB_Ent_Get2ndAdj
@ *************************************************************************/
  int initNumEntSet=0, isIn, numEnt;
  pEntSet eset, pset, listEntSet, setEntSet, new_set;
  std::vector<pEntSet> vecSet;
  std::vector<pEntSet>::iterator lvecIter;
  std::vector<pEntSet>::iterator svecIter;
  pPartEntIter pent_it;
  pMeshEnt ent;
  pSetEntIter esIter, esIterBeg;

  FMDB_Part_GetNumSet (part, &initNumEntSet);
  FMDB_Set_Create(mesh, part, FMDB_PSET, listEntSet);
  FMDB_Set_Create(mesh, part, FMDB_PSET, setEntSet);
  
  int isPartSet, set_type, numEntSet; 
  FMDB_Set_GetType (listEntSet, &set_type);
  CHECK ("\n* Creating Part Set...", set_type==FMDB_PSET);
  FMDB_Set_GetType (setEntSet, &set_type);
  CHECK ("* Creating Part Set...", set_type==FMDB_PSET);

  FMDB_Part_GetNumSet (part, &numEntSet);
  vecSet.clear(); 
  FMDB_Part_GetSet (part, vecSet); 
  lvecIter = std::find (vecSet.begin(), vecSet.end(), listEntSet); 
  svecIter = std::find (vecSet.begin(), vecSet.end(), setEntSet);
  CHECK ("* Checking Part Set Information...", vecSet.end()!=lvecIter && vecSet.end()!=svecIter && initNumEntSet+2==numEntSet);

  pPartSetIter pmset_it, pmset_itBeg; 
  int iterEnd = FMDB_PartSetIter_Init (part, pmset_it);
  pmset_itBeg = pmset_it; 
  int loop_counter = 0, isEnd; 
  while(!iterEnd)
  {
      iterEnd = FMDB_PartSetIter_GetNext(part, pmset_it, eset);
      if(iterEnd) break; 
      ++loop_counter; 
  }
  FMDB_PartSetIter_IsEnd (part, pmset_it, &isEnd);

  iterEnd = FMDB_PartSetIter_Reset (part, pmset_it);
  CHECK ("* Checking Part Set Traversal...", isEnd && loop_counter==numEntSet && pmset_it==pmset_itBeg);

  FMDB_Set_Create(mesh, part, FMDB_PSET, new_set);       // another set type set
  iterEnd = FMDB_PartSetIter_Reset (part, pmset_it);
  loop_counter = 0; 
  while(!iterEnd)
  {
      iterEnd = FMDB_PartSetIter_GetNext(part, pmset_it, eset);
      if(iterEnd) break; 
      ++loop_counter; 
  }
  FMDB_PartSetIter_IsEnd(part, pmset_it, &isEnd);  
  FMDB_Set_Del (mesh, part, new_set); 
  CHECK ("* Checking Part Set Traversal and Reset...", isEnd && loop_counter==numEntSet+1);

  iterEnd = FMDB_PartEntIter_Init(part, FMDB_ALLTYPE, FMDB_ALLTOPO, pent_it);
  loop_counter=0;
  while (!iterEnd)
  {  
    iterEnd = FMDB_PartEntIter_GetNext(pent_it, ent);
    if(iterEnd) break; 
    ++loop_counter;
    FMDB_Set_AddEnt (listEntSet, ent);
    FMDB_Set_AddEnt (listEntSet, ent);
    FMDB_Set_AddEnt (setEntSet, ent);
    FMDB_Set_AddEnt (setEntSet, ent);
  }
  FMDB_Ent_GetPSet (ent, pset);
  CHECK ("* Checking Entity Is In Part Set...", pset==setEntSet);
  
  FMDB_Set_GetNumEnt (listEntSet, &numEnt);
  CHECK ("* Checking Part Set Data Insertion...", numEnt==loop_counter*2);
  FMDB_Set_GetNumEnt (setEntSet, &numEnt);
  CHECK ("* Checking Part Set Data Insertion...", numEnt==loop_counter*2);

  FMDB_Set_Clr (listEntSet);
  FMDB_Set_Clr (setEntSet);
  FMDB_Ent_GetPSet (ent, pset);
  CHECK ("* Checking Entity Is In Part Set...", pset==NULL);

  int isListSetEmpty=0, isSetSetEmpty=0; 
  FMDB_Set_IsEmpty (listEntSet, &isListSetEmpty);
  FMDB_Set_IsEmpty (setEntSet, &isSetSetEmpty);
  CHECK ("* Checking Empty Set...", isListSetEmpty && isSetSetEmpty); 

  iterEnd = FMDB_PartEntIter_Reset(pent_it);
  std::vector<pMeshEnt> vecEnt;
  vecEnt.clear(); 
  while (!iterEnd)
  {  
    iterEnd = FMDB_PartEntIter_GetNext(pent_it, ent);
    if(iterEnd) break; 
    vecEnt.push_back(ent); 
    vecEnt.push_back(ent);
  }
  FMDB_Set_AddEnts (listEntSet, vecEnt);
  FMDB_Set_AddEnts (setEntSet, vecEnt);
  FMDB_Set_GetNumEnt (listEntSet, &numEnt);
  CHECK ("* Checking Part Set Data Vector Insertion...", ((size_t)numEnt)==vecEnt.size());
  FMDB_Set_GetNumEnt (setEntSet, &numEnt);
  CHECK ("* Checking Part Set Data Vector Insertion...", ((size_t)numEnt)==vecEnt.size());
 
  iterEnd = FMDB_PartEntIter_Reset(pent_it);
  int hasEnt, list_flag=1, set_flag=1;
  while (!iterEnd)
  {
    iterEnd = FMDB_PartEntIter_GetNext(pent_it, ent);
    if(iterEnd) break; 
    
    FMDB_Set_FindEnt (listEntSet, ent, &hasEnt);
    if (!hasEnt) list_flag=0;
    FMDB_Set_RmvEnt (listEntSet, ent); 
    FMDB_Set_FindEnt (listEntSet, ent, &hasEnt);
    if (!hasEnt) list_flag=0;

    FMDB_Set_FindEnt (setEntSet, ent, &hasEnt);
    if (!hasEnt) set_flag=0;
    FMDB_Set_RmvEnt (setEntSet, ent); 
    FMDB_Set_FindEnt (setEntSet, ent, &hasEnt);
    if (!hasEnt) set_flag=0;
  }
  CHECK ("* Checking Part Set Data Removal...", list_flag); 
  CHECK ("* Checking Part Set Data Removal...", set_flag);

  iterEnd = FMDB_SetEntIter_Init (listEntSet, FMDB_ALLTYPE, FMDB_POINT, esIter);
  esIterBeg = esIter;
  loop_counter=0;
  while (!iterEnd)
  {
    iterEnd = FMDB_SetEntIter_GetNext(esIter, ent);
    if(iterEnd) break; 
    ++loop_counter;
  }
  FMDB_SetEntIter_IsEnd (esIter, &isEnd);
  FMDB_SetEntIter_Reset (esIter); 
  CHECK ("\n* Checking Part Set Traversal...", isEnd && esIter==esIterBeg);
  FMDB_SetEntIter_Del (esIter);

  loop_counter=0;
  iterEnd = FMDB_SetEntIter_Init (setEntSet, FMDB_VERTEX,  FMDB_ALLTOPO, esIter);
  esIterBeg = esIter; 
  while (!iterEnd)
  {
    iterEnd = FMDB_SetEntIter_GetNext(esIter, ent);
    if(iterEnd) break; 
    ++loop_counter;
  }
  FMDB_SetEntIter_IsEnd (esIter, &isEnd);
  FMDB_SetEntIter_Reset (esIter);
  CHECK ("* Checking Part Set Traversal...", isEnd && esIter==esIterBeg);
  CHECK ("\n* Deleting Part Set Entity Iterator...", !FMDB_SetEntIter_Del (esIter));
  CHECK ("\n* Deleting Part Set...", !FMDB_Set_Del (mesh, part, setEntSet) && !    FMDB_Set_Del (mesh, part, listEntSet));

  return 0;
}

void CLEAR_MESH (pMeshMdl mesh)
{
  std::vector<pPart> parts;
  pMeshEnt ent;
  pPart part;
  pEntSet eset;
  int iterEnd, numEnt;
  std::vector<pEntSet> esets;
  std::vector<pTag> tags;

  // delete mesh sets
  pMeshSetIter mset_it;
  iterEnd = FMDB_MeshSetIter_Init(mesh, mset_it);
  while (!iterEnd)    {
    iterEnd = FMDB_MeshSetIter_GetNext(mesh, mset_it, eset);
    if(iterEnd) break;
    FMDB_Set_Del(mesh, NULL, eset);
  }
  FMDB_MeshSetIter_Del(mset_it);

  FMDB_Mesh_GetSet(mesh, esets);
  assert (esets.empty());

  FMDB_Mesh_GetAllPart (mesh, SCUTIL_CommRank(), parts);
  for (std::vector<pPart>::iterator pit=parts.begin(); pit!=parts.end(); ++pit)
  {
    part = *pit;

    pPartSetIter set_it;
    int iterEnd = FMDB_PartSetIter_Init(part, set_it);
    while (!iterEnd)
    {
      iterEnd = FMDB_PartSetIter_GetNext(part, set_it, eset);
      if(iterEnd) break;
      FMDB_Set_Del(part->getMesh(), part, eset);
    }
    FMDB_PartSetIter_Del (set_it);

    for (int i=FMDB_REGION; i>=FMDB_VERTEX; --i)
    {
      pPartEntIter ent_it;
      int iterEnd = FMDB_PartEntIter_Init(part, i, FMDB_ALLTOPO, ent_it);
      while (!iterEnd)
      {
        iterEnd = FMDB_PartEntIter_GetNext(ent_it, ent);
        if(iterEnd) break;
        FMDB_Ent_Del(part, ent);
      }
      FMDB_PartEntIter_Del(ent_it);
    }
    TEST_PART_DEL_TAG(mesh, part);
    tags.clear(); esets.clear();
    FMDB_Part_GetTag(part, tags);
    FMDB_Part_GetSet(part, esets);
    FMDB_Part_GetNumEnt(part, FMDB_ALLTYPE, FMDB_ALLTOPO, &numEnt);
    assert(esets.empty() && tags.empty() && !numEnt); 
  } // for part

  // delete tag
  tags.clear();
  FMDB_Mesh_GetTag(mesh, tags);
  for (std::vector<pTag>::iterator tag_it=tags.begin(); tag_it!=tags.end(); ++tag_it)
    FMDB_Mesh_DelTag (mesh, *tag_it, 1);   
  tags.clear();
  FMDB_Mesh_GetTag(mesh, tags);
  assert (tags.empty());
}

int TEST_PART_SETGET_TAG(pMeshMdl mesh, pPart part)
{
  pMeshEnt ent_data, ent_tag_data;
  pPartEntIter iter;
  pTag byte_tag, int_tag, dbl_tag, ent_tag, intarr_tag, dblarr_tag, entarr_tag;
  int int_data, byte_size;
  double dbl_data;
  char data[] = "abcdefg";

  FMDB_PartEntIter_Init (part, FMDB_EDGE, FMDB_ALLTOPO, iter);
  FMDB_PartEntIter_GetNext(iter, ent_tag_data); // get an entity to use as tag data
  FMDB_PartEntIter_Del(iter);

  FMDB_Mesh_FindTag(mesh, "byte", byte_tag);
  FMDB_Mesh_FindTag(mesh, "integer", int_tag);
  FMDB_Mesh_FindTag(mesh, "double", dbl_tag);
  FMDB_Mesh_FindTag(mesh, "entity", ent_tag);
  FMDB_Mesh_FindTag(mesh, "integer array", intarr_tag);
  FMDB_Mesh_FindTag(mesh, "double array", dblarr_tag);
  FMDB_Mesh_FindTag(mesh, "entity array", entarr_tag);
 
  assert(!FMDB_Part_SetIntTag (part, int_tag, 1000));
  assert( !FMDB_Part_SetDblTag (part, dbl_tag, 1000.37));
  assert( !FMDB_Part_SetEntTag (part, ent_tag, ent_tag_data));
  assert(!FMDB_Part_SetByteTag (part, byte_tag, data, 7));
  CHECK ("\n* Checking Part Set Tag Functions...", !FMDB_Part_SetIntTag (part, int_tag, 1000) && !FMDB_Part_SetDblTag (part, dbl_tag, 1000.37) &&
        !FMDB_Part_SetEntTag (part, ent_tag, ent_tag_data) && !FMDB_Part_SetByteTag (part, byte_tag, data, 7));  

  FMDB_Part_GetIntTag (part, int_tag, &int_data);
  FMDB_Part_GetDblTag (part, dbl_tag, &dbl_data);
  FMDB_Part_GetEntTag (part, ent_tag, &ent_data);

  void* byte_data = (void*)calloc(128, sizeof(char));
  FMDB_Part_GetByteTag (part, byte_tag, &byte_data, &byte_size);

  CHECK ("* Checking Part Get Tag Functions...", int_data == 1000 && dbl_data == 1000.37 && ent_data == ent_tag_data && !strcmp((char*)byte_data, data) && byte_size==7);

  // FMDB_Part_Set/GetByteTag with integer tag
  char byte_int_data[sizeof(int)];
  int* intPntData = new int(3000);
  char *charPntData = (char*)intPntData;
     
  for ( size_t i=0; i<sizeof(int); i++ )
    byte_int_data[i] =  charPntData[i];
  delete intPntData;

  FMDB_Part_SetByteTag(part, int_tag, byte_int_data, sizeof(int));
  FMDB_Part_GetByteTag (part, int_tag, &byte_data, &byte_size);
  CHECK ("* Checking Part Byte Tag Functions with Interger Tag...", 
        byte_size == (int)sizeof(int) && *(int*)byte_data == 3000);

  // FMDB_Part_Set/GetByteTag with entity tag -- taken from iMesh compliance test
  int iNumChars = sizeof(void*);
  char aChars[iNumChars];
    {
      pMeshEnt* ppvData = new pMeshEnt(ent_tag_data);
      char *pcData = (char*)ppvData;
      for ( int i=0; i<iNumChars; i++ )
        aChars[ i] =  pcData[i] ;
      delete ppvData;
    }
    FMDB_Part_SetByteTag(part, ent_tag, aChars, iNumChars);
    // ... then seeing if it comes back intact
    FMDB_Part_GetByteTag (part, ent_tag, &byte_data, &byte_size);
    CHECK ("* Checking Part Byte Tag Functions with Entity Tag...",  
           byte_size == sizeof(pMeshEnt) && ((pMeshEnt*)byte_data)[0] == ent_tag_data);

  // FMDB_Part_Set/GetIntArrTag with integer arr tag
  int int_arr[] = {4,8,12};
  int arr_size;
  FMDB_Part_SetIntArrTag (part, intarr_tag, int_arr, 3);
  int* int_arr_back = new int[4];
  FMDB_Part_GetIntArrTag (part, intarr_tag, &int_arr_back, &arr_size);
  CHECK ("* Checking Part Integer Array Tag Functions...", 
         arr_size==3 && int_arr_back[0] == 4 && int_arr_back[1] == 8 && int_arr_back[2] == 12);
            
  // FMDB_Part_Set/GetByteTag with integer arr tag
  char byte_intarr_data[sizeof(int)*3];
  charPntData = (char*)int_arr;  // {4,8,12}
     
  for ( size_t i=0; i<sizeof(int)*3; i++ )
    byte_intarr_data[i] =  charPntData[i];

  FMDB_Part_SetByteTag(part, intarr_tag, byte_intarr_data, sizeof(int)*3);
  FMDB_Part_GetByteTag (part, intarr_tag, &byte_data, &byte_size);

  CHECK ("\n* Checking Part Byte Tag Functions with Interger Arr Tag...", 
        byte_size == sizeof(int)*3 && ((int*)byte_data)[0] == 4
        && ((int*)byte_data)[1]==8 && ((int*)byte_data)[2]==12);

  // FMDB_Part_Set/GetDblArrTag
  double dbl_arr[] = {4.1,8.2,12.3};
  FMDB_Part_SetDblArrTag (part, dblarr_tag, dbl_arr, 3);
  double* dbl_arr_back = new double[4];
  FMDB_Part_GetDblArrTag (part, dblarr_tag, &dbl_arr_back, &arr_size);
  CHECK ("* Checking Part Double Array Tag Functions...", 
         arr_size==3 && dbl_arr_back[0] == 4.1 && dbl_arr_back[1] == 8.2 && dbl_arr_back[2] == 12.3);

  // FMDB_Part_Set/GetByteTag with double arr tag
  char byte_dblarr_data[sizeof(double)*3];
  charPntData = (char*)dbl_arr;  // {4.1,8.2,12.3}
     
  for ( size_t i=0; i<sizeof(double)*3; i++ )
    byte_dblarr_data[i] =  charPntData[i];

  FMDB_Part_SetByteTag(part, dblarr_tag, byte_dblarr_data, sizeof(double)*3);
  FMDB_Part_GetByteTag (part, dblarr_tag, &byte_data, &byte_size);

  CHECK ("* Checking Part Byte Tag Functions with Double Arr Tag...", 
        byte_size == sizeof(double)*3 && ((double*)byte_data)[0] == 4.1
        && ((double*)byte_data)[1]==8.2 && ((double*)byte_data)[2]==12.3);

  // FMDB_Part_Set/GetEntArrTag
  pMeshEnt* ent_arr = new pMeshEnt[3];
  ent_arr[0] = ent_arr[1] = ent_arr[2] = ent_tag_data;
  FMDB_Part_SetEntArrTag (part, entarr_tag, ent_arr, 3);
  pMeshEnt* ent_arr_back = new pMeshEnt[4];
  FMDB_Part_GetEntArrTag (part, entarr_tag, &ent_arr_back, &arr_size);
  CHECK ("* Checking Part Entity Array Tag Functions...", 
         arr_size==3 && ent_arr_back[0] == ent_tag_data && ent_arr_back[1] == ent_tag_data && ent_arr_back[2] == ent_tag_data
          && ent_arr[0]==ent_arr_back[0] && ent_arr[1]==ent_arr_back[1] && ent_arr[2] == ent_arr_back[2]);

  // FMDB_Part_Set/GetByteTag with entity arr tag
  char byte_entarr_data[sizeof(pMeshEnt)*3];
  charPntData = (char*)ent_arr;  // {ent_tag_data, ent_tag_data, ent_tag_data}
     
  for ( size_t i=0; i<sizeof(pMeshEnt)*3; i++ )
    byte_entarr_data[i] =  charPntData[i];

  FMDB_Part_SetByteTag(part, entarr_tag, byte_entarr_data, sizeof(pMeshEnt)*3);
  FMDB_Part_GetByteTag (part, entarr_tag, &byte_data, &byte_size);
  CHECK ("\n* Checking Part Byte Tag Functions with Entity Arr Tag...", 
        byte_size == sizeof(pMeshEnt)*3 && ((pMeshEnt*)byte_data)[0] == ent_tag_data
        && ((pMeshEnt*)byte_data)[1]==ent_tag_data && ((pMeshEnt*)byte_data)[2]==ent_tag_data);

  // FMDB_Part_DelTag with non-byte data skipped to test FMDB_Mesh_DelTag with forced deletion option
 // CHECK ("\n* Checking Part Tag Deletion...", !FMDB_Part_DelTag (part, byte_tag));

  delete [] ent_arr;
  delete [] int_arr_back;
  delete [] dbl_arr_back;
  delete [] ent_arr_back;
  free(byte_data);

  return 0;
}

int TEST_PART_DEL_TAG (pMeshMdl mesh, pPart part)
{
  pTag byte_tag, int_tag, dbl_tag, ent_tag, intarr_tag, dblarr_tag, entarr_tag;
  FMDB_Mesh_FindTag(mesh, "byte", byte_tag);
  FMDB_Mesh_FindTag(mesh, "integer", int_tag);
  FMDB_Mesh_FindTag(mesh, "double", dbl_tag);
  FMDB_Mesh_FindTag(mesh, "entity", ent_tag);
  FMDB_Mesh_FindTag(mesh, "integer array", intarr_tag);  
  FMDB_Mesh_FindTag(mesh, "double array", dblarr_tag);  
  FMDB_Mesh_FindTag(mesh, "entity array", entarr_tag);

  CHECK ("\n* Checking Part Tag Deletion...", !FMDB_Part_DelTag (part, byte_tag)
        && !FMDB_Part_DelTag (part, int_tag) 
        && !FMDB_Part_DelTag (part, dbl_tag) 
        && !FMDB_Part_DelTag (part, ent_tag) 
        && !FMDB_Part_DelTag (part, intarr_tag)
        && !FMDB_Part_DelTag (part, dblarr_tag)
        && !FMDB_Part_DelTag (part, entarr_tag));

  return 0;
}

int TEST_SET_SETGET_TAG (pMeshMdl mesh, pEntSet eset)
{
  pMeshEnt ent_data, ent_tag_data;
  pPartEntIter iter;
  pTag byte_tag, int_tag, dbl_tag, ent_tag, intarr_tag, dblarr_tag, entarr_tag;
  int int_data, byte_size;
  double dbl_data;
  char data[] = "abcdefg";

  std::vector<pPart> parts;
  FMDB_Mesh_GetAllPart (mesh, SCUTIL_CommRank(), parts);
  pPart part = parts.at(0);

  FMDB_PartEntIter_Init (part, FMDB_EDGE, FMDB_ALLTOPO, iter);
  FMDB_PartEntIter_GetNext(iter, ent_tag_data); // get an entity to use as tag data
  FMDB_PartEntIter_Del(iter);

  FMDB_Mesh_FindTag(mesh, "byte", byte_tag);
  FMDB_Mesh_FindTag(mesh, "integer", int_tag);
  FMDB_Mesh_FindTag(mesh, "double", dbl_tag);
  FMDB_Mesh_FindTag(mesh, "entity", ent_tag);
  FMDB_Mesh_FindTag(mesh, "integer array", intarr_tag);
  FMDB_Mesh_FindTag(mesh, "double array", dblarr_tag);
  FMDB_Mesh_FindTag(mesh, "entity array", entarr_tag);

    FMDB_Set_SetIntTag (eset, int_tag, 1000);
    FMDB_Set_SetDblTag (eset, dbl_tag, 1000.37);
    FMDB_Set_SetEntTag (eset, ent_tag, ent_tag_data);
    FMDB_Set_SetByteTag (eset, byte_tag, data, 7);  

    FMDB_Set_GetIntTag (eset, int_tag, &int_data);
    FMDB_Set_GetDblTag (eset, dbl_tag, &dbl_data);
    FMDB_Set_GetEntTag (eset, ent_tag, &ent_data);

    void* byte_data = (void*)calloc(128, sizeof(char));
    FMDB_Set_GetByteTag (eset, byte_tag, &byte_data, &byte_size);

    assert(int_data == 1000 && dbl_data == 1000.37 && ent_data == ent_tag_data && 
           !strcmp((char*)byte_data, data) && byte_size==7);

    FMDB_Set_SetIntTag (eset, int_tag, 2000);
    FMDB_Set_SetDblTag (eset, dbl_tag, 2000.37);

    FMDB_Set_GetIntTag (eset, int_tag, &int_data);
    FMDB_Set_GetDblTag (eset, dbl_tag, &dbl_data);

    assert(int_data == 2000 && dbl_data == 2000.37);

    // FMDB_Set_Set/GetByteTag with integer tag
    char byte_int_data[sizeof(int)];
    int* intPntData = new int(3000);
    char* charPntData = (char*) intPntData;
     
    for ( size_t i=0; i<sizeof(int);++i)
      byte_int_data[i] =  charPntData[i];
    delete intPntData;

    FMDB_Set_SetByteTag(eset, int_tag, byte_int_data, sizeof(int));
    FMDB_Set_GetByteTag (eset, int_tag, &byte_data, &byte_size);
    assert(byte_size == (int)sizeof(int) && *(int*)byte_data == 3000);

    // FMDB_Set_Set/GetByteTag with entity tag -- taken from iMesh compliance test
    int iNumChars = sizeof(void*);
    char aChars[iNumChars];
    {
      pMeshEnt* ppvData = new pMeshEnt(ent_tag_data);
      char *pcData = (char*)ppvData;
      for ( int i=0; i<iNumChars; i++ )
        aChars[ i] =  pcData[i] ;
      delete ppvData;
    }
    FMDB_Set_SetByteTag(eset, ent_tag, aChars, iNumChars);
    // ... then seeing if it comes back intact
    FMDB_Set_GetByteTag (eset, ent_tag, &byte_data, &byte_size);
    assert(byte_size == sizeof(pMeshEnt) && ((pMeshEnt*)byte_data)[0] == ent_tag_data);

  // FMDB_Set_Set/GetIntArrTag with integer arr tag
  int int_arr[] = {4,8,12};
  int arr_size;
  FMDB_Set_SetIntArrTag (eset, intarr_tag, int_arr, 3);
  int* int_arr_back = new int[4];
  FMDB_Set_GetIntArrTag (eset, intarr_tag, &int_arr_back, &arr_size);
  assert(arr_size==3 && int_arr_back[0] == 4 && int_arr_back[1] == 8 && int_arr_back[2] == 12);
            
  // FMDB_Set_Set/GetByteTag with integer arr tag
  char byte_intarr_data[sizeof(int)*3];
  charPntData = (char*)int_arr;  // {4,8,12}
     
  for ( size_t i=0; i<sizeof(int)*3; i++ )
    byte_intarr_data[i] =  charPntData[i];

  FMDB_Set_SetByteTag(eset, intarr_tag, byte_intarr_data, sizeof(int)*3);
  FMDB_Set_GetByteTag (eset, intarr_tag, &byte_data, &byte_size);

  assert(byte_size == sizeof(int)*3 && ((int*)byte_data)[0] == 4
        && ((int*)byte_data)[1]==8 && ((int*)byte_data)[2]==12);

  // FMDB_Set_Set/GetDblArrTag
  double dbl_arr[] = {4.1,8.2,12.3};
  FMDB_Set_SetDblArrTag (eset, dblarr_tag, dbl_arr, 3);
  double* dbl_arr_back = new double[4];
  FMDB_Set_GetDblArrTag (eset, dblarr_tag, &dbl_arr_back, &arr_size);
  assert(arr_size==3 && dbl_arr_back[0] == 4.1 && dbl_arr_back[1] == 8.2 && dbl_arr_back[2] == 12.3);

  // FMDB_Set_Set/GetByteTag with double arr tag
  char byte_dblarr_data[sizeof(double)*3];
  charPntData = (char*)dbl_arr;  // {4.1,8.2,12.3}
     
  for ( size_t i=0; i<sizeof(double)*3; i++ )
    byte_dblarr_data[i] =  charPntData[i];

  FMDB_Set_SetByteTag(eset, dblarr_tag, byte_dblarr_data, sizeof(double)*3);
  FMDB_Set_GetByteTag (eset, dblarr_tag, &byte_data, &byte_size);

  assert(byte_size == sizeof(double)*3 && ((double*)byte_data)[0] == 4.1
        && ((double*)byte_data)[1]==8.2 && ((double*)byte_data)[2]==12.3);

  // FMDB_Set_Set/GetEntArrTag
  pMeshEnt* ent_arr = new pMeshEnt[3];
  ent_arr[0] = ent_arr[1] = ent_arr[2] = ent_tag_data;
  FMDB_Set_SetEntArrTag (eset, entarr_tag, ent_arr, 3);
  pMeshEnt* ent_arr_back = new pMeshEnt[4];
  FMDB_Set_GetEntArrTag (eset, entarr_tag, &ent_arr_back, &arr_size);
  assert(arr_size==3 && ent_arr_back[0] == ent_tag_data && ent_arr_back[1] == ent_tag_data && ent_arr_back[2] == ent_tag_data
          && ent_arr[0]==ent_arr_back[0] && ent_arr[1]==ent_arr_back[1] && ent_arr[2] == ent_arr_back[2]);

  // FMDB_Set_Set/GetByteTag with entity arr tag
  char byte_entarr_data[sizeof(pMeshEnt)*3];
  charPntData = (char*)ent_arr;  // {ent_tag_data, ent_tag_data, ent_tag_data}
     
  for ( size_t i=0; i<sizeof(pMeshEnt)*3; i++ )
    byte_entarr_data[i] =  charPntData[i];

  FMDB_Set_SetByteTag(eset, entarr_tag, byte_entarr_data, sizeof(pMeshEnt)*3);
  FMDB_Set_GetByteTag (eset, entarr_tag, &byte_data, &byte_size);
  assert(byte_size == sizeof(pMeshEnt)*3 && ((pMeshEnt*)byte_data)[0] == ent_tag_data
        && ((pMeshEnt*)byte_data)[1]==ent_tag_data && ((pMeshEnt*)byte_data)[2]==ent_tag_data);


  assert(!FMDB_Set_DelTag (eset, byte_tag)
        && !FMDB_Set_DelTag (eset, int_tag) 
        && !FMDB_Set_DelTag (eset, dbl_tag) 
        && !FMDB_Set_DelTag (eset, ent_tag) 
        && !FMDB_Set_DelTag (eset, intarr_tag)
        && !FMDB_Set_DelTag (eset, dblarr_tag)
        && !FMDB_Set_DelTag (eset, entarr_tag));

    free (byte_data);
    delete [] int_arr_back;
    delete [] dbl_arr_back;
    delete [] ent_arr;
    delete [] ent_arr_back;

    return 0;
}

int TEST_SET_DEL_TAG (pMeshMdl mesh, pEntSet eset)
{
  pTag byte_tag, int_tag, dbl_tag, ent_tag, intarr_tag, dblarr_tag, entarr_tag;
  FMDB_Mesh_FindTag(mesh, "byte", byte_tag);
  FMDB_Mesh_FindTag(mesh, "integer", int_tag);
  FMDB_Mesh_FindTag(mesh, "double", dbl_tag);
  FMDB_Mesh_FindTag(mesh, "entity", ent_tag);
  FMDB_Mesh_FindTag(mesh, "integer array", intarr_tag);
  FMDB_Mesh_FindTag(mesh, "double array", dblarr_tag);
  FMDB_Mesh_FindTag(mesh, "entity array", entarr_tag);

  CHECK ("\n* Checking Part Tag Deletion...", !FMDB_Set_DelTag (eset, byte_tag)
        && !FMDB_Set_DelTag (eset, int_tag)
        && !FMDB_Set_DelTag (eset, dbl_tag)
        && !FMDB_Set_DelTag (eset, ent_tag)
        && !FMDB_Set_DelTag (eset, intarr_tag)
        && !FMDB_Set_DelTag (eset, dblarr_tag)
        && !FMDB_Set_DelTag (eset, entarr_tag));

  return 0;
}

int TEST_ENT_SETGET_TAG (pMeshMdl mesh, pMeshEnt ent)
{
  pMeshEnt ent_data, ent_tag_data;
  pPartEntIter iter;
  pTag byte_tag, int_tag, dbl_tag, ent_tag, intarr_tag, dblarr_tag, entarr_tag;
  int int_data, byte_size;
  double dbl_data;
  char data[] = "abcdefg";

  FMDB_Mesh_FindTag(mesh, "byte", byte_tag);
  FMDB_Mesh_FindTag(mesh, "integer", int_tag);
  FMDB_Mesh_FindTag(mesh, "double", dbl_tag);
  FMDB_Mesh_FindTag(mesh, "entity", ent_tag);
  FMDB_Mesh_FindTag(mesh, "integer array", intarr_tag);
  FMDB_Mesh_FindTag(mesh, "double array", dblarr_tag);
  FMDB_Mesh_FindTag(mesh, "entity array", entarr_tag);

  std::vector<pPart> parts;
  FMDB_Mesh_GetAllPart (mesh, SCUTIL_CommRank(), parts);

  FMDB_PartEntIter_Init (parts.at(0), FMDB_EDGE, FMDB_ALLTOPO, iter);
  FMDB_PartEntIter_GetNext(iter, ent_tag_data); // get an entity to use as tag data
  FMDB_PartEntIter_Del(iter);

    FMDB_Ent_SetIntTag (mesh, ent, int_tag, 1000);
    FMDB_Ent_SetDblTag (mesh, ent, dbl_tag, 1000.37);
    FMDB_Ent_SetEntTag (mesh, ent, ent_tag, ent_tag_data);
    FMDB_Ent_SetByteTag (mesh, ent, byte_tag, data, 7);  

    FMDB_Ent_GetIntTag (mesh, ent, int_tag, &int_data);
    FMDB_Ent_GetDblTag (mesh, ent, dbl_tag, &dbl_data);
    FMDB_Ent_GetEntTag (mesh, ent, ent_tag, &ent_data);

    void* byte_data = (void*)calloc(128, sizeof(char));
    FMDB_Ent_GetByteTag (mesh, ent, byte_tag, &byte_data, &byte_size);

    assert(int_data == 1000 && dbl_data == 1000.37 && ent_data == ent_tag_data && 
           !strcmp((char*)byte_data, data) && byte_size==7);

    // FMDB_Ent_Set/GetByteTag with integer tag
    char byte_int_data[sizeof(int)];
    int* intPntData = new int(3000);
    char* charPntData = (char*) intPntData;
     
    for ( size_t i=0; i<sizeof(int);++i)
      byte_int_data[i] =  charPntData[i];
    delete intPntData;

    FMDB_Ent_SetByteTag(mesh, ent, int_tag, byte_int_data, sizeof(int));
    FMDB_Ent_GetByteTag (mesh, ent, int_tag, &byte_data, &byte_size);
    assert(byte_size == (int)sizeof(int) && *(int*)byte_data == 3000);

    // FMDB_Ent_Set/GetByteTag with entity tag -- taken from iMesh compliance test
    int iNumChars = sizeof(void*);
    char aChars[iNumChars];
    {
      pMeshEnt* ppvData = new pMeshEnt(ent_tag_data);
      char *pcData = (char*)ppvData;
      for ( int i=0; i<iNumChars; i++ )
        aChars[ i] =  pcData[i] ;
      delete ppvData;
    }
    FMDB_Ent_SetByteTag(mesh, ent, ent_tag, aChars, iNumChars);
    // ... then seeing if it comes back intact
    FMDB_Ent_GetByteTag (mesh, ent, ent_tag, &byte_data, &byte_size);
    assert(byte_size == sizeof(pMeshEnt) && ((pMeshEnt*)byte_data)[0] == ent_tag_data);

   // FMDB_Ent_Set/GetIntArrTag with integer arr tag
    int int_arr[] = {4,8,12};
    int arr_size;
    FMDB_Ent_SetIntArrTag (mesh, ent, intarr_tag, int_arr, 3);
    int* int_arr_back = new int[4];
    FMDB_Ent_GetIntArrTag (mesh, ent, intarr_tag, &int_arr_back, &arr_size);
    assert(arr_size==3 && int_arr_back[0] == 4 && int_arr_back[1] == 8 && int_arr_back[2] == 12);
            
    // FMDB_Ent_Set/GetByteTag with integer arr tag
    char byte_intarr_data[sizeof(int)*3];
    charPntData = (char*)int_arr;  // {4,8,12}
     
    for ( size_t i=0; i<sizeof(int)*3; i++ )
      byte_intarr_data[i] =  charPntData[i];

    FMDB_Ent_SetByteTag(mesh, ent, intarr_tag, byte_intarr_data, sizeof(int)*3);
    FMDB_Ent_GetByteTag (mesh, ent, intarr_tag, &byte_data, &byte_size);

    assert(byte_size == sizeof(int)*3 && ((int*)byte_data)[0] == 4
        && ((int*)byte_data)[1]==8 && ((int*)byte_data)[2]==12);

    // FMDB_Ent_Set/GetDblArrTag
    double dbl_arr[] = {4.1,8.2,12.3};
    FMDB_Ent_SetDblArrTag (mesh, ent, dblarr_tag, dbl_arr, 3);
    double* dbl_arr_back = new double[4];
    FMDB_Ent_GetDblArrTag (mesh, ent, dblarr_tag, &dbl_arr_back, &arr_size);
    assert(arr_size==3 && dbl_arr_back[0] == 4.1 && dbl_arr_back[1] == 8.2 && 
           dbl_arr_back[2] == 12.3);

    // FMDB_Ent_Set/GetByteTag with double arr tag
    char byte_dblarr_data[sizeof(double)*3];
    charPntData = (char*)dbl_arr;  // {4.1,8.2,12.3}
     
    for ( size_t i=0; i<sizeof(double)*3; i++ )
      byte_dblarr_data[i] =  charPntData[i];

    FMDB_Ent_SetByteTag(mesh, ent, dblarr_tag, byte_dblarr_data, sizeof(double)*3);
    FMDB_Ent_GetByteTag (mesh, ent, dblarr_tag, &byte_data, &byte_size);

    assert(byte_size == sizeof(double)*3 && ((double*)byte_data)[0] == 4.1
          && ((double*)byte_data)[1]==8.2 && ((double*)byte_data)[2]==12.3);

    // FMDB_Ent_Set/GetEntArrTag
    pMeshEnt* ent_arr = new pMeshEnt[3];
    ent_arr[0] = ent_arr[1] = ent_arr[2] = ent_tag_data;
    FMDB_Ent_SetEntArrTag (mesh, ent, entarr_tag, ent_arr, 3);
    pMeshEnt* ent_arr_back = new pMeshEnt[4];
    FMDB_Ent_GetEntArrTag (mesh, ent, entarr_tag, &ent_arr_back, &arr_size);
    assert(arr_size==3 && ent_arr_back[0] == ent_tag_data && ent_arr_back[1] == 
           ent_tag_data && ent_arr_back[2] == ent_tag_data
           && ent_arr[0]==ent_arr_back[0] && ent_arr[1]==ent_arr_back[1] && 
           ent_arr[2] == ent_arr_back[2]);

    // FMDB_Ent_Set/GetByteTag with entity arr tag
    char byte_entarr_data[sizeof(pMeshEnt)*3];
    charPntData = (char*)ent_arr;  // {ent_tag_data, ent_tag_data, ent_tag_data}
     
    for ( size_t i=0; i<sizeof(pMeshEnt)*3; i++ )
      byte_entarr_data[i] =  charPntData[i];
  
    FMDB_Ent_SetByteTag(mesh, ent, entarr_tag, byte_entarr_data, sizeof(pMeshEnt)*3);
    FMDB_Ent_GetByteTag (mesh, ent, entarr_tag, &byte_data, &byte_size);
    assert(byte_size == sizeof(pMeshEnt)*3 && ((pMeshEnt*)byte_data)[0] == 
           ent_tag_data && ((pMeshEnt*)byte_data)[1]==ent_tag_data && 
           ((pMeshEnt*)byte_data)[2]==ent_tag_data);

    free (byte_data);
    delete [] int_arr_back;
    delete [] dbl_arr_back;
    delete [] ent_arr;
    delete [] ent_arr_back;
    return 0;
}

int TEST_ENT_DEL_TAG (pMeshMdl mesh, pMeshEnt ent)
{
  pTag byte_tag, int_tag, dbl_tag, ent_tag, intarr_tag, dblarr_tag, entarr_tag;
  FMDB_Mesh_FindTag(mesh, "byte", byte_tag);
  FMDB_Mesh_FindTag(mesh, "integer", int_tag);
  FMDB_Mesh_FindTag(mesh, "double", dbl_tag);
  FMDB_Mesh_FindTag(mesh, "entity", ent_tag);
  FMDB_Mesh_FindTag(mesh, "integer array", intarr_tag);
  FMDB_Mesh_FindTag(mesh, "double array", dblarr_tag);
  FMDB_Mesh_FindTag(mesh, "entity array", entarr_tag);

  assert(!FMDB_Ent_DelTag (ent, byte_tag)
        && !FMDB_Ent_DelTag (ent, int_tag)
        && !FMDB_Ent_DelTag (ent, dbl_tag)
        && !FMDB_Ent_DelTag (ent, ent_tag)
        && !FMDB_Ent_DelTag (ent, intarr_tag)
        && !FMDB_Ent_DelTag (ent, dblarr_tag)
        && !FMDB_Ent_DelTag (ent, entarr_tag));

  return 0;
}

int TEST_MESH_SET(pMeshMdl mesh, pPart part)
{
  // create three mesh set
  int hasEnt, initNumEntSet, set_type, numEntSet;
  pPartEntIter pent_it;
  pMeshEnt ent;
  int numEnt;
  FMDB_Mesh_GetNumSet (mesh, &initNumEntSet);
  pEntSet listEntSet;
  pEntSet setEntSet;
  FMDB_Set_Create(mesh, part, FMDB_LSET, listEntSet);
  FMDB_Set_Create(mesh, part, FMDB_SSET, setEntSet);

  FMDB_Set_GetType (listEntSet, &set_type);
  CHECK ("\n* Creating List-type Set...", set_type==FMDB_LSET);
  FMDB_Set_GetType (setEntSet, &set_type);
  CHECK ("* Creating Set-type Set...", set_type==FMDB_SSET);

  FMDB_Mesh_GetNumSet (mesh, &numEntSet);

  std::vector<pEntSet> vecSet;
  FMDB_Mesh_GetSet (mesh, vecSet);
  std::vector<pEntSet>::iterator lvecIter, svecIter;
  lvecIter = std::find (vecSet.begin(), vecSet.end(), listEntSet);
  svecIter = std::find (vecSet.begin(), vecSet.end(), setEntSet);
  CHECK ("* Checking Mesh Set Information...", vecSet.end()!=lvecIter && vecSet.end()!=svecIter && initNumEntSet+2==numEntSet);

  pMeshSetIter mset_it, mset_itBeg;
  pEntSet eset;
  int iterEnd = FMDB_MeshSetIter_Init (mesh, mset_it);
  mset_itBeg = mset_it;
  int loop_counter = 0;
  while(!iterEnd)
  {
      iterEnd = FMDB_MeshSetIter_GetNext(mesh, mset_it, eset);
      if(iterEnd) break;
      ++loop_counter;
  }
  int isEnd;
  FMDB_MeshSetIter_IsEnd(mesh, mset_it, &isEnd);
  CHECK ("* Checking Mesh Set Traversal...", isEnd && loop_counter==numEntSet && mset_it==mset_itBeg);

  pEntSet new_set;
  FMDB_Set_Create(mesh, part, FMDB_SSET, new_set);       // another set type set

  // do manipulation
  FMDB_PartEntIter_Init(part,FMDB_ALLTYPE,FMDB_ALLTOPO,pent_it);
  iterEnd = FMDB_PartEntIter_Reset(pent_it);
  while (!iterEnd)
  {  
    iterEnd = FMDB_PartEntIter_GetNext(pent_it, ent);
    if(iterEnd) break; 
    FMDB_Set_AddEnt (listEntSet, ent);
    FMDB_Set_AddEnt (listEntSet, ent);
    FMDB_Set_AddEnt (setEntSet, ent);
    FMDB_Set_AddEnt (setEntSet, ent);
  }
  FMDB_Set_GetNumEnt (listEntSet, &numEnt);
  FMDB_Set_GetNumEnt (setEntSet, &numEnt);

  FMDB_Set_Clr (listEntSet);
  FMDB_Set_Clr (setEntSet);
  int isListSetEmpty=0, isSetSetEmpty=0; 
  FMDB_Set_IsEmpty (listEntSet, &isListSetEmpty);
  FMDB_Set_IsEmpty (setEntSet, &isSetSetEmpty);
  CHECK ("* Checking Empty Set...", isListSetEmpty && isSetSetEmpty); 

  std::vector<pMeshEnt> vecEnt; 
  iterEnd = FMDB_PartEntIter_Reset(pent_it);
  while (!iterEnd)
  {  
    iterEnd = FMDB_PartEntIter_GetNext(pent_it, ent);
    if(iterEnd) break; 
    vecEnt.push_back(ent); 
    vecEnt.push_back(ent);
  }
  FMDB_Set_AddEnts (listEntSet, vecEnt);
  FMDB_Set_AddEnts (setEntSet, vecEnt);
  FMDB_Set_GetNumEnt (listEntSet, &numEnt);
  FMDB_Set_GetNumEnt (setEntSet, &numEnt);
 
  iterEnd = FMDB_PartEntIter_Reset(pent_it);
  int list_flag=1, set_flag=1;
  while (!iterEnd)
  {
    iterEnd = FMDB_PartEntIter_GetNext(pent_it, ent);
    if(iterEnd) break; 
    FMDB_Set_FindEnt (listEntSet, ent, &hasEnt);
    if (!hasEnt) list_flag=0;
    FMDB_Set_RmvEnt (listEntSet, ent); 
    FMDB_Set_FindEnt (listEntSet, ent, &hasEnt);
    if (!hasEnt) list_flag=0;

    FMDB_Set_FindEnt (setEntSet, ent, &hasEnt);
    if (!hasEnt) set_flag=0;
    FMDB_Set_RmvEnt (setEntSet, ent); 
    FMDB_Set_FindEnt (setEntSet, ent, &hasEnt);
    if (hasEnt) set_flag=0;
  }
  CHECK ("* Checking List-type Set Data Removal...", list_flag); 
  CHECK ("* Checking Set-type Set Data Removal...", set_flag);

  loop_counter=0;
  pSetEntIter esIter;
  iterEnd = FMDB_SetEntIter_Init (setEntSet, FMDB_VERTEX,  FMDB_ALLTOPO, esIter);
  pSetEntIter esIterBeg = esIter; 
  while (!iterEnd)
  {
    iterEnd = FMDB_SetEntIter_GetNext(esIter, ent);
    if(iterEnd) break; 
    ++loop_counter;
  }
  FMDB_SetEntIter_IsEnd (esIter, &isEnd);
  FMDB_SetEntIter_Del (esIter);
  CHECK ("* Checking Set-type Set Traversal...", isEnd && (loop_counter==0) && esIter==esIterBeg);
  CHECK ("\n* Deleting Set Entity Iterator...", !FMDB_SetEntIter_Del (esIter));

  iterEnd = FMDB_MeshSetIter_Reset (mesh, mset_it);
  while(!iterEnd)
  {
      iterEnd = FMDB_MeshSetIter_GetNext(mesh, mset_it, eset);
      if(iterEnd) break;
      FMDB_Set_Clr(eset);
  }

  FMDB_MeshSetIter_Del(mset_it);

  return 0;
}
